#!/usr/bin/python

########################################################################
#
# tbnoiseped_plots.py
#
# Creates plots from the output files produced by the NA60 or
# TELL1 data anaylis during VELO production in Liverpool.
# The plots follow the same color convention as the plots created
# by the VELO production data base frontend.  You can invoke this
# script manually for debugging purposes.  In the standard use case
# it will be called by the steering script that runs the analysis.
#
# Usage:
#
#   tbnoiseped_plots.py  -tell1 <tell1> -indir <dir> [-outdir <dir>]
#
# 2006-11-11, Kurt Rinnert, <kurt.rinnert@cern.ch>
#
########################################################################

from sys import argv
from sys import exit
from os import system


########## CONFIG ##########

########################################################################
#
# histogram appearance configuration
#
defaultColor = 3  # green
mod32Color   = 1  # black
badColor     = 2  # red
markerStyle  = 20 # filled circle
markerSize   = 0.3
bgColor      = 10

noiseMin = 0
noiseMax = 10
pedMin   = 480
pedMax   = 580

xTitle = "chip channel"
yTitle = "ADC"

########################################################################
#
# exit codes
#
ERR_BAD_ARG = 61
ERR_BAD_DIR = 62
ERR_BAD_NUM = 63

########################################################################
#
# command line cinfiguration cache 
#
inDir   = ''
outDir  = ''
tell1   = ''
markBad = 0

########################################################################
#
# globals to be read form input files and related
# configuration
#
nChips  = 16
nStrips = nChips*128

badChannels = [ [] for chip in range(nChips)]
strip       = [ [] for chip in range(nChips)]
noise       = [ [] for chip in range(nChips)]
rawNoise    = [ [] for chip in range(nChips)]
pedestal    = [ [] for chip in range(nChips)]


########## FUNCTIONS ##########

########################################################################
#
# printUsage()
#
# print command line usage to console
#
def printUsage():
    print '\nUsage:'
    print '\ttbnoiseped_plots.py -tell1 <tell1> -indir <dir> [-outdir <dir>]'
    print ''
# end of printUsage()

########################################################################
#
# parseArgs()
#
# parse argument and perform some simple validity checks
#
def parseArgs():
    global inDir;  
    global outDir; 
    global tell1; 
    global markBad
    
    #
    # parse command line arguments
    #
    switches = {'-indir':"", '-outdir':"",'-tell1':"",'-markbad':"1"}

    onTags     = ('on','On','ON','1','y','Y','yes','Yes','YES','enable','Enable','ENABLE')
    offTags    = ('off','Off','OFF','0','n','N','no','No','NO','disable','Disable','DISABLE')

    while len(argv) > 2:
        if argv[1] in switches:
            switch = argv.pop(1)
            value  = argv.pop(1)
            switches[switch] = value
        else:
            break
        
    inDir   = switches['-indir']
    outDir  = switches['-outdir']
    
    tell1  = switches['-tell1']
		
    markBad = switches['-markbad']
    if markBad in onTags:
        markBad = 1
    elif markBad in offTags:
        markBad = 0
    else:
        printUsage()
        exit(ERR_BAD_ARG)

    if len(outDir) == 0:
        outDir = inDir
# end of parseArgs()

########################################################################
#
# importROOTBindingsAndConf()
#
# import ROOT bindings
# this has to be done *after* parsing the arguments
# and removing all arguments handled by this script from argv[].
# otherwise ROOT will change the current working directory
# if one argument in the list is a directory name or give you some
# other nasty surprises.
#
def importROOTBindingsAndConf():
    global TH1F
    global TCanvas
    global gROOT
    global gStyle
    
    from ROOT import TH1F
    from ROOT import TCanvas
    from ROOT import gROOT
    from ROOT import gStyle


    #
    # global Root configuration
    #
    gROOT.SetBatch()
    gROOT.SetStyle("Plain")
    gStyle.SetOptStat(0)
    gStyle.SetMarkerStyle(markerStyle)
    gStyle.SetMarkerColor(defaultColor)
    gStyle.SetMarkerSize(markerSize)
    gStyle.SetCanvasColor(bgColor)
    gStyle.SetTitleFillColor(bgColor)
    gStyle.SetTitleBorderSize(0)
# end of importROOTBindingsAndConf()

########################################################################
#
# readBadChannels()
#
# produces the noise plots for a the chip given as argument
#
def readBadChannels():
    global badChannels
    global tell1
    global inDir

    #
    # the bad channel list will be read from a
    # file expected to follow the
    # standard naming convention and exist in
    # the same directory as the input data file
    #
    tagFilePath = inDir + "/TELL1_%s_summary.dat" % (tell1)

    tagFile = open(tagFilePath)
    for line in tagFile:
        if line.startswith("#"):
					continue
        values = line.split()
        chip = int(values[1])
        tag = int(values[3])
        if tag < 0:
          badChannels[chip].append(int(values[2]))
    tagFile.close()
# end of readBadChannels()

########################################################################
#
# perChipPlots(chip)
#
# produces the noise plots for a the chip given as argument
#
def perChipPlots(chip):
    global tell1
    global badChannels
    global strip
    global noise
    global rawNoise
    global pedestal

    #
    # read the input file, making sure to skip the comments
    # in the beginning
    #
    fileName = "TELL1_%s_noise_chip%i.dat" % (tell1,chip)
    filePath = inDir + "/" + fileName

    datFile = open(filePath)
    for line in datFile:
        if line.startswith("#"):
            continue
        values = line.split()
        strip[chip].append(int(values[0]))
        noise[chip].append(float(values[3]))
        pedestal[chip].append(float(values[4]))
        rawNoise[chip].append(float(values[5]))
    datFile.close()

    #
    # create and fill the histograms.
    # there are separate histograms for channels 0,32,64,128
    # as well as for channels flagged as bad, this is necessary
    # to display these in a different color.
    #

    #
    # raw noise
    #
    rawNoiseName    = "rawNoiseH%i" % (chip)
    rawNoiseName32  = "rawNoiseH32%i" % (chip)
    rawNoiseNameBad = "rawNoiseHBad%i" % (chip)
    rawNoiseTitle   = "raw noise,TELL1 %s CHIP %s" % (tell1,chip)
    rawNoiseEntries = len(rawNoise[chip])
    rawNoiseH    = TH1F(rawNoiseName,rawNoiseTitle,rawNoiseEntries,-0.5,rawNoiseEntries-0.5)
    rawNoiseH32  = TH1F(rawNoiseName32,rawNoiseTitle,rawNoiseEntries,-0.5,rawNoiseEntries-0.5)
    rawNoiseHBad = TH1F(rawNoiseNameBad,rawNoiseTitle,rawNoiseEntries,-0.5,rawNoiseEntries-0.5)
    for channel in range(rawNoiseEntries):
        if badChannels[chip].count(channel) != 0:
            rawNoiseHBad.SetBinContent(channel+1,rawNoise[chip][channel])
            rawNoiseH.SetBinContent(channel+1,noiseMin-1)
            rawNoiseH32.SetBinContent(channel+1,noiseMin-1)
        elif channel%32 == 0:
            rawNoiseH32.SetBinContent(channel+1,rawNoise[chip][channel])
            rawNoiseH.SetBinContent(channel+1,noiseMin-1)
            rawNoiseHBad.SetBinContent(channel+1,noiseMin-1)
        else: 
            rawNoiseH.SetBinContent(channel+1,rawNoise[chip][channel])
            rawNoiseHBad.SetBinContent(channel+1,noiseMin-1)
            rawNoiseH32.SetBinContent(channel+1,noiseMin-1)
    rawNoiseH32.SetMarkerColor(mod32Color)
    rawNoiseHBad.SetMarkerColor(badColor)

    #
    # noise
    #
    noiseName    = "noiseH%i" % (chip)
    noiseName32  = "noiseH32%i" % (chip)
    noiseNameBad = "noiseHBad%i" % (chip)
    noiseTitle   = "noise,TELL1 %s CHIP %s" % (tell1,chip)
    noiseEntries = len(noise[chip])
    noiseH    = TH1F(noiseName,noiseTitle,noiseEntries,-0.5,noiseEntries)
    noiseH32  = TH1F(noiseName32,noiseTitle,noiseEntries,-0.5,noiseEntries)
    noiseHBad = TH1F(noiseNameBad,noiseTitle,noiseEntries,-0.5,noiseEntries)
    for channel in range(noiseEntries):
        if badChannels[chip].count(channel) != 0:
            noiseHBad.SetBinContent(channel+1,noise[chip][channel])
            noiseH.SetBinContent(channel+1,noiseMin-1)
            noiseH32.SetBinContent(channel+1,noiseMin-1)
        elif channel%32 == 0:
            noiseH32.SetBinContent(channel+1,noise[chip][channel])
            noiseH.SetBinContent(channel+1,noiseMin-1)
            noiseHBad.SetBinContent(channel+1,noiseMin-1)
        else: 
            noiseH.SetBinContent(channel+1,noise[chip][channel])
            noiseHBad.SetBinContent(channel+1,noiseMin-1)
            noiseH32.SetBinContent(channel+1,noiseMin-1)
    noiseH32.SetMarkerColor(mod32Color)
    noiseHBad.SetMarkerColor(badColor)

    #
    # pedestal
    #
    pedName    = "pedH%i" % (chip)
    pedName32  = "pedH32%i" % (chip)
    pedNameBad = "pedHBad%i" % (chip)
    pedTitle   = "pedestals,TELL1 %s CHIP %s" % (tell1,chip)
    pedEntries = len(pedestal[chip])
    pedH    = TH1F(pedName,pedTitle,pedEntries,-0.5,pedEntries)
    pedH32  = TH1F(pedName32,pedTitle,pedEntries,-0.5,pedEntries)
    pedHBad = TH1F(pedNameBad,pedTitle,pedEntries,-0.5,pedEntries)
    for channel in range(pedEntries):
        if badChannels[chip].count(channel) != 0:
            pedHBad.SetBinContent(channel+1,pedestal[chip][channel])
            pedH.SetBinContent(channel+1,pedMin-1)
            pedH32.SetBinContent(channel+1,pedMin-1)
        elif channel%32 == 0:
            pedH32.SetBinContent(channel+1,pedestal[chip][channel])
            pedH.SetBinContent(channel+1,pedMin-1)
            pedHBad.SetBinContent(channel+1,pedMin-1)
        else:
            pedH.SetBinContent(channel+1,pedestal[chip][channel])
            pedHBad.SetBinContent(channel+1,pedMin-1)
            pedH32.SetBinContent(channel+1,pedMin-1)
    pedH32.SetMarkerColor(mod32Color)
    pedHBad.SetMarkerColor(badColor)

    #
    # settings common to groups of histograms
    #
    noiseHistos = [rawNoiseH,rawNoiseH32,rawNoiseHBad,noiseH,noiseH32,noiseHBad]
    pedHistos   = [pedH,pedH32,pedHBad]
    histos      = noiseHistos + pedHistos
    for h in histos:
        h.SetXTitle(xTitle)
        h.SetYTitle(yTitle)
    for h in noiseHistos:
        h.SetMaximum(noiseMax)
        h.SetMinimum(noiseMin)
    for h in pedHistos:
        h.SetMaximum(pedMax)
        h.SetMinimum(pedMin)
    #
    # draw the histograms
    #
    cvName  = "cv%i" % (chip)
    cvTitle = "Pedestals and Noise, TELL1 %s CHIP %s" % (tell1,chip)
    cv = TCanvas(cvName,cvTitle,900,300)

    cv.Divide(3,1)
    cv.cd(1)
    rawNoiseH.Draw("P")
    rawNoiseH32.Draw("PSame")
    rawNoiseHBad.Draw("PSame")
    cv.cd(2)
    noiseH.Draw("P")
    noiseH32.Draw("PSame")
    noiseHBad.Draw("PSame")
    cv.cd(3)
    pedH.Draw("P")
    pedH32.Draw("PSame")
    pedHBad.Draw("PSame")
    cv.Update()

    #
    # save plots to file
    #
    plotFileNameEps = outDir + "/" + fileName[:fileName.rfind(".dat")] + ".eps"
    plotFileNamePng = outDir + "/" + fileName[:fileName.rfind(".dat")] + ".png"
    cv.SaveAs(plotFileNameEps)
    epsToPngCmd = "convert " + plotFileNameEps + " " + plotFileNamePng
    system(epsToPngCmd)
# end of perChipPlots(chip)

########################################################################
#
# perSensorPlots()
#
# creates raw noise, noise and pedestal plots versus
# software strip number for the whole sensor
#
def perSensorPlots():
    global outDir
    global hybrid
    global sensor
    global status
    global badChannels
    global strip
    global rawNoise
    global noise
    global pedestal

    #
    # create histograms
    #
    rawNoiseH    = TH1F("rawNoiseH","raw noise vs. strip, TELL1 %s" % (tell1),nStrips,-0.5,nStrips-0.5)
    rawNoiseH32  = TH1F("rawNoiseH32","raw noise vs. strip, TELL1 %s" % (tell1),nStrips,-0.5,nStrips-0.5)
    rawNoiseHBad = TH1F("rawNoiseHBad","raw noise vs. strip, TELL1 %s" % (tell1),nStrips,-0.5,nStrips-0.5)
    rawNoiseH32.SetMarkerColor(mod32Color)
    rawNoiseHBad.SetMarkerColor(badColor)
    noiseH       = TH1F("noiseH","noise vs. strip, TELL1 %s" % (tell1),nStrips,-0.5,nStrips-0.5)
    noiseH32     = TH1F("noiseH32","noise vs. strip, TELL1 %s" % (tell1),nStrips,-0.5,nStrips-0.5)
    noiseHBad    = TH1F("noiseHBad","noise vs. strip, TELL1 %s" % (tell1),nStrips,-0.5,nStrips-0.5)
    noiseH32.SetMarkerColor(mod32Color)
    noiseHBad.SetMarkerColor(badColor)
    pedH         = TH1F("pedH","pedestal vs. strip, TELL1 %s" % (tell1),nStrips,-0.5,nStrips-0.5)
    pedH32       = TH1F("pedH32","pedestal vs. strip, TELL1 %s" % (tell1),nStrips,-0.5,nStrips-0.5)
    pedHBad      = TH1F("pedHBad","pedestal vs. strip, TELL1 %s" % (tell1),nStrips,-0.5,nStrips-0.5)
    pedH32.SetMarkerColor(mod32Color)
    pedHBad.SetMarkerColor(badColor)

    #
    # fill histograms from previously read in data
    #
    for chip in range(nChips):
        nChannels = len(strip[chip])
        for channel in range(nChannels):
            bin = strip[chip][channel]+1
            if badChannels[chip].count(channel) != 0:
                rawNoiseHBad.SetBinContent(bin,rawNoise[chip][channel])
                rawNoiseH32.SetBinContent(bin,noiseMin-1)
                rawNoiseH.SetBinContent(bin,noiseMin-1)

                noiseHBad.SetBinContent(bin,noise[chip][channel])
                noiseH32.SetBinContent(bin,noiseMin-1)
                noiseH.SetBinContent(bin,noiseMin-1)

                pedHBad.SetBinContent(bin,pedestal[chip][channel])
                pedH32.SetBinContent(bin,pedMin-1)
                pedH.SetBinContent(bin,pedMin-1)
            elif channel%32 == 0:
                rawNoiseH32.SetBinContent(bin,rawNoise[chip][channel])
                rawNoiseHBad.SetBinContent(bin,noiseMin-1)
                rawNoiseH.SetBinContent(bin,noiseMin-1)

                noiseH32.SetBinContent(bin,noise[chip][channel])
                noiseHBad.SetBinContent(bin,noiseMin-1)
                noiseH.SetBinContent(bin,noiseMin-1)

                pedH32.SetBinContent(bin,pedestal[chip][channel])
                pedHBad.SetBinContent(bin,pedMin-1)
                pedH.SetBinContent(bin,pedMin-1)
            else:
                rawNoiseH.SetBinContent(bin,rawNoise[chip][channel])
                rawNoiseHBad.SetBinContent(bin,noiseMin-1)
                rawNoiseH32.SetBinContent(bin,noiseMin-1)

                noiseH.SetBinContent(bin,noise[chip][channel])
                noiseHBad.SetBinContent(bin,noiseMin-1)
                noiseH32.SetBinContent(bin,noiseMin-1)

                pedH.SetBinContent(bin,pedestal[chip][channel])
                pedHBad.SetBinContent(bin,pedMin-1)
                pedH32.SetBinContent(bin,pedMin-1)

    #
    # set histogram limits and axis titles
    #
    noiseHistos = [rawNoiseH,rawNoiseH32,rawNoiseHBad,noiseH,noiseH32,noiseHBad]
    pedHistos   = [pedH,pedH32,pedHBad]
    histos      = noiseHistos + pedHistos
    for h in histos:
        h.SetXTitle('strip number')
        h.SetYTitle(yTitle)
    for h in noiseHistos:
        h.SetMaximum(noiseMax)
        h.SetMinimum(noiseMin)
    for h in pedHistos:
        h.SetMaximum(pedMax)
        h.SetMinimum(pedMin)

    #
    # create canvases and draw the histograms
    #
    cvRawNoise  = TCanvas("cvRawNoise","Raw Noise, %s" % (tell1),900,300)
    cvRawNoise.cd()
    rawNoiseH.Draw("P")
    rawNoiseH32.Draw("PSame")
    rawNoiseHBad.Draw("PSame")

    cvNoise     = TCanvas("cvNoise","Noise, %s" % (tell1),900,300)
    cvNoise.cd()
    noiseH.Draw("P")
    noiseH32.Draw("PSame")
    noiseHBad.Draw("PSame")

    cvPedestals = TCanvas("cvPedestals","Pedestals, %s" % (tell1),900,300)
    cvPedestals.cd()
    pedH.Draw("P")
    pedH32.Draw("PSame")
    pedHBad.Draw("PSame")
    
    #
    # save plots to eps and png
    #
    plotFileNameEps = outDir + "/TELL1_%s_raw_noise_all.eps" % (tell1)
    plotFileNamePng = outDir + "/TELL1_%s_raw_noise_all.png" % (tell1)
    cvRawNoise.SaveAs(plotFileNameEps)
    epsToPngCmd = "convert " + plotFileNameEps + " " + plotFileNamePng
    system(epsToPngCmd)

    plotFileNameEps = outDir + "/TELL1_%s_noise_all.eps" % (tell1)
    plotFileNamePng = outDir + "/TELL1_%s_noise_all.png" % (tell1)
    cvNoise.SaveAs(plotFileNameEps)
    epsToPngCmd = "convert " + plotFileNameEps + " " + plotFileNamePng
    system(epsToPngCmd)

    plotFileNameEps = outDir + "/TELL1_%s_pedestal_all.eps" % (tell1)
    plotFileNamePng = outDir + "/TELL1_%s_pedestal_all.png" % (tell1)
    cvPedestals.SaveAs(plotFileNameEps)
    epsToPngCmd = "convert " + plotFileNameEps + " " + plotFileNamePng
    system(epsToPngCmd)
# end of perSensorPlots()

########## MAIN ##########

########################################################################
#
# main part of the script.  call the functions that do the
# work
#

#
# configuration
#
parseArgs()
importROOTBindingsAndConf()

#
# read bad channel list, if applicable
#
if markBad:
    readBadChannels()

#
# create single chip plots
#
for chip in range(nChips):
    perChipPlots(chip)
    
#
# create plots vs. strip number for the whole
# range of sensor software strip numbers
#
perSensorPlots()

########## THE END ##########
