// $Id: VeloClusterMoni.cpp,v 1.10 2006/10/28 18:09:50 mtobin Exp $
// Include files 

// from Gaudi
#include "GaudiKernel/AlgFactory.h"

// local
#include "VeloClusterMoni.h"

// velo
#include "Kernel/IMCVeloFEType.h"
#include "VeloDet/DeVelo.h"

//-----------------------------------------------------------------------------
// Implementation file for class : VeloClusterMoni
//
// 2005-11-30 : Tomasz Szumlak
//-----------------------------------------------------------------------------

// Declaration of the Algorithm Factory
DECLARE_ALGORITHM_FACTORY( VeloClusterMoni );


//=============================================================================
// Standard constructor, initializes variables
//=============================================================================
VeloClusterMoni::VeloClusterMoni( const std::string& name,
                                  ISvcLocator* pSvcLocator)
  : GaudiTupleAlg ( name , pSvcLocator ),
    m_procInfos ( 0 ),
    m_procInfosLoc ( VeloProcessInfoLocation::Default ),
    m_clusterCont ( LHCb::InternalVeloClusterLocation::Default ),
    m_feCont ( LHCb::MCVeloFELocation::Default ),
    m_nVeloClusters ( 0. ),
    m_nVeloClusters2 ( 0. ),
    m_nVeloClustersS ( 0. ),
    m_nVeloClustersN ( 0. ),
    m_nVeloClustersO ( 0. ),
    m_nOneStrip ( 0. ),
    m_nTwoStrip ( 0. ),
    m_nThreeStrip ( 0. ),
    m_nFourStrip ( 0. ),
    m_numberOfEvents ( 0 ),
    m_veloDet ( 0 ),

    m_convergenceLimit ( ),
    m_isInitialized ( false ), // test for initialisation

    // hit tree variables, giving data about each cluster
    m_firstPosX (-1), // Position of first strip in cluster
    m_nStrips (0), // No of strips in cluster
    m_centrePosX (-1), // strip no of strip with largest cluster
    m_centreADC (-1), // ADC value of strip with biggest signal
    m_clusterADC (0), // Total signal in cluster

    // Variables for writing event trees
    // no of hits on each detector
    m_numberOfHits (0),

    // Extra event information
    m_bunchCounter (0),
    m_BCntLoop (0),  /// In this version, this variable both fixes the problems caused by the bunch counter looping back to zero, and also increases with the BATTrig value. So, by sorting the tree by this variable we can get all the "events" from the TELL1 data file in the right order, ready for analysis.
    m_doesCounterLoop (false),
    m_BATTrig (0),
    m_orderRaw (0),


    m_evtInfos ( 0 ),
    m_evtInfoLoc ( EvtInfoLocation::Default )

{
  declareProperty( "PrintInfo", m_printInfo = false );
  declareProperty("HeaderWidth", m_headerWidth=0);
  declareProperty("nSamples", m_nSamples=1);
  declareProperty("ModuleNumber", m_moduleNumber=0);
  declareProperty("DetectorNumber", m_detectorNumber=0);
}
//=============================================================================
// Destructor
//=============================================================================
VeloClusterMoni::~VeloClusterMoni() {};
//=============================================================================
// Initialization
//=============================================================================
StatusCode VeloClusterMoni::initialize() {
  StatusCode sc = GaudiAlgorithm::initialize(); // must be executed first
  if ( sc.isFailure() ) return sc;  // error printed already by GaudiAlgorithm
  //
  debug() << "==> Initialize" << endmsg;
  m_veloDet = getDet<DeVelo>( DeVeloLocation::Default );
  m_feTypeTool=tool<IMCVeloFEType>("MCVeloFEType/feTypeTool");
  setHistoTopDir("Velo/");
  //
  return StatusCode::SUCCESS;
};
//=============================================================================
// Main execution
//=============================================================================
StatusCode VeloClusterMoni::execute() {

  debug() << "==> Execute" << endmsg;
  //
  m_numberOfEvents++;

  // Get process info, including the convergence limit
  if(!m_isInitialized){
    if(!exist<VeloProcessInfos>(m_procInfosLoc)){
      debug()<< " ==> No init info available! " <<endmsg;
    }else{
      m_procInfos=get<VeloProcessInfos>(m_procInfosLoc);
    }
    m_isInitialized=true;
    VeloProcessInfos::const_iterator procIt=m_procInfos->begin();
    if(!((*procIt)->convLimit())){
      Error("Convergence Limit not set!", StatusCode::FAILURE);
    }else{
      m_convergenceLimit=(*procIt)->convLimit();
    }
  }

  //debug() << "==> Event no " << m_numberOfEvents << ",  Convergence limit " << m_convergenceLimit << endmsg;
  if(m_numberOfEvents>m_convergenceLimit){
    StatusCode sc=getData();
    if(sc) veloClusterMonitor();
  }
  //
  return StatusCode::SUCCESS;
};
//=============================================================================
//  Finalize
//=============================================================================
StatusCode VeloClusterMoni::finalize() {

  debug() << "==> Finalize" << endmsg;
  // set value if there is no event processed
  if(m_numberOfEvents==0) m_numberOfEvents=1;
  //
  m_nVeloClusters/=m_numberOfEvents;
  m_nVeloClusters2/=m_numberOfEvents;
  double errnVeloClusters=
   sqrt((m_nVeloClusters2-(m_nVeloClusters*m_nVeloClusters))/m_numberOfEvents);
  //
  m_nOneStrip/=m_numberOfEvents;
  m_nTwoStrip/=m_numberOfEvents;
  m_nThreeStrip/=m_numberOfEvents;
  m_nFourStrip/=m_numberOfEvents;
  double all=m_nOneStrip+m_nTwoStrip+m_nThreeStrip+m_nFourStrip;
  //
  info()<< "======================================================" <<endmsg;
  info()<< "              - VeloClusterMoni table -               " <<endmsg;
  info()<< "------------------------------------------------------" <<endmsg;
  info()<< "| Number of Clusters/Event: " << m_nVeloClusters << " +/- " 
        << errnVeloClusters <<endmsg;
  double allClusters=m_nVeloClustersS+m_nVeloClustersN+m_nVeloClustersO;
  if(allClusters>0){
    info().precision(4);
    info()<< "| Clusters from signal:                      " 
          << (m_nVeloClustersS/allClusters)*100
          << "%" <<endmsg;
    info().precision(4);
    info()<< "| Clusters from noise:                       " 
          << (m_nVeloClustersN/allClusters)*100
          << "%" <<endmsg;
    info().precision(4);
    info()<< "| Clusters from other (spillover/coupling):  " 
          << (m_nVeloClustersO/allClusters)*100
          << "%" <<endmsg;
    info()<< "------------------------------------------------------" <<endmsg;
    info()<< "              - Size of clusters (all)-               " <<endmsg;
    info()<< "------------------------------------------------------" <<endmsg;
    info().precision(4);
    info()<< "| 1 strip clusters/event:     "
          << m_nOneStrip << " (" << (m_nOneStrip/all)*100 << "%)"
          <<endmsg;
    info().precision(4);
    info()<< "| 2 strip clusters/event:     "
          << m_nTwoStrip << " (" << (m_nTwoStrip/all)*100 << "%)"
          <<endmsg;
    info().precision(3);
    info()<< "| 3 strip clusters/event:     "
          << m_nThreeStrip << " (" << (m_nThreeStrip/all)*100 << "%)"
          <<endmsg;
    info().precision(3);
    info()<< "| 4 strip clusters/event:     "
          << m_nFourStrip << " (" << (m_nFourStrip/all)*100 << "%)"
          <<endmsg;
  }else{
    debug()<< "| ==> No VeloClusters found! " <<endmsg;
  }
  debug()<< "======================================================" <<endmsg;
  //
  return GaudiAlgorithm::finalize();  // must be called after all other actions
}
//=============================================================================
StatusCode VeloClusterMoni::getData()
{
  debug()<< " ==> getData() " <<endmsg;
  //
  if(!exist<LHCb::InternalVeloClusters>(m_clusterCont)){
    error()<< " ==> There are no VeloClusters in TES! " <<endmsg;
    return (StatusCode::FAILURE);
  }else{
    m_veloClusters=get<LHCb::InternalVeloClusters>(m_clusterCont);
  }
  
  /* Ignore this object for now. Comes from simulation code.
  if(!exist<LHCb::MCVeloFEs>(m_feCont)){
    error()<< " ==> There are no MCVeloFEs in TES! " <<endmsg;
    return (StatusCode::FAILURE);
  }else{
    m_veloFEs=get<LHCb::MCVeloFEs>(m_feCont);
  }
  */
  //
  if(m_printInfo){
    info()<< " ==> Number of clusters found in TES: "
          << m_veloClusters->size() <<endmsg;
   /* info()<< " ==> Number of velo FEs found in TES: "
          << m_veloFEs->size() <<endmsg; */
  }
  //
  return (StatusCode::SUCCESS);
}
//==============================================================================
StatusCode VeloClusterMoni::veloClusterMonitor()
{
  debug()<< " ==> veloClusterMonitor() " <<endmsg;
  //
  int contSize=m_veloClusters->size(); // No of clusters in the event
  m_nVeloClusters+=double(contSize);
  m_nVeloClusters2+=double(contSize*contSize);
  plot(contSize, 100, "Number of VeloClusters/event", 0., 3000., 50);
  //
  LHCb::InternalVeloClusters::iterator cluIt;
  //

  //**************** Deal with event-related info for all sensors
    // Ntuples for hits
  Tuple hitNtupleCluster=nTuple("ClusterHit", "hitsTreeCluster", CLID_ColumnWiseTuple);

  // Ntuple for events
  Tuple eventNtupleCluster=nTuple("ClusterEvent", "eventTreeCluster", CLID_ColumnWiseTuple);


  // Added this in to get event info
  if(!exist<EvtInfos>(m_evtInfoLoc)){
    info()<< " ==> There is no event info at: "
           << m_evtInfoLoc <<endmsg;
  }else{  
    // get data banks from default TES location
    m_evtInfos=get<EvtInfos>(m_evtInfoLoc);
    debug()<< " ==> The EvtInfos have been read-in from location: "
           << m_evtInfoLoc
           << ", size of data container (number of read-out TELL1s): "
           << m_evtInfos->size() <<endmsg;
  }

  EvtInfo* anInfo=m_evtInfos->object(1); // get info on TELL1 no 1

  // Get bunch counter 

  m_bunchCounter=anInfo->bunchCounter();

  // Given that we record m_nSamples consecutive samples every time the BAT sees a trigger, we need to keep track of which BAT trigger each sample corresponds to. Within each m_nSamples samples in the raw data, we also add an index. NOTE that within each batch of m_nSamples samples, they are NOT automatically consecutive; at some point, we will need to re-arrange them so that we have m_nSamples consecutive BunchCounter numbers.

  m_BATTrig=(m_numberOfEvents-1)/m_nSamples; 
  m_orderRaw=m_numberOfEvents-m_BATTrig*m_nSamples-1; 

      // Then, when we get the first sample of each group, we want to change the m_counterLoops flag, so that we can create a m_BCntLoops variable where we no longer have the problem of the BunchCounter looping from 4097 to 0.

  if(m_orderRaw==0){
    if((m_bunchCounter>=(4097-m_nSamples))||(m_bunchCounter<m_nSamples)){
      m_doesCounterLoop=true;}
  else{
      m_doesCounterLoop=false;}
      }

  // With the flag set, we can then decide if we need to add 4096 to the low bunch counter nos. In this version we also add 5000 * the BAT trigger no, so that by sorting by this variable we can get all the events in order. After that sorting, we still need to fix the time bins.

  if((m_doesCounterLoop==true)&&(m_bunchCounter<=(2*m_nSamples))){
    m_BCntLoop=m_BATTrig*5000+m_bunchCounter+4096;}
  else{
    m_BCntLoop=m_BATTrig*5000+m_bunchCounter;}

  // Setup to write Event data:
  /// Write the event-related data here

  eventNtupleCluster->column("Event", m_numberOfEvents);

  eventNtupleCluster->column("BATTrig", m_BATTrig);

  eventNtupleCluster->column("OrderRaw", m_orderRaw);

  eventNtupleCluster->column("BunchCounter", m_bunchCounter);

  eventNtupleCluster->column("BCntLoop", m_BCntLoop);

  // Temp storage for no of hits
  unsigned int tempHits=0;

  //
  int iStrip=0; // Used when looping over strips
  int valStrip=0; // Temp store of strip value
 

  // Set first and last strip in detector, going by the module and detector nos

  int firstStrip=0;
  int lastStrip=127;

  // Each module has 512 strips plus 64 dummies. Each detector has 128.
  firstStrip=m_moduleNumber*576+m_detectorNumber*128;
  lastStrip=m_moduleNumber*576+m_detectorNumber*128+127;



  //***************** LOOP OVER ALL THE CLUSTERS IN THE CURRENT EVENT
  
  for(cluIt=m_veloClusters->begin(); cluIt!=m_veloClusters->end(); cluIt++){

    //****** FOR CONVENIENCE, KEEP MY WORKING SEPARATE FROM THE HISTOGRAM STUFF

  // Initialise cluster-related variables

    m_firstPosX=-1; // Position of first strip in cluster
    m_nStrips=0; // No of strips in cluster
    m_centrePosX=-1; // strip no of strip with largest cluster
    m_centreADC=0; // ADC value of strip with biggest signal
    m_clusterADC=0; // Total signal in cluster

    // For each cluster we encounter, get all the appropriate information
   
    m_nStrips=(*cluIt)->size(); // No of strips in the cluster
    m_firstPosX=(*cluIt)->strip(0); // get no of first strip 

    // Loop over the strips in the cluster and grab info
    for(iStrip=0; iStrip<m_nStrips; iStrip++){
      valStrip=((*cluIt)->adcValue(iStrip)); // grab strip signal
      m_clusterADC+=valStrip; // add to cluster total
      if ((valStrip*valStrip)>(m_centreADC*m_centreADC)) { // if this is the strip with the largest MAGNITUDE signal, update the "centre" strip no and value.
	m_centrePosX=(*cluIt)->strip(iStrip);
	m_centreADC=valStrip;
      }
    }
    //

    // Once we've extracted the info, need to check which sensor the cluster belongs to. For now, will use the same approach as with LCMS (with re-ordering)
    // Omit clusters centred on a header strip
    // Should maybe have headers customisable using options? For now, just do it myself
    // test to see if cluster falls within Gla2 - strips 0 to 127 - and also check for header
    // Test becomes a little different now - check it starts and ends within the sensor, or work based on the main strip? For now, will check that the centre of the cluster isn't in the header.


    // Use firstStrip and lastStrip as the test. Compare this to the centre position
    if(((m_centrePosX%32)>=m_headerWidth)&&(m_firstPosX>firstStrip)&&((m_firstPosX+m_nStrips-1)<lastStrip)){
      ++tempHits;
      hitNtupleCluster->column("FirstPosX", m_firstPosX);
      hitNtupleCluster->column("nStrips", m_nStrips);
      hitNtupleCluster->column("CentrePosX", m_centrePosX);
      hitNtupleCluster->column("CentreADC", m_centreADC);
      hitNtupleCluster->column("ClusterADC", m_clusterADC);
      hitNtupleCluster->column("BunchCounter", m_bunchCounter);
      hitNtupleCluster->column("BCntLoop", m_BCntLoop);
      hitNtupleCluster->column("Event", m_numberOfEvents);
      hitNtupleCluster->column("BATTrig", m_BATTrig);
      hitNtupleCluster->column("OrderRaw", m_orderRaw);
      hitNtupleCluster->write();
    }


    // Code for histograms

    // Count up the no of events with a given no of strips. Not so useful.
    double strips=(*cluIt)->size(); // No of strips in the cluster
    switch(int(strips)){
    case 1: m_nOneStrip++; break;
    case 2: m_nTwoStrip++; break;
    case 3: m_nThreeStrip++; break;
    case 4: m_nFourStrip++; break;
    default : debug()<< " ==> cluster has more than four strips!" <<endmsg;
    }
    //
    bool signal=false, noise=false, other=false;
    LHCb::InternalVeloCluster* cluster=(*cluIt);
    
    // clusterType function tries to get data from nonexistent object. Changed this so it always returns signal
    clusterType(cluster, signal, noise, other);
    // printout some info about cluster
    if(m_printInfo){
      debug()<< " ==> VeloCluster: " << " sensor number: "
             << (*cluIt)->sensor() << ", first strip in cluster: "
             << (*cluIt)->strip(0) <<endmsg;
    }
    // printout some info about strips
    int cluSize=(*cluIt)->size();
    if(m_printInfo){
      for(iStrip=0; iStrip<cluSize; iStrip++){
        debug()<< " ==> VeloCluster: " << " strip: "
               << (*cluIt)->strip(iStrip) << ", signal on strip: "
               << (*cluIt)->adcValue(iStrip) <<endmsg;
      }
    }
    //
    double adcSum=0.;
    //
    for(int iStrip=0; iStrip<cluSize; iStrip++){
      adcSum+=double((*cluIt)->adcValue(iStrip));
    }
    //
    unsigned int sensor=(*cluIt)->sensor();
    const DeVeloSensor* sens=m_veloDet->sensor(sensor);
    if(sens->isPileUp())
    debug()<< "sensor number: " << sensor <<endmsg;
    double zPosOfClu=sens->z();
    debug()<< " z pos: " << zPosOfClu <<endmsg;
    //
    plot2D((*cluIt)->sensor(), (*cluIt)->strip(0), 102,
           "Sensor and first strip number",
           0., 132., 0., 2050., 132, 50);
    plot(adcSum, 103,
         "ADC sum",
         -0.5, 255.5, 256);
    //
    if(signal){
      plot(adcSum, 104,
           "Signal dominated - ADC sum",
           -0.5, 255.5, 256);
      m_nVeloClustersS++;
      plot2D(zPosOfClu/Gaudi::Units::cm, sensor, 114,
             "Z position and sensor number for signal dominated clusters",
             -50., 100., 0., 135, 150, 135);
    }
    if(noise){
      plot(adcSum, 105,
           "Noise dominated - ADC sum",
           -0.5, 255.5, 256);
      m_nVeloClustersN++;
      plot2D(zPosOfClu/Gaudi::Units::cm, sensor, 115,
             "Z position and sensor number for noise dominated clusters",
             -50., 100., 0., 135, 150, 135);
    }
    if(other){
      plot(adcSum, 106,
           "SpillOver/couplings dominated - ADC sum",
           -0.5, 255.5, 256);
      m_nVeloClustersO++;
      plot2D(zPosOfClu, sensor, 116,
             "Z position and sensor number for spillOver/couplings"
             " dominated clusters",
             -50., 100., 0., 135, 150, 135);
    }
    //
    plot(cluSize, 107,
         "Number of strips in cluster",
         -0.5, 5.5, 6);
    if(signal){
      plot(cluSize, 108,
           "Number of strips in cluster - signal dominated",
           -0.5, 5.5, 6);
    }
    if(noise){
      plot(cluSize, 109,
           "Number of strips in cluster - noise dominated",
           -0.5, 5.5, 6);
    }
    if(other){
      plot(cluSize, 110,
           "Number of strips in cluster - spillover/couplings dominated",
           -0.5, 5.5, 6);
    }

  } //*************** END LOOP OVER ALL CLUSTERS

  // Once outside the loop, start writing Event info

  m_numberOfHits=tempHits;

  eventNtupleCluster->column("nHits", m_numberOfHits);

  eventNtupleCluster->write();

  //
  plot(m_nVeloClustersS, 111,
       "Number of velo clusters/event - signal dominated",
       0., 3000., 50);
  plot(m_nVeloClustersN, 112,
       "Number of velo clusters/event - noise dominated",
       0., 3000., 50);
  plot(m_nVeloClustersO, 113,
       "Number of velo clusters/event - spillover/couplings dominated",
       0., 3000., 50);
  //
  return (StatusCode::SUCCESS);
}











//==============================================================================
StatusCode VeloClusterMoni::clusterType(LHCb::InternalVeloCluster* clu,
                                        bool& s, bool& n, bool& o)
{

  // This entire method is a minefield, because it calls on a nonexistent data object
  // Need to alter it so I can get code running in some form.  
  // The function  is called with: clusterType(cluster, signal, noise, other);
  // It returns three boolean values, and takes an InternalVeloCluster as its input
  // I've commented out most of this, and just set it up to return "signal"

  /* LHCb::MCVeloFE* myFE=0;
  int feType=0;
  int nOfStrips=clu->size();
  for(int iStrip=0; iStrip<nOfStrips; iStrip++){
    LHCb::VeloChannelID chanID=clu->channelID(iStrip);
    LHCb::MCVeloFE* FE=m_veloFEs->object(chanID);
    if(0==FE){
      warning()<< " ==> Not found FE for given ChannelID: " 
               << chanID <<endmsg;
    }else{
      if(iStrip==0) myFE=FE;
      if(FE->charge()>myFE->charge()) myFE=FE;
    }
  }
  // s - signal, n - noise, o - coupling or spillover
  if((!s)&&(!n)&&(!o)){
    m_feTypeTool->FEType(myFE, feType);
  }else{
    s=false;
    n=false;
    o=false;
    m_feTypeTool->FEType(myFE, feType);
  }
  //
  switch(feType){
  case 0: s=true; break;
  case 1: n=true; break;
  case 2: o=true; break;
  default: error()<< " ==> Wrong FE Type flag! " <<endmsg;
    return (StatusCode::FAILURE);
  } */

  // My inserted code
    s=true;
    n=false;
    o=false;

  //
  return (StatusCode::SUCCESS);
}
