// $Id: VeloClusterDataMonitor.cpp,v 1.19 2007/07/06 13:42:06 arasp Exp $

// Include files 


// from Gaudi
#include "GaudiKernel/AlgFactory.h"
#include "GaudiKernel/SystemOfUnits.h"

// velo

#include "VeloDet/DeVelo.h"
#include "VeloDet/DeVeloSensor.h"

#include "Kernel/VeloChannelID.h"
#include "Event/VeloCluster.h"

//boost
#include <boost/lexical_cast.hpp>

#include <math.h>

// local
#include "VeloClusterDataMonitor.h"

////////////////////////////////////////////////////////////
//Uncomment this when the time is right for Online monitoring
//This is needed for the histogram publishing to DIM:
//#include "AIDA/IHistogram1D.h"
//#include "AIDA/IHistogram2D.h"
/////////////////////////////////////////////////////////////

using namespace Gaudi::Units;

//-----------------------------------------------------------------------------
// Implementation file for class : VeloClusterDataMonitor
//
// 2006-04-30:   Aras Papadelis, Thijs Versloot
//
// A cluster monitoring algorithm for the VELO detector. It produces a set of 
// histograms from the cluster bank in TES.
// We suggest that you add new histograms from a separate method (see method
// userHistos in this file). This makes it easier for you to move your own 
// code to new versions of this monitoring algorithm.
//
//-----------------------------------------------------------------------------

// Declaration of the Algorithm Factory
DECLARE_ALGORITHM_FACTORY( VeloClusterDataMonitor );


//=============================================================================
// Standard constructor, initializes variables
//=============================================================================
VeloClusterDataMonitor::VeloClusterDataMonitor( const std::string& name,
                                                ISvcLocator* pSvcLocator)
  : GaudiTupleAlg ( name , pSvcLocator ),
    m_veloDet ( 0 ),
    m_numberOfEvents ( 0 ),
    m_ntupleSize(3000)
{
  declareProperty( "PrintInfo"		, m_printInfo = false);
  declareProperty( "ClusterMoni"	, m_clusterInfo = true );
  declareProperty( "OutputNtuple"	, m_ntuple = false );
  declareProperty( "OccupancyHistos" , m_sensorOcc = false);
  declareProperty( "MultiplicityHistos"	, m_sensorMul = false);
  declareProperty( "HitmapHistos", m_hitmapPlots=true);
  declareProperty( "VeloClusterLocation",m_clusterCont=LHCb::VeloClusterLocation::Default);
  declareProperty( "HitmapResolution", m_hitmapres=2);
  declareProperty( "HitmapADCCut", m_hitmapadccut=20);
  

}
//=============================================================================
// Destructor
//=============================================================================
VeloClusterDataMonitor::~VeloClusterDataMonitor() {} 

//=============================================================================
// Initialization
//=============================================================================
StatusCode VeloClusterDataMonitor::initialize() {
  StatusCode sc = GaudiTupleAlg::initialize(); // must be executed first
  if ( sc.isFailure() ) return sc;  // error printed already by GaudiTupleAlg

  debug() << "==> Initialize" << endmsg;

  m_veloDet = getDet<DeVelo>( DeVeloLocation::Default );

  m_xrange = 70;
  m_yrange = 50;
  m_xbins = (int) (2* m_xrange / m_hitmapres);
  m_ybins = (int) (2* m_yrange / m_hitmapres);
  

  ///////////////////////////////////////////////////////////////////////
  //Only uncomment this once Vetra is compatible with the online environment.
  //  histogramBooking();
  ///////////////////////////////////////////////////////////////////////

  
  return StatusCode::SUCCESS;
}

//=============================================================================
// Main execution
//=============================================================================
StatusCode VeloClusterDataMonitor::execute() {

  debug() << "==> Execute" << endmsg;

  m_numberOfEvents++;
  
  plot1D(0, "nEvents", "Number of events", 0,1,1);

  if(m_numberOfEvents==1){
    std::vector<DeVeloSensor*>::const_iterator itSns;

    for(itSns=m_veloDet->sensorsBegin();itSns!=m_veloDet->sensorsEnd();++itSns){
      
      unsigned int sensorID=(*itSns)->sensorNumber();
      unsigned int tell1ID;
     
      if(false == sensorIsReadOut(sensorID))
        continue;
      if(m_veloDet->tell1IdBySensorNumber(sensorID, tell1ID)){
    
        plot2D(tell1ID, sensorID,"Tell1ID_vs_SensorID" 
               ,"Sensor mapping: Tell1 ID (x-axis) vs Sensor ID (y-axis), draw in \"col\" mode.",-1.5,139.5,-1.5,139.5,141,141);
        plot2D(tell1ID, sensorID,"Tell1ID_vs_SensorID_weighted" 
               ,"Sensor mapping: Tell1 ID (x-axis) vs Sensor ID (y-axis), #bin entries = sensor ID, draw in \"text\" mode."
               ,-1.5,139.5,-1.5,139.5,141,141,sensorID);
      }
      
    }
  }
  
  
  

  StatusCode sc=getClusterContainer();
  if(!sc)
    return StatusCode::SUCCESS;

  if(true==m_clusterInfo) {
    clusterMonitor();
  }
  
  if(true==m_ntuple) {
    clusterNtupleMonitor();
  }

  return StatusCode::SUCCESS;
}

//=============================================================================
//  Finalize
//=============================================================================
StatusCode VeloClusterDataMonitor::finalize() {

  debug() << "==> Finalize" << endmsg;
  
  return GaudiTupleAlg::finalize();  // must be called after all other actions
}

//=============================================================================

//=============================================================================
// Gets the cluster container
//=============================================================================
StatusCode VeloClusterDataMonitor::getClusterContainer()
{
  
  debug()<< "==> getClusterContainer() " <<endmsg;
  
  if (m_clusterInfo) {
    
    debug() << "Retrieving " << m_clusterCont << endmsg;
    if ( !exist<LHCb::VeloClusters>(m_clusterCont ) ) {
      
      debug() << " ----  No VeloClusters container retrieved --- "
        
              << endmsg;
      
      return (StatusCode::FAILURE);
      
    }else{
      m_clusters=get<LHCb::VeloClusters>(m_clusterCont );
      debug() << "---- VeloCluster Container retrieved from default location ----"
              << endmsg;
      
      if(m_printInfo){
        debug()<< "Number of clusters found in TES: "
	       << m_clusters->size() <<endmsg;
      }

      
      
    }
  }
  return (StatusCode::SUCCESS);
}



//============================================================================
//Cluster monitoring histograms
//===========================================================================
StatusCode VeloClusterDataMonitor::clusterMonitor() 
{


  //Number of clusters per event

  double size=m_clusters->size();
  plot1D(size,"nClusters","Number of VeloClusters per Event",0,3000,3000);


  if(!size) return (StatusCode::SUCCESS);
  
  
  m_nClustersPerSensor.assign(140,0);
  
  std::string histoTitle;
  std::string histoName;

//Iterate over clusters in m_cluster container
  LHCb::VeloClusters::iterator itcv;
  for (itcv=m_clusters->begin(); itcv!=m_clusters->end(); itcv++) {

    LHCb::VeloCluster* cluster = (*itcv);
    
    //Get all the cluster information and put it in variables:
      
    LHCb::VeloChannelID idtemp = cluster->channelID();
      
    int clsns = idtemp.sensor();
    int clstr = idtemp.strip();
    int clsize = cluster->size();
    
    double adc = cluster->totalCharge();
    double strfr = cluster->interStripFraction();
 
   const DeVeloSensor* temps = m_veloDet->sensor(clsns);
   
   //Create a string with the number of the sensor inside:
   std::string sensor = boost::lexical_cast<std::string>(clsns);

   //ADC-value (total charge) of clusters
      
    plot1D(adc,"ClusterADC","ADC-Value per cluster",0,128*4,128*4);
    plot1D(clsize,"ClusterSize","Number of Strips per Cluster",-0.5,4.5,5);
    
    plot2D(clsns,clsize,"ClusterSizeVsSensor","Number of Strips per Cluster vs Sensor"
           ,-0.5,105.5,-0.5,4.5,106,5);
    
    //Clusters versus position   
    unsigned int zone;
    double testRadius;
    double testRadius2;
    double sensorZ=temps->z()/cm;
   
    m_nClustersPerSensor[clsns]++;


    if(m_hitmapPlots==true){
      hitmapHistos(temps,cluster);
      debug() << "Hitmaps have been filled" << endmsg;
    }

    //R SENSOR HISTOGRAMS AND NTUPLE
    if (temps->isR()) {
const DeVeloRType* rsns = m_veloDet->rSensor(clsns);
      testRadius=rsns->rOfStrip(clstr,strfr);
      zone=rsns->zoneOfStrip(clstr);
      if (m_printInfo) {
        debug() << " sensorZ " << sensorZ
		            << " sensor " << clsns  
                << " strip " << clstr
                << " zone " << zone
                << " testRadius " << testRadius 
                << " testRadius/cm " << testRadius/cm 
                << endmsg;
      }

      std::string szone = boost::lexical_cast<std::string>(zone);
      std::string histoName="Cluster R vs Z(cm), zone " + szone;
      std::string histoTitle="ClusterRvsZ_z" + szone;
      

      plot2D(sensorZ,testRadius/cm,histoTitle,histoName,  -20.5,99.5,0,5,120,50);
      
      plot2D(sensorZ,testRadius/cm,"ClusterRvsZ","Cluster R vs Z(cm) All zones",
             -20.5,99.5,0,5,120,50);
      
      plot1D(clsns,"nClustersVsRsensor","Number of Clusters vs R sensor",
             -0.5,49.5,50);

      
      histoTitle="Cluster R, sensor " + sensor;
      histoName="ClusterR_s" + sensor;

      plot1D(testRadius/mm, histoName, histoTitle, 0,100,100);
      
      plot1D(testRadius/mm,"ClusterR","Cluster radius, all sensors"
             ,0,50,50);

      //Fill R specific ntuple vectors:

      m_clusterCentrePosR.push_back(clstr+strfr);
      m_clusterTotalChargeR.push_back((int)adc);
      m_clusterSizeR.push_back(clsize);
      m_clusterGeomPosR.push_back(testRadius);
      m_clusterZR.push_back(sensorZ);
      
      m_clusterSensorIDR.push_back(clsns);

    } //end of if R sensor
    
  
    //PHI SENSOR HISTOGRAMS
    else if(temps->isPhi()) {
      
      const DeVeloPhiType* phisns = m_veloDet->phiSensor(clsns);
      zone=phisns->zoneOfStrip(clstr);
      testRadius=phisns->rMin(zone);
      testRadius2=phisns->rMax(zone);
      double testRad = (testRadius + testRadius2 ) / 2;
      double testPhi=phisns->phiOfStrip(clstr,strfr,testRad);
      double phiCoor=phisns->localPhiToGlobal(testPhi);
      if (m_printInfo) {      
		    debug() << " sensorZ " << sensorZ
                << " sensor " << clsns
                << " strip " << clstr
                << " zone " << zone
                << " RadiusMin " << testRadius
                << " RadiusMax " << testRadius2 
                << " phiCoor (global) " << phiCoor/degree 
                << endmsg;
      }
      
      if (0==zone)
        plot2D(sensorZ,phiCoor/degree,"ClusterPhiVsZ_Inner",
               "Cluster Phi vs Z (cm) Inner", -20,100,-180,180,120,50);
	    else 
        plot2D(sensorZ,phiCoor/degree,"ClusterPhiVsZ_Outer",
               "Cluster Phi vs Z (cm) Outer", -20,100,-180,180,120,50);
      
      plot2D(sensorZ,phiCoor/degree,"ClusterPhiVsZ",
             "Cluster Phi vs Z (cm) All strips", -20,100,-180,180,120,50);

      
	    plot1D(clsns,"nClustersVsPhiSensor","Number of Clusters vs phi sensor",
             60.5,110.5,50);
      
      histoTitle="Cluster phi, sensor " + sensor;
      
      histoName="ClusterPhi_s" + sensor;
      

      plot1D(phiCoor/degree, histoName, histoTitle, -100,100,200);
      
      plot1D(phiCoor/degree,"ClusterPhi","Cluster phi, all sensors"
             ,-100,100,200);

      //Phi specific ntuple vectors:
      
      m_clusterCentrePosPhi.push_back(clstr+strfr);
      m_clusterTotalChargePhi.push_back((int) adc);
      m_clusterSizePhi.push_back(clsize);
      m_clusterGeomPosPhi.push_back(phiCoor);
      m_clusterZPhi.push_back(sensorZ);
      m_clusterSensorIDPhi.push_back(clsns);
      
            
      
    } //end of if phi sensor
    
    for(unsigned int istrip=0;istrip<cluster->size();istrip++){
      histoTitle="Cluster strips ADC value vs strip number, sensor " + sensor;
      histoName="ClusterADCvsStrip_s"+sensor;
      

      plot2D((cluster->stripValues())[istrip].first,(cluster->stripValues())[istrip].second,
             histoName,histoTitle,-0.5,2047.5,0,4*127,2048,4*127); 

      histoTitle="Cluster strips ADC value vs electronics channel, sensor " + sensor;
      histoName="ClusterADCvsElectrChannel_s"+sensor;
      
      plot2D(
             temps->StripToChipChannel( (cluster->stripValues())[istrip].first)
             ,(cluster->stripValues())[istrip].second,
             histoName,histoTitle,-0.5,2047.5,0,4*127,2048,4*127); 
    } 
    
    
  } // End of cluster iteration for-loop

  std::vector<DeVeloSensor*>::const_iterator itSns;
  for(itSns=m_veloDet->sensorsBegin();itSns!=m_veloDet->sensorsEnd();++itSns){
    
    unsigned int isns=(*itSns)->sensorNumber();
    std::string sensN = boost::lexical_cast<std::string>(isns);

    if(false==sensorIsReadOut(isns))
      continue;
    
    histoName="nClusters_s" + sensN;
    histoTitle="Number of clusters, sensor " + sensN;

    plot1D(m_nClustersPerSensor[isns],histoName,histoTitle,-0.5,255.5,256);

    if((*itSns)->isR()){  
    
     histoName="nClusters_mod" + sensN;
     histoTitle="Number of cluster, module " + sensN;
     
     plot1D(m_nClustersPerSensor[isns]+m_nClustersPerSensor[isns+64],histoName,histoTitle,-0.5,511.5,512);

     if(( m_nClustersPerSensor[isns]==0 && m_nClustersPerSensor[isns+64]==0 )) continue;
     
     histoName="nR_vs_nPhi_mod" + sensN;
     histoTitle="#R clusters vs # Phi clusters, module " + sensN;
     plot2D(m_nClustersPerSensor[isns], m_nClustersPerSensor[isns+64], 
            histoName,histoTitle,-0.5,99.5,-0.5,99.5,100,100);

     if(m_nClustersPerSensor[isns+64]==0) m_nClustersPerSensor[isns+64]=-100;
     plot2D(isns, (double) (m_nClustersPerSensor[isns]/m_nClustersPerSensor[isns+64])
	     ,"RPhiRatioVsModule","R/Phi cluster ratio vs module",
	     -0.5,49.5,-5.45,5.55,50,110);    
    }
  }
  return StatusCode::SUCCESS;
}

const bool VeloClusterDataMonitor::sensorIsReadOut(int sensorN) const
{
  verbose() << "Requesting \'isReadOut()\' for sensor # " << sensorN <<endmsg;

  const DeVeloSensor *sensor = m_veloDet->sensor(sensorN);
  if(!sensor) return 0;
  verbose() << "Module " << sensor->module()
      << " sensor " << sensor->sensorNumber()
      << " type " << sensor->fullType() 
      << " z = " << sensor->z()
      << endmsg;

  return sensor->isReadOut();
}

  


//=============================================================================
// Cluster monitoring ntuple fill
//=============================================================================
StatusCode VeloClusterDataMonitor::clusterNtupleMonitor()
{

  // IMPORTANT!  If you add more ntuple vectors, 
  // don't forget to clear them at the end of this function!

  const int s=m_ntupleSize;
  
  int nClustersR = m_clusterCentrePosR.size();
  int nClustersPhi = m_clusterCentrePosPhi.size();

  
  if( nClustersR >=m_ntupleSize || nClustersPhi >=m_ntupleSize){
    warning() << "Event is too busy. Ntuple will not be filled for this event. " << endreq;
    
    return StatusCode::FAILURE;
  }
  
  
  Tuple tuple=nTuple(100, "VeloNtuple", CLID_ColumnWiseTuple );

  tuple->farray("ClusterCentrePositionR",m_clusterCentrePosR,"nClustersR",s);
  tuple->farray("ClusterCentreSizeR",m_clusterSizeR,"nClustersR",s);
  tuple->farray("ClusterCentreGeomPositionR",m_clusterGeomPosR,"nClustersR",s);
  tuple->farray("ClusterTotalChargeR",m_clusterTotalChargeR,"nClustersR",s);
  tuple->farray("ClusterZR",m_clusterZR,"nClustersR",s);
  tuple->farray("ClusterSensorIDR", m_clusterSensorIDR, "nClustersR",s);
  

  tuple->farray("ClusterCentrePositionPhi",m_clusterCentrePosPhi, "nClustersPhi",s);
  tuple->farray("ClusterTotalChargePhi",m_clusterTotalChargePhi,"nClustersPhi",s);
  tuple->farray("ClusterSizePhi",m_clusterSizePhi,"nClustersPhi",s);
  tuple->farray("ClusterGeomPositionPhi", m_clusterGeomPosPhi,"nClustersPhi",s);
  tuple->farray("ClusterZPhi",m_clusterZPhi,"nClustersPhi",s);
  tuple->farray("ClusterSensorIDPhi", m_clusterSensorIDPhi, "nClustersPhi",s);

  
  if(m_printInfo)
    debug() << "Writing to ntuple" << endreq;
  
  tuple->write();
  

  //Clear Ntuple vectors: very important!
  
  m_clusterCentrePosR.clear();
  m_clusterTotalChargeR.clear();
  m_clusterSizeR.clear();
  m_clusterGeomPosR.clear();
  m_clusterZR.clear();
  m_clusterSensorIDR.clear();

  

  m_clusterCentrePosPhi.clear();
  m_clusterTotalChargePhi.clear();
  m_clusterSizePhi.clear();
  m_clusterGeomPosPhi.clear();
  m_clusterZPhi.clear();
  m_clusterSensorIDPhi.clear();
    
  return StatusCode::SUCCESS;    
  
}


//===========================================================================
//WARNING!
//This method fills histograms with hitmaps, based only on cluster information.
// The danger with these histograms is that there is no unambiguous way to link a an r-cluster
//to a phi-cluster without tracks. This method will plot ALL combinations of R and Phi clusters
//that pass a ADC cut, set with the option HitmapADCCut = int. Default is 20.
//It is only advised to use this when track information is not available!
//===========================================================================

StatusCode VeloClusterDataMonitor::hitmapHistos(const DeVeloSensor* temps, LHCb::VeloCluster* cluster)
{
  debug() << " ===> hitmapHistos() " << endmsg;
  if (temps->isR()==false){
    debug() << "This method requires an R-cluster to work." << endmsg;
    return StatusCode::FAILURE;
  }
  
  std::string histoTitle;
  std::string histoName;
    
  LHCb::VeloChannelID idtemp = cluster->channelID();    
  int clsns = idtemp.sensor();
  double strfr = cluster->interStripFraction();
  int clstr = idtemp.strip();
  
  if(cluster->totalCharge() < m_hitmapadccut)
    return StatusCode::SUCCESS;
  
  std::string sensN = boost::lexical_cast<std::string>(clsns); //module number
  
  LHCb::VeloClusters::iterator itcv2;
  bool phiIsUsed=false;
  
  for (itcv2=m_clusters->begin(); itcv2!=m_clusters->end(); itcv2++) {
    
    LHCb::VeloCluster* cluster2 = (*itcv2);
    LHCb::VeloChannelID idtemp2 = cluster2->channelID();

    if(cluster2->totalCharge() < m_hitmapadccut )
      continue;

    int clsns2 = idtemp2.sensor();
    const DeVeloSensor* temps2 = m_veloDet->sensor(clsns2);

    verbose() << "clsns : " << clsns << " clsns2 : " << clsns2+64  << "  isPhi = " << temps2->isPhi() << endreq;

    if (clsns2==clsns+64 && temps2->isPhi()) {
      phiIsUsed=true;

      int clstr2 = idtemp2.strip();
      double strfr2 = cluster2->interStripFraction();

      const DeVeloPhiType* phisns = m_veloDet->phiSensor(clsns2);
      const DeVeloRType* rsns = m_veloDet->rSensor(clsns);

      double iSFphi=strfr2;
      double rCoor = rsns->rOfStrip(clstr,strfr);

      double localPhiCoor = phisns->phiOfStrip(clstr2,iSFphi,rCoor); //in radians for pos
      double phiCoor=phisns->localPhiToGlobal(localPhiCoor);
      
      //transfer to xy plane
      double x_pos = rCoor * cos(phiCoor);
      double y_pos = rCoor * sin(phiCoor);

      double x_pos_local= rCoor * cos(localPhiCoor);
      double y_pos_local = rCoor * sin(localPhiCoor);      

      verbose() << "x-pos : " << x_pos << " y-pos : " << y_pos << endreq;
      
      std::vector<std::string> frame(2);
      frame[0]="_global";
      frame[1]="_local";
      
      for(int i=0;i<2;i++){
        
        histoName="Cluster_hitmap_mod" + sensN + frame[i];
        plot2D(x_pos, y_pos, histoName, histoName, -1*m_xrange, m_xrange, -1*m_yrange, m_yrange, m_xbins, m_ybins);
        
        histoName="Cluster_hitmap_mod"+ sensN + "_yproj" + frame[i];
        plot1D(y_pos,histoName,histoTitle,-50,50,100);
	
        histoName="Cluster_hitmap_mod"+sensN + "_xproj" + frame[i];
        plot1D(x_pos,histoName,histoTitle,-50,50,100);
	
        x_pos=x_pos_local;
        y_pos=y_pos_local;
	
      }
    }
  }
  return StatusCode::SUCCESS;
  
}


//////////////////////////////////////////////////////////////////////
// This method is only needed for online monitoring. It will be uncommented once there
// is a stable Online release working with Vetra.
/////////////////////////////////////////////////////////////////////
// StatusCode VeloClusterDataMonitor::histogramBooking(){

// //Counters

//   declareInfo("EventCounter" , m_numberOfEvents, "Number of Events");
 

//   //1D Histograms without variables

//   AIDA::IHistogram1D *h1_1 = book1D("nEvents", "Number of events" , 0, 1, 1);
//   declareInfo("nEvents", h1_1, "Number of events");
 
//   AIDA::IHistogram1D *h1_2 = book1D("nClusters", "Number of VeloClusters per Event" , 0, 3000, 3000);
//   declareInfo("nClusters", h1_2, "Number of VeloClusters per Events");

//   AIDA::IHistogram1D *h1_3 = book1D("ClusterADC", "ADC-Value per cluster", 0, 128*4, 128*4);
//   declareInfo("ClusterADC", h1_3, "ADC-Value per cluster");

//   AIDA::IHistogram1D *h1_4 = book1D("ClusterSize", "Number of Strips per Cluster", -0.5, 4.5, 5);
//   declareInfo("ClusterSize", h1_4, "Number of Strips per Cluster");

//   AIDA::IHistogram1D *h1_5 = book1D("nClustersVsRsensor","Number of Clusters vs R sensor", -0.5, 49.5, 50);
//   declareInfo("nClustersVsRsensor", h1_5, "Number of Clusters vs R sensor");

//   AIDA::IHistogram1D *h1_6 = book1D("ClusterR", "Cluster radius, all sensors", 0, 50, 50);
//   declareInfo("ClusterR", h1_6, "Cluster radius, all sensors");

//   AIDA::IHistogram1D *h1_7 = book1D("nClustersVsPhiSensor", "Number of Clusters vs phi sensor", 60.5, 110.5, 50);
//   declareInfo("nClustersVsPhiSensor", h1_7, "Number of Clusters vs phi sensor");

//   AIDA::IHistogram1D *h1_8 = book1D("ClusterPhi", "Cluster phi, all sensors", -100, 100, 200);
//   declareInfo("ClusterPhi", h1_8, "Cluster phi, all sensors");

//   //2D Histograms without variables
  
//   AIDA::IHistogram2D *h2_3 = book2D("ClusterSizeVsSensor",
//                                     "Number of Strips per Cluster vs Sensor", -0.5, 105.5, 106, -0.5, 4.5, 5);
//   declareInfo("ClusterSizeVsSensor", h2_3, "Number of Strips per Cluster vs Sensor");
 
//   AIDA::IHistogram2D *h2_4 = book2D("ClusterRvsZ",
//                                     "Cluster R vs Z(cm) All zones", -20.5, 99.5, 120, 0, 5, 50);
//   declareInfo("ClusterRvsZ", h2_4, "Cluster R vs Z(cm) All zones");

//   AIDA::IHistogram2D *h2_5 = book2D("ClusterPhiVsZ_Inner",
//                                     "Cluster Phi vs Z (cm) Inner", -20, 100, 120, -180, 180, 50);
//   declareInfo("ClusterPhiVsZ_Inner", h2_5, "Cluster Phi vs Z (cm) Inner");

//   AIDA::IHistogram2D *h2_6 = book2D("ClusterPhiVsZ_Outer", 
//                                     "Cluster Phi vs Z (cm) Outer", -20, 100, 120, -180, 180, 50);
//   declareInfo("ClusterPhiVsZ_Outer", h2_6, "Cluster Phi vs Z (cm) Outer");

//   AIDA::IHistogram2D *h2_7 = book2D("ClusterPhiVsZ",
//                                     "Cluster Phi vs Z (cm) All strips", -20, 100, 120, -180, 180, 50);
//   declareInfo("ClusterPhiVsZ", h2_7, "Cluster Phi vs Z (cm) All strips");

//   AIDA::IHistogram2D *h2_8 = book2D("RPhiRatioVsModule", 
//                                     "R/Phi cluster ratio vs module", -0.5, 49.5, 50, -5.45, 5.55, 110);  
//   declareInfo("RPhiRatioVsModule", h2_8, "R/Phi cluster ratio vs module");

  
//   //Loop over zones

//   int NumberOfZones = 4;
//   std::vector<AIDA::IHistogram2D*> h2_9(NumberOfZones);

//   for(int zone=0; zone<NumberOfZones; ++zone) { 
                
//     std::string szone = boost::lexical_cast<std::string>(zone);

//     std::string histoName="Cluster R vs Z(cm), zone " + szone;
//     std::string histoTitle="ClusterRvsZ_z" + szone;
//     h2_9[zone]=book2D(histoName, histoTitle, -20.5, 99.5, 120, 0, 5, 50);
//     declareInfo(histoTitle,h2_9[zone],histoName);
    
//   }

//   //Loop over sensors

//   int maxSensorID = 140;

//   std::vector<AIDA::IHistogram1D*> h1_11(maxSensorID);
//   std::vector<AIDA::IHistogram1D*> h1_12(maxSensorID);

//   std::vector<AIDA::IHistogram2D*> h2_10(maxSensorID);  
//   std::vector<AIDA::IHistogram2D*> h2_11(maxSensorID);

//   for(int sensorID = 0; sensorID<maxSensorID; ++sensorID) {
//     if(true == sensorIsReadOut(sensorID)) {
//       std::string ssensorID = boost::lexical_cast<std::string>(sensorID);

//       std::string histoTitle= "Cluster strips ADC value vs strip number, sensorID " + ssensorID;
//       std::string histoName= "ClusterADCvsStrip_s" + ssensorID;

//       h2_10[sensorID]=book2D(histoName, histoTitle, -0.5, 2047.5, 2048, 0, 512, 512);
//       declareInfo(histoTitle,h2_10[sensorID],histoName);

//       if(m_veloDet->sensor(sensorID)->isR()) {

// 	histoTitle= "#R clusters vs # Phi clusters, module " + ssensorID;
// 	histoName= "nR_vs_nPhi_mod" + ssensorID;
// 	h2_11[sensorID]=book2D(histoName, histoTitle, -0.5, 99.5, 100, -0.5, 99.5, 100);
// 	declareInfo(histoTitle,h2_11[sensorID],histoName);

// 	histoTitle="Cluster R, sensor " + ssensorID;
// 	histoName="ClusterR_s" + ssensorID;
// 	h1_11[sensorID]=book1D(histoName, histoTitle, 0, 100, 100);
// 	declareInfo(histoTitle,h1_11[sensorID],histoName);
//       }
//       else if(m_veloDet->sensor(sensorID)->isPhi()) {

// 	histoTitle="Cluster phi, sensor " + ssensorID;  
// 	histoName="ClusterPhi_s" + ssensorID;
// 	h1_12[sensorID]=book1D(histoName, histoTitle, -100, 100, 200);
// 	declareInfo(histoTitle,h1_12[sensorID],histoName);

//       }
//     }
//   }
  

//   //Book XY hitmaps for modules ( = only use R-sensors):

//   int NumberOfModules = 44;
//   std::vector<AIDA::IHistogram1D*> h_hitmapXYprojX(NumberOfModules);
//   std::vector<AIDA::IHistogram1D*> h_hitmapXYprojY(NumberOfModules);
//   std::vector<AIDA::IHistogram2D*> h_hitmap(NumberOfModules);
//   std::vector<AIDA::IHistogram2D*> h_hitmap_rot(NumberOfModules);
 
  
//   std::vector<AIDA::IHistogram1D*> h_hitmapXYprojX_loc(NumberOfModules);
//   std::vector<AIDA::IHistogram1D*> h_hitmapXYprojY_loc(NumberOfModules);
//   std::vector<AIDA::IHistogram2D*> h_hitmap_loc(NumberOfModules);
  
  
//   for(int isensor=0; isensor< NumberOfModules ; ++isensor){

//     if(sensorIsReadOut(isensor)==false)
//       continue;
//     if(m_veloDet->sensor(isensor)->isR() == false)
//       continue;
    
//     std::string sensor = boost::lexical_cast<std::string>(isensor);
    
//     std::string histoName = "Cluster_hitmap_mod" + sensor + "_global";
//     h_hitmap[isensor]=book2D(histoName,histoName,-1*m_xrange, m_xrange, m_xbins, -1 * m_yrange, m_yrange, m_ybins);
//     declareInfo(histoName, h_hitmap[isensor],histoName ); 
    
//     histoName = "Cluster_hitmap_mod"+sensor + "_yproj_global";
//     h_hitmapXYprojY[isensor]=book1D(histoName,histoName, -50,50,100);
//     declareInfo(histoName, h_hitmapXYprojY[isensor], histoName);
    
//     histoName = "Cluster_hitmap_mod"+sensor + "_xproj_global";
//     h_hitmapXYprojX[isensor]=book1D(histoName,histoName,-50,50,100);
//     declareInfo(histoName, h_hitmapXYprojX[isensor],histoName);
    
    
//     histoName = "Cluster_hitmap_mod" + sensor + "_local"; 
    
//     h_hitmap_loc[isensor]=book2D(histoName,histoName,
// 				 -1*m_xrange, m_xrange, m_xbins, -1 * m_yrange, m_yrange, m_ybins);
//     declareInfo(histoName, h_hitmap_loc[isensor],histoName ); 

        
//   }
  
//   return StatusCode::SUCCESS;
  
// }




// This method draws the contours of all the edge strips of the read out sensors and the beginning
// and end point of all the strips. Only call it once!

StatusCode VeloClusterDataMonitor::drawSensorContour()
{
  
  
  for(int iSensor=0; iSensor<144;iSensor++){
    if(sensorIsReadOut(iSensor)==false)
      continue;
    
    const DeVeloSensor* sensor = m_veloDet->sensor(iSensor);
  
    std::string histoName;
    histoName="SensorContour" +  boost::lexical_cast<std::string>(iSensor);
    

    for(int strip=0;strip<2048;strip++){

      std::pair<Gaudi::XYZPoint,Gaudi::XYZPoint> stripLimits = sensor->globalStripLimits(strip);
    
      Gaudi::XYZPoint stripBegin = stripLimits.first;
      Gaudi::XYZPoint stripEnd   = stripLimits.second;
    
      if(stripLimits.first.x()>stripLimits.second.x()){
        stripBegin = stripLimits.second;
        stripEnd= stripLimits.first;
      }
    
      ///  debug() << "Strip limits. Begin = " << stripBegin.x() << "," << stripBegin.y() << endmsg;
      //  debug() << "Strip limits. End   = " << stripEnd.x() << "," << stripEnd.y() << endmsg;

      bool drawStrip=false;
      if(sensor->isPhi()){
        if(strip==0 || strip==682 || strip ==683 || strip ==2047)
          drawStrip=true;
      }else
        if(sensor->isR()){
          if(strip==511 
             || strip==1023 
             || strip==1535
             || strip==2047
             || strip==512 
             || strip==1024 
             || strip==1536
             || strip==0)
            drawStrip=true;
	
        }

      if(drawStrip==true){
        if(sensor->isPhi()){
          double gradient = (stripEnd.y()-stripBegin.y()) / (stripEnd.x()-stripBegin.x());  
      
          double deltax=stripEnd.x()-stripBegin.x();
          for(double x=0;x<=deltax;x+=0.05){
            plot2D(stripBegin.x()+x,stripBegin.y() + gradient*x,histoName,histoName,-70,70,-50,50,1400,1000);
          }
        }
        else{
      
          double phiBegin=stripBegin.phi();
          double phiEnd=stripEnd.phi();
          double radius=stripBegin.Rho();
	  
          if(phiBegin>phiEnd){
            double temp=phiBegin;
            phiBegin=phiEnd;
            phiEnd=temp;
          }
	  
          double deltaphi=phiEnd-phiBegin;
	  
          for(double phi=0;phi<=deltaphi;phi+=.01){
	    
            double x=radius * cos (phi+phiBegin);
            double y=radius * sin (phi+phiBegin);

            plot2D(x,y,histoName,histoName,-70,70,-50,50,1400,1000);
          } 
        }
      }
      else{
        plot2D(stripBegin.x(),stripBegin.y(),histoName,histoName,-70,70,-50,50,1400,1000); 
        plot2D(stripEnd.x(),stripEnd.y(),histoName,histoName,-70,70,-50,50,1400,1000);
      }
    }
  }

  return StatusCode::SUCCESS;
  
}
