// $Id: VeloClusterCompare.cpp,v 1.3 2007/04/11 12:00:26 szumlat Exp $
// Include files 

// from Gaudi
#include "GaudiKernel/DeclareFactoryEntries.h" 
#include "GaudiKernel/AlgFactory.h"

// local
#include "VeloClusterCompare.h"

// velo
#include "VeloDet/DeVelo.h"

// Kernel
#include "Kernel/VeloEventFunctor.h"

// boost
#include "boost/lexical_cast.hpp"

//-----------------------------------------------------------------------------
// Implementation file for class : VeloClusterCompare
//
// 2007-01-25 : Tomasz Szumlak
//-----------------------------------------------------------------------------

// Declaration of the Algorithm Factory
DECLARE_ALGORITHM_FACTORY( VeloClusterCompare );

using namespace Gaudi::Units;
using namespace LHCb;

//=============================================================================
// Standard constructor, initializes variables
//=============================================================================
VeloClusterCompare::VeloClusterCompare( const std::string& name,
                                        ISvcLocator* pSvcLocator)
  : GaudiTupleAlg ( name , pSvcLocator ),
    m_tell1ClustersLoc ( LHCb::VeloClusterLocation::Default ),
    m_emuClustersLoc ( LHCb::VeloClusterLocation::Emulated ),
    m_evtNumber ( 0 ),
    m_nRawClusters ( 0., 0. ),
    m_nRawClusters2 ( 0., 0.  ),
    m_nOneStr ( 0., 0. ),
    m_nTwoStr ( 0., 0. ),
    m_nThreeStr ( 0., 0. ),
    m_nFourStr ( 0., 0. )
{
}
//=============================================================================
// Destructor
//=============================================================================
VeloClusterCompare::~VeloClusterCompare() {}
//=============================================================================
// Initialization
//=============================================================================
StatusCode VeloClusterCompare::initialize() {
  StatusCode sc = GaudiAlgorithm::initialize(); // must be executed first
  if ( sc.isFailure() ) return sc;  // error printed already by GaudiAlgorithm

  debug() << "==> Initialize" << endmsg;
  //
  setHistoTopDir("Velo/");
  m_veloDet=getDet<DeVelo>( DeVeloLocation::Default );
  //
  return StatusCode::SUCCESS;
}
//=============================================================================
// Main execution
//=============================================================================
StatusCode VeloClusterCompare::execute() {

  debug() << "==> Execute" << endmsg;
  //
  m_evtNumber++;
  StatusCode inputDataStatus=getData();
  if(inputDataStatus.isSuccess()){
    sortClusters();
    clustersMonitor(m_tell1Clusters, TELL1);
    clustersMonitor(m_emuClusters, EMU);
  }
  //
  return StatusCode::SUCCESS;
}
//=============================================================================
//  Finalize
//=============================================================================
StatusCode VeloClusterCompare::finalize() {

  debug() << "==> Finalize" << endmsg;

  return GaudiAlgorithm::finalize();  // must be called after all other actions
}
//=============================================================================
StatusCode VeloClusterCompare::getData()
{
  debug()<< " ==> getData() " <<endmsg;
  // take the TELL1 clusters
  if(!exist<LHCb::VeloClusters>(m_tell1ClustersLoc)){
    error()<< " ==> There is no VeloClusters in TES " <<endmsg;
    return ( StatusCode::FAILURE );
  }else{
    m_tell1Clusters=get<LHCb::VeloClusters>(m_tell1ClustersLoc);
  }
  // take the emulated clusters
  if(!exist<LHCb::VeloClusters>(m_emuClustersLoc)){
    error()<< " ==> There is no VeloClusters in TES " <<endmsg;
    return ( StatusCode::FAILURE );
  }else{
    m_emuClusters=get<LHCb::VeloClusters>(m_emuClustersLoc);
  }
  //
  debug()<< " ==> Number of the tell1 clusters found in TES at: "
         << m_tell1ClustersLoc << " is:"
         << m_tell1Clusters->size() <<endmsg;
  //
  debug()<< " ==> Number of emulated clusters found in TES at: "
         << m_emuClustersLoc << " is: " 
         << m_emuClusters->size() <<endmsg;
  //
  return ( StatusCode::SUCCESS );
}
//============================================================================
void VeloClusterCompare::sortClusters()
{
  debug()<< " ==> sortClusters() " <<endmsg;
  //
  std::stable_sort(m_tell1Clusters->begin(),m_tell1Clusters->end(),
		   VeloEventFunctor::Less_by_key<const LHCb::VeloCluster*>());
  std::stable_sort(m_emuClusters->begin(),m_emuClusters->end(),
		   VeloEventFunctor::Less_by_key<const LHCb::VeloCluster*>());
  //
}
//============================================================================
StatusCode VeloClusterCompare::clustersMonitor(
           LHCb::VeloClusters* inCont, const int procType)
{
  debug()<< " ==> rawVeloClusterMonitor() " <<endmsg;
  //
  int contSize=inCont->size();
  if(procType==TELL1){
    m_nRawClusters.first+=double(contSize);
    m_nRawClusters2.first+=double(contSize*contSize);
  }
  if(procType==EMU){
    m_nRawClusters.second+=double(contSize);
    m_nRawClusters2.second+=double(contSize*contSize);
  }
  plot(contSize, 100+procType*100,
       "Number of Velo Clusters/event"+
       boost::lexical_cast<std::string>(procType),
       0., 200., 100);
  //  
  LHCb::VeloClusters::const_iterator cluIt;
  //
  for(cluIt=inCont->begin(); cluIt!=inCont->end(); cluIt++){
    // local cluster
    LHCb::VeloCluster* cluster=(*cluIt);
    LHCb::VeloChannelID channel=cluster->channelID();
    unsigned int sensNumber=channel.sensor();
    if(procType==TELL1&&sensNumber==99) continue;
    if(procType==TELL1&&sensNumber==39) continue;
    if((cluster->totalCharge())<10) continue;
    // cluster size - number of strips in cluster
    unsigned int clusterSize=cluster->size();
    debug()<< " ==> Cluster size: " << clusterSize <<endmsg;
    // adc values
    debug()<< " The given cluster has: " << clusterSize << " strip(s) "
          << " the ADC value(s) is/are: " <<endmsg;
    //
    if(procType==TELL1){
      switch(int(clusterSize)){
      case 1: m_nOneStr.first++; break;
      case 2: m_nTwoStr.first++; break;
      case 3: m_nThreeStr.first++; break;
      case 4: m_nFourStr.first++; break;
      default : debug()<< " ==> Cluster has more than four strips! " <<endmsg;
      }
    }
    if(procType==EMU){
      switch(int(clusterSize)){
      case 1: m_nOneStr.second++; break;
      case 2: m_nTwoStr.second++; break;
      case 3: m_nThreeStr.second++; break;
      case 4: m_nFourStr.second++; break;
      default : debug()<< " ==> Cluster has more than four strips! " <<endmsg;
      }
    }
    //
    // method adcValue(stripNumber) returns value of the ADC count
    // from given strip, adcValue() calls stripValues() method which
    // returns ADCVector of pairs (stripNumber, adcValue)
    //
    for(unsigned int i=0; i<clusterSize; i++){
      debug()<< " Strip from cluster [" << i 
             << "] = " << cluster->adcValue(i) 
             << ", number of the strip on sensor: " 
             << cluster->strip(i) <<endmsg;
    }
    //
    // channelID info (this channel is used to create VeloLightCluster)
    //
    const DeVeloSensor* sensor=m_veloDet->sensor(sensNumber);
    plot2D(sensor->z()/cm, sensNumber, 101+procType*100,
           "Z position and sensor number of the cluster"+
           boost::lexical_cast<std::string>(procType),
           -50., 100., 0, 135, 150, 135);
    //
    debug()<< " Central channel of given cluster: " 
          << channel << ", sensor number: " << sensNumber
          << ", strip: " << channel.strip() <<endmsg;
    //
    // total charge colected on the strips
    double tCharge=cluster->totalCharge();
    if(tCharge>126.&&tCharge<128.){
      debug()<< "clusters ADC: " << tCharge
            << "this clu strip: " << cluster->channelID().strip()
            << "and sensor: " << cluster->channelID().sensor() <<endmsg;
    }
    debug()<< " Total charge of the cluster: " << tCharge <<endmsg;
    plot(tCharge, 102+procType*100,
         "ADC sum for all clusters"+
         boost::lexical_cast<std::string>(procType),
         -0.5, 260.5, 261);
    plot2D(sensNumber, cluster->firstChannel().strip(), 103+procType*100,
           "Sensor and first strip number"+
           boost::lexical_cast<std::string>(procType),
           0., 135., 0., 2050., 135, 2050);
    plot(clusterSize, 104+procType*100,
         "Number of strips in cluster"+
         boost::lexical_cast<std::string>(procType),
         -0.5, 5.5, 6);
    //
    // fractional position of the centre of the given cluster
    double fracPos=cluster->interStripFraction();
    debug()<< " Fractional position: " << fracPos <<endmsg;
    plot(fracPos, 114+procType*100,
         "Inter strip position of cluster centre"+
         boost::lexical_cast<std::string>(procType),
         0., 1., 100);
    //
    // pseudo size of the cluster
    unsigned int pseudoSize=cluster->pseudoSize();
    debug()<< " Pseudo size: " << pseudoSize <<endmsg;
    plot(pseudoSize, 115+procType*100,
         "Cluster pseudo size"+
         boost::lexical_cast<std::string>(procType),
         -0.5, 5.5, 6);
    //
    // high threshold bit
    bool thresholdBit=cluster->highThreshold();
    debug()<< " Signal bit: " << thresholdBit <<endmsg;
    plot(thresholdBit, 116+procType*100, 
         "Cluster high threshold"+
         boost::lexical_cast<std::string>(procType),
         -1., 2., 30);
    //
    // first and last channels
    LHCb::VeloChannelID firstChan=cluster->firstChannel();
    LHCb::VeloChannelID lastChan=cluster->lastChannel();
    debug()<< " First channel info: strip - " << firstChan.strip()
          << ", sensor - " << firstChan.sensor() <<endmsg;
    debug()<< " Last channel info: strip - " << lastChan.strip()
          << ", sensor - " << lastChan.sensor() <<endmsg;
    // channels vector
    std::vector<LHCb::VeloChannelID> channels=cluster->channels();
    debug()<< " Number of VeloChannelIDs inside the channel vector: "
          << channels.size() <<endmsg;
    // type of the cluster
    if(cluster->isRType()) debug()<< " RType cluster " <<endmsg;
    if(cluster->isPhiType()) debug()<< " PhiType cluster " <<endmsg;
    if(cluster->isPileUp()) debug()<< " PileUpType cluster" <<endmsg;
    // ADC signals
    LHCb::VeloCluster::ADCVector localADC;
    localADC=cluster->stripValues();
    for(unsigned int i=0; i<localADC.size(); i++){
      int stripNum=localADC[i].first;
      unsigned int adcValue=localADC[i].second;
      debug()<< " ADC value of the strip " << stripNum << " is: "
             << adcValue <<endmsg;
    }
  }
  //
  return ( StatusCode::SUCCESS );
}
//=============================================================================
