// $Id: $
// Include files 

// from Gaudi
#include "GaudiKernel/AlgFactory.h" 

// local
#include "VeloClusterizationAdaptor.h"

// VeloDet
#include "VeloDet/DeVelo.h"

// general
#include <stdlib.h>

//-----------------------------------------------------------------------------
// Implementation file for class : VeloClusterizationAdaptor
//
// 2007-11-28 : Tomasz Szumlak
//-----------------------------------------------------------------------------

// Declaration of the Algorithm Factory
DECLARE_ALGORITHM_FACTORY( VeloClusterizationAdaptor );
//=============================================================================
// Standard constructor, initializes variables
//=============================================================================
VeloClusterizationAdaptor::VeloClusterizationAdaptor( const std::string& name,
                                                      ISvcLocator* pSvcLocator)
  : GaudiAlgorithm ( name , pSvcLocator ),
    m_velo ( 0 ),
    m_inData ( 0 ),
    m_outData ( 0 )
{ 
  declareProperty("InDataLocation",
                  m_inDataLoc=LHCb::VeloTELL1DataLocation::CMSuppressedADCs);
  declareProperty("OutDataLocation",
                  m_outDataLoc=LHCb::VeloDigitLocation::Default);
  declareProperty("SensorNumber", m_sensorNumber=25);
  declareProperty("SetDummySensor", m_useDummySensor=false);
  declareProperty("ADCThreshold", m_adcThreshold=3);
}
//=============================================================================
// Destructor
//=============================================================================
VeloClusterizationAdaptor::~VeloClusterizationAdaptor() {} 

//=============================================================================
// Initialization
//=============================================================================
StatusCode VeloClusterizationAdaptor::initialize() {
  StatusCode sc = GaudiAlgorithm::initialize(); // must be executed first
  if ( sc.isFailure() ) return sc;  // error printed already by GaudiAlgorithm

  debug() << "==> Initialize" << endmsg;
  m_velo=getDet<DeVelo>(DeVeloLocation::Default);
  return StatusCode::SUCCESS;
}
//=============================================================================
// Main execution
//=============================================================================
StatusCode VeloClusterizationAdaptor::execute() {

  debug() << "==> Execute" << endmsg;
  StatusCode jobStatus=getData();
  if(jobStatus.isSuccess()){
    jobStatus=translateData();
    if(jobStatus.isSuccess()) jobStatus=storeDigiData();
  }else{
    return ( StatusCode::FAILURE );
  }
  //
  return ( jobStatus );
}
//=============================================================================
//  Finalize
//=============================================================================
StatusCode VeloClusterizationAdaptor::finalize() {

  debug() << "==> Finalize" << endmsg;

  return GaudiAlgorithm::finalize();  // must be called after all other actions
}
//=============================================================================
StatusCode VeloClusterizationAdaptor::getData()
{
  debug()<< " getData " <<endmsg;
  //
  if(!exist<LHCb::VeloTELL1Datas>(m_inDataLoc)){
    debug()<< " ==> There is no input data: "
           << m_inDataLoc <<endmsg;
    return ( StatusCode::FAILURE );
  }else{  
    // get data banks from default TES location
    m_inData=get<LHCb::VeloTELL1Datas>(m_inDataLoc);
    debug()<< " ==> The data have been read-in from location: "
           << m_inDataLoc
           << ", size of the data container: "
           << m_inData->size() <<endmsg;
    // vector for reordered data
    m_outData=new LHCb::VeloDigits();
  }
  //
  return ( StatusCode::SUCCESS );
}
//=============================================================================
StatusCode VeloClusterizationAdaptor::translateData()
{
  debug()<< " translateData() " <<endmsg;
  //
  LHCb::VeloTELL1Datas::const_iterator sensIt=m_inData->begin();
  // 
  for( ; sensIt!=m_inData->end(); ++sensIt){
    // get pointer to the velo sensor related with given TELL1
    unsigned int sensNb=0;
    if(!m_useDummySensor){
      const unsigned int tell1=static_cast<unsigned int>((*sensIt)->key());
      const DeVeloSensor* sens=m_velo->sensorByTell1Id(tell1);
      if(sens==0){
        Error(" ==> No match between TELL1 and sensor found! ");
        Error(" ==> Check your XML conditions file! ");
        return ( StatusCode::FAILURE );
      }
      debug()<< " sensor number: " << sensNb << " for TELL1 number: "
             << tell1 <<endmsg;
      sensNb=sens->sensorNumber();
    }else{
      sensNb=m_sensorNumber;
    }
    // get data vector and create digits
    std::vector<signed int> adcs=(*sensIt)->data();
    std::vector<signed int>::const_iterator adcsIt=adcs.begin();
    unsigned int chan=0;
    for( ; adcsIt!=adcs.end(); ++adcsIt){
      debug()<< "Entered loop over ADC entries at channel " << chan <<endmsg;
      // make a cut on adc value
      if(abs((*adcsIt))>m_adcThreshold){
        LHCb::VeloChannelID chanID(sensNb, chan, LHCb::VeloChannelID::RType);
        LHCb::VeloDigit* aDigi=new LHCb::VeloDigit(chanID);
        debug()<< "Just before setting ADC val of aDigi" <<endmsg;
        aDigi->setADCValue(*adcsIt);
        debug()<< "Just before inserting aDigi into m_outData" <<endmsg;
        m_outData->insert(aDigi);
        debug()<< "Inserted aDigi into m_outData" <<endmsg;
      }
      ++chan;
      if (chan==2048) break;
    }
  }
  return ( StatusCode::SUCCESS );
}
//=============================================================================
StatusCode VeloClusterizationAdaptor::storeDigiData()
{
  debug()<< "storeDigiData() " <<endmsg;
  //
  //if(m_outData->size()==0){
    //return ( StatusCode::FAILURE );
  //}else{
    put(m_outData, m_outDataLoc);
  //}
  return ( StatusCode::SUCCESS );
}
//

    
