// $Id: VeloTELL1ClusterMaker.cpp,v 1.13 2007/08/26 16:39:14 szumlat Exp $
// Include files 

// from Gaudi
#include "GaudiKernel/DeclareFactoryEntries.h" 

// boost
#include <boost/assign/std/vector.hpp>

// stl
#include <vector>
#include <algorithm>

// local
#include "VeloTELL1ClusterMaker.h"

// kernel
#include "Kernel/ContPrinter.h"
#include "Kernel/BitReader.h"

// engine class
#include "TELL1Engine/TELL1ZSProcessEngine.h"

//-----------------------------------------------------------------------------
// Implementation file for class : VeloTELL1ClusterMaker
//
// 2006-10-02 : Tomasz Szumlak
//-----------------------------------------------------------------------------

// Declaration of the Algorithm Factory
DECLARE_ALGORITHM_FACTORY( VeloTELL1ClusterMaker );

using namespace boost::assign;

//=============================================================================
// Standard constructor, initializes variables
//=============================================================================
VeloTELL1ClusterMaker::VeloTELL1ClusterMaker( const std::string& name,
                                              ISvcLocator* pSvcLocator)
  : VeloTELL1Algorithm ( name , pSvcLocator ),
    m_rawEvent ( 0 ),
    m_rawEventLoc ( RawEventLocation::Emulated ),
    m_clusters ( ),
    m_clustersMem ( ),
    m_adcs ( ),
    m_adcsMem ( ),
    m_eventNumber ( 0 ),
    m_bankVersion ( V2 ),
    m_zsEngine ( 0 )
{
  setTELL1Process(CLUSTER_MAKER);
  setAlgoName("CLUSTER_MAKER");
  setAlgoType("TELL1 Cluster Maker");
  // fill vector with number of first strips in each processing channel
  m_boundaryStripsR+=0, 64, 128, 192, 256, 320, 384, 448, 512, 512,
                    576, 640, 704, 768, 832, 896, 960, 1024, 1024,
                    1088, 1152, 1216, 1280, 1344, 1408, 1472, 1536,
                    1536, 1600, 1664, 1728, 1792, 1856, 1920, 1984, 0;
  m_boundaryStripsPhi+=0, 64, 128, 683, 747, 811, 875, 939, 1003, 171,
                       235, 299, 1024, 1088, 1152, 1216, 1280, 1344,
                       342, 406, 470, 1365, 1429, 1493, 1557, 1621, 
                       1685, 512, 576, 640, 1707, 1771, 1835, 1899, 
                       1963, 2027;
  declareProperty("InputDataLoc",
                  m_inputDataLoc=VeloTELL1DataLocation::CMSuppressedADCs);
  declareProperty("PrintInfo", m_printInfo=false);
  declareProperty("ValidationRun", m_validationRun=false);
  declareProperty("HitThresholdValue", m_hitThresholdValue=HIGH_THRESHOLD);
  declareProperty("LowThresholdValue", m_lowThresholdValue=LOW_THRESHOLD);
  declareProperty("SumThresholdValue", m_sumThresholdValue=ADC_SUM_THRESHOLD);
}
//=============================================================================
// Destructor
//=============================================================================
VeloTELL1ClusterMaker::~VeloTELL1ClusterMaker() {}
//=============================================================================
// Initialization
//=============================================================================
StatusCode VeloTELL1ClusterMaker::initialize() {
  StatusCode sc = VeloTELL1Algorithm::initialize(); // must be executed first
  if ( sc.isFailure() ) return sc;  // error printed already by GaudiAlgorithm

  info() << "==> Initialize" << endmsg;
  HitThresholdVec tempHit(2304, m_hitThresholdValue);
  m_hitThresholds=tempHit;
  Thresholds tempLow(36, m_lowThresholdValue);
  m_lowThresholds=tempLow;
  Thresholds tempSum(36, m_sumThresholdValue);
  //
  return StatusCode::SUCCESS;
}
//=============================================================================
// Main execution
//=============================================================================
StatusCode VeloTELL1ClusterMaker::execute() {

  debug() << "==> Execute" << endmsg;
  ++m_eventNumber;
  //
  if(m_validationRun){
    INIT();
    prepareEngine();
    StatusCode rawEvtStatus=createRawEvent();
    StatusCode exitStatus;
    if(rawEvtStatus.isSuccess()){
      exitStatus=validationRun();
    }
    return ( exitStatus );
  }else{
    if(!isInitialized()){
      INIT();
      prepareEngine();
    }
    StatusCode rawEvtStatus=createRawEvent();
    if(isEnable()&&(m_eventNumber>convergenceLimit())){
      StatusCode dataStatus=getData();
      StatusCode cluMakerStatus;
      if(dataStatus.isSuccess()&&rawEvtStatus.isSuccess()){
        dataStatus=inputStream(inputData());
        if(dataStatus.isSuccess()){
          cluMakerStatus=runClusterMaker();
          if(cluMakerStatus.isSuccess()){ 
            fillAndWriteRawEvent();
          }else{
            Error(" The cluster maker was not successful! ");
            return ( cluMakerStatus );
          }
        }
      }
      flushMemory();
    }
  }
  //
  return ( StatusCode::SUCCESS );
}
//=============================================================================
//  Finalize
//=============================================================================
StatusCode VeloTELL1ClusterMaker::finalize() {

  debug() << "==> Finalize" << endmsg;
  // must be called after all other actions
  delete m_zsEngine;
  return VeloTELL1Algorithm::finalize();
}
//=============================================================================
StatusCode VeloTELL1ClusterMaker::createRawEvent()
{
  debug()<< " createRawEvent() " <<endmsg;
  // create raw event - that is the output of the algorithm 
  if(exist<RawEvent>(m_rawEventLoc)){
    m_rawEvent=get<RawEvent>(m_rawEventLoc);
  }else{
    // create RawEvent structure
    m_rawEvent=new RawEvent();
    eventSvc()->registerObject(m_rawEventLoc, m_rawEvent);
    debug()<< " registered RawEvent " <<endmsg;
  }
  //
  return ( StatusCode::SUCCESS );
}
//=============================================================================
StatusCode VeloTELL1ClusterMaker::runClusterMaker()
{
  debug()<< " ==> runClusterMaker() " <<endmsg;
  //
  LHCb::VeloTELL1Datas::iterator sensIt=inputData()->begin();
  // loop over all sensors
  debug()<< " Size of input container: " << inputData()->size() <<endmsg;
  for( ; sensIt!=inputData()->end(); ++sensIt){
    m_clusters.clear();
    m_adcs.clear();
    // get info about current sensor
    const unsigned int tell1=static_cast<unsigned int>((*sensIt)->key());
    const DeVeloSensor* sens=deVelo()->sensorByTell1Id(tell1);
    // check if the sensor pointer is valid
    if(sens==0){
      Error(" ==> No match between TELL1 and sensor found! ");
      Error(" ==> Check your XML/DB conditions file! ");
      return ( StatusCode::FAILURE );
    }
    unsigned int sensNb=sens->sensorNumber();
    debug()<< " the sensor number: " << sensNb << " for TELL1 number: "
          << tell1 <<endmsg;
    // input data for clusterization
    if(sens->isReadOut()){
      // cache the current read-out sensor number
      if(m_eventNumber==(convergenceLimit()+1)){
        m_sensors.push_back(tell1); 
      }
      sdataVec data=(*sensIt)->data();
      if(sens->isR()||sens->isPileUp()){
        // processing R sensor set start strip sequence for R
        m_zsEngine->setBoundaryStrips(m_boundaryStripsR);
      }else if(sens->isPhi()){
        // processing Phi sensor
        m_zsEngine->setBoundaryStrips(m_boundaryStripsPhi);
      }
      if(m_eventNumber>convergenceLimit()&&m_printInfo)
        contPrinter(data, 10);
      m_zsEngine->setInData(data);
      m_zsEngine->runZeroSuppression();
      // store clusters, adcs and corresponding counters in memory
      m_clusters=m_zsEngine->outClusters();
      m_adcs=m_zsEngine->outADCS();
    }
    // store clusters, adcs and corresponding counters in memory
    m_clustersMem.push_back(m_clusters);
    m_adcsMem.push_back(m_adcs);   
  }
  //
  return ( StatusCode::SUCCESS );
}
//=============================================================================
void VeloTELL1ClusterMaker::flushMemory()
{
  debug()<< " ==> flushMemory() " <<endmsg;
  //
  m_clustersMem.clear();
  m_adcsMem.clear();
}
//=============================================================================
StatusCode VeloTELL1ClusterMaker::fillAndWriteRawEvent()
{
  debug()<< " ==> fillAndWriteRawEvent() " <<endmsg;
  //
  // coding of the RawBank bank body: header word, clusters and adcs
  // create one RawBank per sensor, to do so loop over container
  // with sensor numbers, use const_iterator - see Kernel/DecodeCore.h
  unsigned int rawBankCnt=0;
  cdatIt IT=m_sensors.begin();
  debug()<< " size of sens vec: " << m_sensors.size() <<endmsg;
  debug()<< " first sens: " << (*IT) <<endmsg;
  for( ; IT!=m_sensors.end(); ++IT){
    // clear bank body for next bank
    m_bankBody.clear();
    TELL1ClusterVec clusters=m_clustersMem[rawBankCnt];
    unsigned int clusterNumber=clusters.size();
    // create and sort a header word
    SiHeaderWord aHWord(clusterNumber, 0, 0);
    SiDAQ::buffer_word aHeader=aHWord.value();
    m_bankBody.push_back(aHeader);
    // check if clusters were produced for current sensor
    // check if the cluster number is even
    unsigned bufferWords=0;
    if((clusterNumber)&&(clusterNumber%CLUSTER_PER_WORD==0)){
      // calculate number of buffer_words needed to code clusters
      bufferWords=
      static_cast<unsigned int>(floor(
      static_cast<double>(clusterNumber)/
      static_cast<double>(CLUSTER_PER_WORD)));
      for(unsigned int word=0; word<bufferWords; ++word){
        SiDAQ::buffer_word codedClusters=0x0, tempBuffWord=0x0;
        for(int cluCnt=0; cluCnt<CLUSTER_PER_WORD; ++cluCnt){
          tempBuffWord=clusters[CLUSTER_PER_WORD*word+cluCnt];
          tempBuffWord<<=(cluCnt*CLUSTER_SHIFT);
          codedClusters|=tempBuffWord;
        }
        if(m_printInfo) bitReader<u_int32_t>(codedClusters);
        m_bankBody.push_back(codedClusters);
      }
    }else if((clusterNumber)&&((clusterNumber%CLUSTER_PER_WORD!=0))){
      unsigned int fullWords=
      static_cast<unsigned int>(floor(
      static_cast<double>(clusterNumber)/
      static_cast<double>(CLUSTER_PER_WORD)));
      // take into account last cluster
      bufferWords=fullWords+1;
      for(unsigned int word=0; word<bufferWords; ++word){
        SiDAQ::buffer_word codedClusters=0x0, tempBuffWord=0x0;
        // write first full words (2 clusters per word)
        if(clusterNumber==1){
          codedClusters|=clusters[CLUSTER_PER_WORD*word];
          //codedClusters<<=CLUSTER_SHIFT;
          if(m_printInfo) bitReader<u_int32_t>(codedClusters);
          m_bankBody.push_back(codedClusters);
        }else if((clusterNumber>1)&&(word<fullWords)){
          for(int cluCnt=0; cluCnt<CLUSTER_PER_WORD; ++cluCnt){
            tempBuffWord=clusters[CLUSTER_PER_WORD*word+cluCnt];
            tempBuffWord<<=(cluCnt*CLUSTER_SHIFT);
            codedClusters|=tempBuffWord;
          }
          if(m_printInfo) bitReader<u_int32_t>(codedClusters);
          m_bankBody.push_back(codedClusters);
        }else if((clusterNumber>1)&&(word==fullWords)){
          codedClusters|=clusters[CLUSTER_PER_WORD*word];
          if(m_printInfo) bitReader<u_int32_t>(codedClusters);
          m_bankBody.push_back(codedClusters);
        }
      }
    }
    // store adcs in raw bank
    TELL1ADCVec adcs=m_adcsMem[rawBankCnt];
    unsigned int adcNumber=adcs.size();
    if(clusterNumber&&(adcNumber%ADCS_PER_WORD==0)){
      // calculate number of buffer_words needed to code the adcs 
      bufferWords=
      static_cast<unsigned int>(floor(
      static_cast<double>(adcNumber)/static_cast<double>(ADCS_PER_WORD)));
      for(unsigned int word=0; word<bufferWords; ++word){
        SiDAQ::buffer_word codedADCS=0x0, tempBuffWord=0x0;
        for(int adcCnt=0; adcCnt<ADCS_PER_WORD; ++adcCnt){
          tempBuffWord=adcs[ADCS_PER_WORD*word+adcCnt];
          tempBuffWord<<=(adcCnt*ADC_SHIFT);
          codedADCS|=tempBuffWord;
          //if(adcCnt<ADCS_PER_WORD-1) codedADCS<<=ADC_SHIFT;
        }
        m_bankBody.push_back(codedADCS);
      } 
    }else if(clusterNumber&&(adcNumber%ADCS_PER_WORD!=0)){
      unsigned int fullWords=
      static_cast<unsigned int>(floor(
      static_cast<double>(adcNumber)/static_cast<double>(ADCS_PER_WORD)));
      // take into account remaining adcs
      bufferWords=fullWords+1;
      for(unsigned int word=0; word<bufferWords; ++word){
        SiDAQ::buffer_word codedADCS=0x0, tempBuffWord=0x0;
        // write first full word (4 adcs per word)
        if(word<fullWords){
          for(int adcCnt=0; adcCnt<ADCS_PER_WORD; ++adcCnt){
            debug()<< " adc: " << adcs[ADCS_PER_WORD*word+adcCnt] <<endmsg;
            tempBuffWord=adcs[ADCS_PER_WORD*word+adcCnt];
            tempBuffWord<<=(adcCnt*ADC_SHIFT);
            if(m_printInfo) bitReader<u_int32_t>(tempBuffWord);
            codedADCS|=tempBuffWord;
            //if(adcCnt<ADCS_PER_WORD-1) codedADCS<<=ADC_SHIFT;
            if(m_printInfo) bitReader<u_int32_t>(codedADCS);
          }
          m_bankBody.push_back(codedADCS);
        }else if(word==fullWords){
          // calculate number of adcs for the last buffer word
          unsigned int lastADCS=adcNumber%ADCS_PER_WORD;
          for(unsigned int adcCnt=0; adcCnt<lastADCS; ++adcCnt){
            tempBuffWord=adcs[ADCS_PER_WORD*word+adcCnt];
            tempBuffWord<<=(adcCnt*ADC_SHIFT);
            if(m_printInfo) bitReader<u_int32_t>(tempBuffWord);
            codedADCS|=tempBuffWord;
            //codedADCS<<=ADC_SHIFT;
            if(m_printInfo) bitReader<u_int32_t>(codedADCS);
          }
          m_bankBody.push_back(codedADCS);
        }
      }
    }
    // calculate bank size in bytes:
    int paddingSpace=adcNumber%ADCS_PER_WORD;
    m_bankBodySize=sizeof(SiDAQ::buffer_word)*m_bankBody.size()-
                   (paddingSpace?sizeof(SiDAQ::buffer_word)-
                    paddingSpace*sizeof(u_int8_t):0);
    // some checks
    for(unsigned int i=0; i<m_bankBody.size(); ++i){
      debug()<< " m_bankBody[" << i << "]=" << m_bankBody[i] <<endmsg;
    }
    // store raw bank
    RawBank* aBank=
    m_rawEvent->createBank(static_cast<SiDAQ::buffer_word>(*IT),
    RawBank::Velo, m_bankVersion, m_bankBodySize, &(m_bankBody[0]));
    // 
    m_rawEvent->adoptBank(aBank,true);
    rawBankCnt++;
  }
  //
  return ( StatusCode::SUCCESS );
}
//=============================================================================
StatusCode VeloTELL1ClusterMaker::validationRun()
{
  info()<< " validationRun() " <<endmsg;
  //
  sdataVec localInput;
  readDataFromFile("d_lcms_sub_test.txt", localInput);
  // validate actual clu maker code
  m_sensors.push_back(25);
  debug()<< " m_sensors: " << m_sensors.size() <<endmsg;
  m_zsEngine->setInData(localInput);
  m_zsEngine->runZeroSuppression();
  info()<< " from clu maker: " 
        << static_cast<int>(*(m_hitThresholds.begin())) <<endmsg;
  TELL1ClusterVec locClusters=m_zsEngine->outClusters();
  TELL1ADCVec locADC=m_zsEngine->outADCS();
  // store clusters, adcs and corresponding counters in memory
  m_clustersMem.push_back(m_zsEngine->outClusters());
  m_adcsMem.push_back(m_zsEngine->outADCS());
  // create output
  fillAndWriteRawEvent();
  flushMemory();
  //
  return ( StatusCode::SUCCESS );
}
//=============================================================================
void VeloTELL1ClusterMaker::prepareEngine()
{
  m_zsEngine=new TELL1ZSProcessEngine();
  m_zsEngine->setProcessEnable(isEnable());
  m_zsEngine->setHitThresholds(m_hitThresholds);
  m_zsEngine->setLowThresholds(m_lowThresholds);
  m_zsEngine->setSumThresholds(m_sumThresholds);
}
//--
