// $Id: VeloTELL1FIRFilter.cpp,v 1.2 2007/08/26 16:39:14 szumlat Exp $
// Include files 

// from Gaudi
#include "GaudiKernel/DeclareFactoryEntries.h" 

// local
#include "VeloTELL1FIRFilter.h"

// Kernel
#include "Kernel/ContPrinter.h"

// stl
#include <vector>
#include <algorithm>

// engine classes
#include "TELL1Engine/TELL1FIRProcessEngine.h"

//-----------------------------------------------------------------------------
// Implementation file for class : VeloTELL1FIRFilter
//
// 2007-01-30 : Tomasz Szumlak
//-----------------------------------------------------------------------------

// Declaration of the Algorithm Factory
DECLARE_ALGORITHM_FACTORY( VeloTELL1FIRFilter );

using namespace LHCb;

//=============================================================================
// Standard constructor, initializes variables
//=============================================================================
VeloTELL1FIRFilter::VeloTELL1FIRFilter( const std::string& name,
                                        ISvcLocator* pSvcLocator)
  :  VeloTELL1Algorithm( name , pSvcLocator ),
     m_firEngine ( 0 )
{
  setTELL1Process(FIR_FILTER);
  setAlgoName("FIR_FILTER");
  setAlgoType("TELL1 FIR Filter");
  declareProperty("ValidationRun", m_validationRun=false);
  declareProperty("InputDataLoc",
                  m_inputDataLoc=VeloTELL1DataLocation::ADCs);
  declareProperty("OutputDataLoc",
                  m_outputDataLoc=VeloTELL1DataLocation::FIRCorrectedADCs);
  // Switches on FIR filtering
  declareProperty("FIRFiltering", m_FIRFiltering=false);
  declareProperty("K0", m_K0=1.0);
  declareProperty("K1", m_K1=0.0);
  declareProperty("K2", m_K2=0.0);
}
//=============================================================================
// Destructor
//=============================================================================
VeloTELL1FIRFilter::~VeloTELL1FIRFilter() {} 

//=============================================================================
// Initialization
//=============================================================================
StatusCode VeloTELL1FIRFilter::initialize() {
  StatusCode sc=VeloTELL1Algorithm::initialize(); // must be executed first
  if ( sc.isFailure() ) return sc;  // error printed already by GaudiAlgorithm

  debug() << "==> Initialize" << endmsg;
  //
  return StatusCode::SUCCESS;
}
//=============================================================================
// Main execution
//=============================================================================
StatusCode VeloTELL1FIRFilter::execute() {

  debug() << "==> Execute" << endmsg;
  //
  m_evtNumber++;
  if(!isInitialized()){ 
    INIT();
    prepareEngine();
  }
  StatusCode dataStatus=getData();
  if(m_FIRFiltering){
    if(dataStatus.isSuccess()){            /// check if data is in TES
      dataStatus=inputStream(inputData());
      if(dataStatus.isSuccess()){          /// check if not empty
        runFIRFilter();
        debug()<< " About to write to FIR output location"   <<endmsg;
        writeData();  // writes m_outputData to the output location
      }else{
        return ( dataStatus );
      }
    }else{
      return ( dataStatus );
    }
  }else{
    if(dataStatus.isSuccess()){
      cloneData();
    }else{
      return ( dataStatus );
    }
  }
  return StatusCode::SUCCESS;
}
//=============================================================================
//  Finalize
//=============================================================================
StatusCode VeloTELL1FIRFilter::finalize() {

  debug() << "==> Finalize" << endmsg;
  delete m_firEngine;
  // must be called after all other actions
  return VeloTELL1Algorithm::finalize();
}
//=============================================================================
StatusCode VeloTELL1FIRFilter::runFIRFilter()
{

  float K0test=0.987;
  float K1test=0.987;
  float K2test=0.987;

  debug()<< " ==> runFIRFilter() " <<endmsg;
  // iterate over all sensors
  VeloTELL1Datas::const_iterator sensIt;
  sensIt=inputData()->begin();
  // 
debug()<< "About to start main loop in runFIRFilter"   <<endmsg;
  for( ; sensIt!=inputData()->end(); ++sensIt){
     debug()<< "Start of main loop in runFIRFilter"   <<endmsg;
    // get pointer to the velo sensor related with given TELL1
    const unsigned int tell1=static_cast<unsigned int>((*sensIt)->key());
    const DeVeloSensor* sens=deVelo()->sensorByTell1Id(tell1);
    if(sens==0){
      Error(" ==> No match between TELL1 and sensor found! ");
      Error(" ==> Check your XML conditions file! ");
      return ( StatusCode::FAILURE );
    }
    unsigned int sensNb=sens->sensorNumber();
    debug()<< " sensor number: " << sensNb << " for TELL1 number: "
          << tell1 <<endmsg;
    // check if the sensor is read out
    if(sens->isReadOut()){
      debug()<< "Just after sens->isReadOut"   <<endmsg;
      // create a new VeloTELL1Data object for FIR corrected data
      VeloTELL1Data* firData=new VeloTELL1Data(tell1, VeloFull);
      // data from current sensor
      sdataVec inData=(*sensIt)->data();
      m_firEngine->setInData(inData); // Set input data for FIR engine


      m_firEngine->setK0(m_K0); // These bits are added so that the filter coefficients are set by the options file
      m_firEngine->setK1(m_K1);
      m_firEngine->setK2(m_K2);

      // Check the values we're dealing with
      
      debug()<< " Test of temp storage for K0test " << K0test <<endmsg;

      debug()<< " m_K0: " << m_K0  <<endmsg;
      debug()<< " m_K1: " << m_K1  <<endmsg;
      debug()<< " m_K2: " << m_K2  <<endmsg;

      K0test=m_firEngine->getK0();
      K1test=m_firEngine->getK1();
      K2test=m_firEngine->getK2();

      debug()<< " m_firEngine->getK0(): " << K0test <<endmsg;
      debug()<< " m_firEngine->getK1(): " << K1test <<endmsg;
      debug()<< " m_firEngine->getK2(): " << K2test <<endmsg;


      m_firEngine->runFilter(); // run FIR engine
      firData->setDecodedData(m_firEngine->outData()); // Gets the output data from the FIR engine, sticks it into the firData object. 
      // store the fir data
      debug()<< " About to put FIR data into m_outputData"   <<endmsg;
      outputData(firData); // Puts firData into m_outputData
      debug()<< " Just put FIR data into m_outputData"   <<endmsg;
    }
    debug()<< "Now, we're just past the bracket"   <<endmsg;
  }
  debug()<< " Just exiting runFIRFilter"   <<endmsg;
  return ( StatusCode::SUCCESS );
}
//=============================================================================
void VeloTELL1FIRFilter::prepareEngine()
{
  debug()<< " prepareEngine() " <<endmsg;
  //
  m_firEngine=new TELL1FIRProcessEngine();
  m_firEngine->setProcessEnable(1);
 }
//--
