// $Id: VeloTELL1LCMS.cpp,v 1.4 2007/08/26 16:39:14 szumlat Exp $
// Include files

// from Gaudi
#include "GaudiKernel/DeclareFactoryEntries.h" 

// stl
#include <vector>
#include <algorithm>

// local
#include "VeloTELL1LCMS.h"
#include "Kernel/ContPrinter.h"

// engine classes
#include "TELL1Engine/TELL1LCMSProcessEngine.h"

//-----------------------------------------------------------------------------
// Implementation file for class : VeloTELL1LCMS
//
// 2007-02-01 : Tomasz Szumlak
//-----------------------------------------------------------------------------

// Declaration of the Algorithm Factory
DECLARE_ALGORITHM_FACTORY( VeloTELL1LCMS );

using namespace LHCb;
using namespace VeloTELL1;

//=============================================================================
// Standard constructor, initializes variables
//=============================================================================
VeloTELL1LCMS::VeloTELL1LCMS( const std::string& name,
                              ISvcLocator* pSvcLocator)
  : VeloTELL1Algorithm( name , pSvcLocator ),
    m_evtNumber ( 0 ),
    m_testMode ( false ),
    m_tmpCnt ( 0 ),
    m_lcmsEngine ( 0 )
{
  setTELL1Process(COMMON_MODE_SUBTRACTION);
  setAlgoType("TELL1 Linear Common Mode Subtraction");
  declareProperty("InputDataLoc",
                  m_inputDataLoc=VeloTELL1DataLocation::ReorderedADCs);
  declareProperty("OutputDataLocation",
                  m_outputDataLoc=VeloTELL1DataLocation::CMSuppressedADCs);
  declareProperty("AlgorithmName",
                  m_algorithmName="LCMS - after reordering");
  declareProperty("ValidationRun", m_validationRun=false);
}
//=============================================================================
// Destructor
//=============================================================================
VeloTELL1LCMS::~VeloTELL1LCMS() {}
//=============================================================================
// Initialization
//=============================================================================
StatusCode VeloTELL1LCMS::initialize() {
  StatusCode sc = VeloTELL1Algorithm::initialize(); // must be executed first
  if ( sc.isFailure() ) return sc;  // error printed already by GaudiAlgorithm

  debug() << "==> Initialize" << endmsg;
  //
  return StatusCode::SUCCESS;
}
//=============================================================================
// Main execution
//=============================================================================
StatusCode VeloTELL1LCMS::execute() {

  debug() << "==> Execute" << endmsg;
  //
  if(m_validationRun){
    INIT();
    prepareEngine();
    StatusCode exitStatus=validationRun();
    return ( exitStatus );
  }else{
    m_evtNumber++;
    if(!isInitialized()){
      INIT();
      prepareEngine();
    }
    StatusCode dataStatus;
    if(isEnable()&&(m_evtNumber>convergenceLimit())){
      dataStatus=getData();
      if(dataStatus.isSuccess()){
        dataStatus=inputStream(inputData());
        if(dataStatus.isSuccess()){
          runLCMSuppression();
          writeData();
        }else{
          return ( dataStatus );
        }
      }else{
        return ( dataStatus );
      }
    }else{
      dataStatus=getData();
      if(dataStatus.isSuccess()){
        cloneData();
      }else{
        return ( dataStatus );
      }
    }
  }
  //
  return ( StatusCode::SUCCESS );
}
//=============================================================================
//  Finalize
//=============================================================================
StatusCode VeloTELL1LCMS::finalize() {

  debug() << "==> Finalize" << endmsg;
  delete m_lcmsEngine;
  return VeloTELL1Algorithm::finalize();  // must be called after all other actions
}
//=============================================================================
StatusCode VeloTELL1LCMS::runLCMSuppression()
{
  debug()<< " runLCMSuppression() " <<endmsg;
  //
  VeloTELL1Datas::const_iterator sensIt;
  sensIt=inputData()->begin();
  // 
  for( ; sensIt!=inputData()->end(); ++sensIt){
    // get pointer to the velo sensor related with given TELL1
    const unsigned int tell1=static_cast<unsigned int>((*sensIt)->key());
    const DeVeloSensor* sens=deVelo()->sensorByTell1Id(tell1);
    if(sens==0){
      Error(" ==> No match between TELL1 and sensor found! ");
      Error(" ==> Check your XML conditions file! ");
      return ( StatusCode::FAILURE );
    }
    unsigned int sensNb=sens->sensorNumber();
    debug()<< " sensor number: " << sensNb << " for TELL1 number: "
          << tell1 <<endmsg;
    // check if the sensor is read out
    if(sens->isReadOut()){
      // create a new VeloTELL1Data object for cm suppressesd data
      VeloTELL1Data* lcmsData=new VeloTELL1Data(tell1, VeloFull);
      // data from current sensor
      sdataVec inData=(*sensIt)->data();
      m_lcmsEngine->setInData(inData);
      m_lcmsEngine->runLCMS();
      lcmsData->setDecodedData(m_lcmsEngine->outData());
      // store the corected data
      outputData(lcmsData);
    }
  }
  return ( StatusCode::SUCCESS );
}
//=============================================================================
StatusCode VeloTELL1LCMS::validationRun()
{
  info()<< " validationRun() " <<endmsg;
  // get test data from files
  if(isEnable()){
    info()<< " enabled " <<endmsg;
  }
  sdataVec inData, outData;
  StatusCode dataStatus=readDataFromFile("d_source.txt", inData);
  if(!dataStatus.isSuccess()){
    return ( dataStatus );
  }
  m_lcmsEngine->setInData(inData);
  m_lcmsEngine->runLCMS();
  outData=m_lcmsEngine->outData();
  writeDataToFile("d_lcms_sub.txt", outData);
  //
  return ( StatusCode::SUCCESS );
}
//=============================================================================
void VeloTELL1LCMS::prepareEngine()
{
  m_lcmsEngine=new TELL1LCMSProcessEngine();
  m_lcmsEngine->setProcessEnable(isEnable());
}
//--
