// $Id: VeloTELL1PedestalSubtractor.cpp,v 1.4 2007/08/26 16:39:14 szumlat Exp $
// Include files 

// from Gaudi
#include "GaudiKernel/DeclareFactoryEntries.h" 

// stl
#include <vector>
#include <algorithm>

// local
#include "Kernel/ContPrinter.h"
#include "VeloTELL1PedestalSubtractor.h"

// engine classes
#include "TELL1Engine/TELL1PedestalProcessEngine.h"
#include "TELL1Engine/TELL1UpdateProcessEngine.h"

#include "Kernel/TableInOut.h"

//-----------------------------------------------------------------------------
// Implementation file for class : VeloTELL1PedestalSubtractor
//
// 2007-02-08 : Tomasz Szumlak
//-----------------------------------------------------------------------------

// Declaration of the Algorithm Factory
DECLARE_ALGORITHM_FACTORY( VeloTELL1PedestalSubtractor );

using namespace LHCb;
using namespace VeloTELL1;

//=============================================================================
// Standard constructor, initializes variables
//=============================================================================
VeloTELL1PedestalSubtractor::VeloTELL1PedestalSubtractor(
                           const std::string& name, ISvcLocator* pSvcLocator)
  : VeloTELL1Algorithm ( name , pSvcLocator ),
    m_sensorData ( ),
    m_pedestalLine ( 2048, PEDESTAL ),
    m_pedestalSumMemory ( ),
    m_subPeds ( 0 ),
    m_subPedsLoc ( LHCb::VeloTELL1DataLocation::SubPeds ),
    m_pedestalEngine ( 0 ),
    m_linkPedestal ( 2048, PEDESTAL ),
    m_linkMask ( 2048, 0 ),
    m_headers ( 256, 10 ),
    m_headerCorrectionValueStrip0 ( 4, 40 ),
    m_updateEngine ( 0 )
{
  setTELL1Process(PED_SUBTRACTION);
  setAlgoName("PEDESTAL_SUBTRACTION");
  setAlgoType("TELL1 Pedestal Subtraction");
  // 
  declareProperty("SubtractConstPedestal", m_subtractConstPedestal=false);
  declareProperty("SubtractPedestal", m_subtractPedestal=true);
  declareProperty("InputDataLoc",
                  m_inputDataLoc=VeloTELL1DataLocation::FIRCorrectedADCs);
  declareProperty("OutputDataLoc",
                  m_outputDataLoc=VeloTELL1DataLocation::PedSubADCs);
  declareProperty("ScalingMode", m_scalingMode=CUT_MSB);
  declareProperty("ValidationRun", m_validationRun=false);
  declareProperty("HeaderCorrection", m_headerCorrection=0);
}
//=============================================================================
// Destructor
//=============================================================================
VeloTELL1PedestalSubtractor::~VeloTELL1PedestalSubtractor() {}
//=============================================================================
// Initialization
//=============================================================================
StatusCode VeloTELL1PedestalSubtractor::initialize() {
  StatusCode sc = VeloTELL1Algorithm::initialize(); // must be executed first
  if ( sc.isFailure() ) return sc;  // error printed already by GaudiAlgorithm

  debug() << "==> Initialize" << endmsg;
  //
  return StatusCode::SUCCESS;
}
//=============================================================================
// Main execution
//=============================================================================
StatusCode VeloTELL1PedestalSubtractor::execute() {

  debug() << "==> Execute" << endmsg;
  if(m_validationRun){
    prepareEngine();
    StatusCode exitStatus=validationRun();
    return ( exitStatus );
  }else{
    m_evtNumber++;
    if(!isInitialized()){
      INIT();
      prepareEngine();
    }
    StatusCode dataStatus=getData();
    if(isEnable()){
      if(dataStatus.isSuccess()){            /// check if data is in TES
        dataStatus=inputStream(inputData());
        if(dataStatus.isSuccess()){          /// check if not empty
          runPedestalSub();
          writeData();
          writePedestals();
        }else{
          return ( dataStatus );
        }
      }else{
        return ( dataStatus );
      }
    }else{
      if(dataStatus.isSuccess()){
        cloneData();
      }else{
        return ( dataStatus );
      }
    }
  }
  //
  return StatusCode::SUCCESS;
}
//=============================================================================
//  Finalize
//=============================================================================
StatusCode VeloTELL1PedestalSubtractor::finalize() {

  debug() << "==> Finalize" << endmsg;
  delete m_pedestalEngine;
  delete m_updateEngine;
  // must be called after all other actions
  return VeloTELL1Algorithm::finalize();
}
//=============================================================================
StatusCode VeloTELL1PedestalSubtractor::runPedestalSub()
{
  debug()<< " runPedestalSub() " <<endmsg;
  //
  VeloTELL1Datas::const_iterator sensIt;
  sensIt=inputData()->begin();
  // variable used to count sensors
  unsigned int sensCnt=0;
  // create output container for pedestals if ped algorithm in use
  // and prepare memory for pedestal sum
  if(m_subtractPedestal){
    if(m_evtNumber==1) prepareMemory();
    m_subPeds=new VeloTELL1Datas();
    debug()<< " memory prepared " <<endmsg;
  }
  //
  for( ; sensIt!=inputData()->end(); ++sensIt){
    // get pointer to the velo sensor related with given TELL1
    const unsigned int tell1=static_cast<unsigned int>((*sensIt)->key());
    const DeVeloSensor* sens=deVelo()->sensorByTell1Id(tell1);
    debug()<< " TELL1 number: " << tell1 <<endmsg;
    if(sens==0){
      Error(" ==> No match between TELL1 and sensor found! ");
      Error(" ==> Check your XML conditions file! ");
      return ( StatusCode::FAILURE );
    }
    unsigned int sensNb=sens->sensorNumber();
    debug()<< " sensor number: " << sensNb << " for TELL1 number: "
          << tell1 <<endmsg;
    // check if the sensor is read out
    if(sens->isReadOut()){
      // data from current sensor
      m_sensorData.clear();
      m_sensorData=(*sensIt)->data();
      assert(m_sensorData.size()==ALL_STRIPS);
      // pedestal subtraction - const pedestal value
      if(m_subtractConstPedestal) subtractConstPedestal(tell1);
      // pedestal subtraction - const pedestal value
      ///
      // pedestal subtraction - weighted average
      if(m_subtractPedestal) subtractPedestal(tell1, sensCnt);
      // pedestal subtraction - weighted average
    }
    ++sensCnt;
  }
  return ( StatusCode::SUCCESS );
}
//=============================================================================
StatusCode VeloTELL1PedestalSubtractor::subtractConstPedestal(
                                        const unsigned int tell1)
{
  debug()<< " subtractConstPedestal() " <<endmsg;
  // for the time being just const pedestal subtraction
  // is implemented according to tell1lib_v1.9.2
  sdataVec subPedestals (2304, 0);
  std::transform(m_sensorData.begin(), m_sensorData.end(),
                 m_pedestalLine.begin(),
                 subPedestals.begin(),
                 std::minus<signed int>());
  // data scaling -> mode 0
  sdatIt iT=subPedestals.begin();
  for( ; iT!=subPedestals.end(); ++iT){
    if((*iT)>MAX_ADC_8_BIT) (*iT)=MAX_ADC_8_BIT;
    else if((*iT)<(-MIN_ADC_8_BIT)) (*iT)=(-MAX_ADC_8_BIT);
  }
  // create a new VeloTELL1Data object for subtracted data
  VeloTELL1Data* subData=new VeloTELL1Data(tell1, VeloFull);
  subData->setDecodedData(subPedestals);
  outputData(subData);
  //
  return ( StatusCode::SUCCESS );
}
//=============================================================================
StatusCode VeloTELL1PedestalSubtractor::subtractPedestal(
           const unsigned int tell1, const unsigned int sensCnt)
{
  debug()<< " ==> subractPedestal() " <<endmsg;
  //
  VeloTELL1Data* subPeds=new VeloTELL1Data(tell1, VeloPedestal);
  VeloTELL1Data* subData=new VeloTELL1Data(tell1, VeloFull);
  // rawADCs    - data before ped subtraction
  // subPedADCs - data after correction
  // peds       - subtracted pedestals
  sdataVec rawADCs, subPedADCs(2048, 0), peds(2048, 0);
  // get raw ADCs for current sensor
  rawADCs=m_sensorData;
  // get sum of the pedestals
  lpIT lIT=m_linkPedestal.begin();
  sdataVec pedestalSum=m_pedestalSumMemory[sensCnt];
  sdatIt psIT=pedestalSum.begin();
  // transform pedestal sum into pedestals
  for( ; psIT!=pedestalSum.end(); ++psIT){
    *lIT++=((*psIT)>>PED_SUM_SHIFT);
  }
  // subtract pedestals that were calculated for the previous event
  // pass input data to the engine object
  m_pedestalEngine->setInData(rawADCs);
  // pedestal sum is in different format - c-modules issue
  extractPedestals();
  m_pedestalEngine->setLinkPedestal(m_linkPedestal);
  m_pedestalEngine->runSubtraction();
  subPedADCs=m_pedestalEngine->outData();
  // translate pedestals from u_int16_t into signed int
  lIT=m_linkPedestal.begin();
  psIT=peds.begin();
  for( ; lIT!=m_linkPedestal.end(); ++lIT){
    *psIT++=static_cast<signed int>(*lIT);
  }
  // switch off the data storage if in validation mode
  if(!m_validationRun){
    subData->setDecodedData(subPedADCs);
    subPeds->setDecodedData(peds);
    outputData(subData);
    m_subPeds->insert(subPeds);
  }
  // calculate pedestals for the next event by refreshing ped sum
  m_updateEngine->setInData(rawADCs);
  m_updateEngine->setPedestalSum(pedestalSum);
  m_updateEngine->runUpdate();
  m_pedestalSumMemory[sensCnt]=m_updateEngine->updatedPedestalSum();
  // write out text files if running in validation mode
  if(m_validationRun){
    writeDataToFile("p_file.txt", m_pedestalSumMemory[sensCnt]);
    writeDataToFile("d_ped_sub.txt", subPedADCs);
  }
  //
  return ( StatusCode::SUCCESS );
}
//=============================================================================
StatusCode VeloTELL1PedestalSubtractor::prepareMemory()
{
  debug()<< " ==> prepareMemory() " <<endmsg;
  //
  if(m_validationRun){
    // create storage space just for data from one sensor
    sensorRAM pedSumRAM;
    m_pedestalSumMemory.push_back(pedSumRAM);
  }else{
    VeloTELL1Datas* dataPtr=inputData();
    int memorySize=dataPtr->size();
    // create the place where all pedestals will be stored
    for(int sensMem=0; sensMem<memorySize; ++sensMem){
      // pedestal sum storage
      sensorRAM pedSumRAM(2048, SUM_RAM_INIT);
      // add dummy strips
      prepareData(pedSumRAM);
      m_pedestalSumMemory.push_back(pedSumRAM);
    }
  }
  //
  return ( StatusCode::SUCCESS );
}
//=============================================================================
StatusCode VeloTELL1PedestalSubtractor::writePedestals()
{
  debug()<< " ==> writePedestals() " <<endmsg;
  //
  put(m_subPeds, m_subPedsLoc);
  //
  return ( StatusCode::SUCCESS );
}
//=============================================================================
StatusCode VeloTELL1PedestalSubtractor::validationRun()
{
  debug()<< " validationRun() " <<endmsg;
  // prepare memory for generated data - we will emulate just one sensor
  // the member variable m_validationRun will be checked
  prepareMemory();
  // read data and pedestals
  m_sensorData.clear();
  StatusCode dataStatus;
  dataStatus=readDataFromFile("d_source.txt", m_sensorData);
  if(!dataStatus.isSuccess()){
    return ( dataStatus );
  }
  dataStatus=readDataFromFile("pp_file.txt", m_pedestalSumMemory[0]);
  if(!dataStatus.isSuccess()){
    return ( dataStatus );
  }
  StatusCode testRunStatus=subtractPedestal(0, 0);
  //
  return ( testRunStatus );
}
//=============================================================================
void VeloTELL1PedestalSubtractor::extractPedestals()
{
  debug()<< " ==> extractPedestals() " <<endmsg;
  // use this method to extract pedestals - validation run
  unsigned int proc=0;
  lpIT iT1, iT2;
  LinkPedestalVec temp;
  while(proc<NumberOfPPFPGA){
    iT1=m_linkPedestal.begin()+proc*(STRIPS_IN_PPFPGA+DUMMY_STRIPS);
    iT2=iT1+(STRIPS_IN_PPFPGA);
    for( ; iT1!=iT2; ++iT1){
      temp.push_back(*iT1);
    }
    ++proc;
  }
  assert(temp.size()==SENSOR_CHANNELS);
  m_linkPedestal=temp;
}
//=============================================================================
void VeloTELL1PedestalSubtractor::prepareEngine()
{
  debug()<< " ==> prepareEngine() "<<endmsg;
  // pedestal engine
  m_pedestalEngine=new TELL1PedestalProcessEngine();
  m_pedestalEngine->setProcessEnable(isEnable());
  m_pedestalEngine->setScalingMode(m_scalingMode);
  m_pedestalEngine->setHeaderCorrectionFlag(m_headerCorrection);
  m_pedestalEngine->setZSEnable(1);
  m_pedestalEngine->setHeaders(m_headers);
  m_pedestalEngine->setLinkMask(m_linkMask);
  m_pedestalEngine->setHeaderCorrectionValues(m_headerCorrectionValueStrip0);
  // update engine - end of pedestal following cycle
  m_updateEngine=new TELL1UpdateProcessEngine();
  m_updateEngine->setProcessEnable(isEnable());
  m_updateEngine->setHeaderCorrectionFlag(0);
  m_updateEngine->setHeaders(m_headers);
}
//--
