// $Id: VeloTELL1Reordering.cpp,v 1.2 2007/08/26 16:39:15 szumlat Exp $
// Include files 

// from Gaudi
#include "GaudiKernel/DeclareFactoryEntries.h" 

// stl
#include <vector>
#include <algorithm>

// local
#include "Kernel/ContPrinter.h"
#include "VeloTELL1Reordering.h"

// boost
#include <boost/assign/std/vector.hpp>
#include "boost/lexical_cast.hpp"

// engine classes
#include "TELL1Engine/TELL1ReorderProcessEngine.h"

//-----------------------------------------------------------------------------
// Implementation file for class : VeloTELL1Reordering
//
// 2007-02-08 : Tomasz Szumlak
//-----------------------------------------------------------------------------

// Declaration of the Algorithm Factory
DECLARE_ALGORITHM_FACTORY( VeloTELL1Reordering );

using namespace boost::assign;
using namespace LHCb;

//=============================================================================
// Standard constructor, initializes variables
//=============================================================================
VeloTELL1Reordering::VeloTELL1Reordering( const std::string& name,
                                          ISvcLocator* pSvcLocator)
  : VeloTELL1Algorithm ( name , pSvcLocator ),
    m_reorderedData ( ),
    m_printInfo ( false ),
    m_reorderingMode ( NO_REORDERING ),
    m_reorderEngine ( 0 )
{
  setTELL1Process(CHANNEL_REORDERING);
  setAlgoName("CHANNEL_REORDERING");
  setAlgoType("TELL1 channel reordering");
  declareProperty("ValidationRun", m_validationRun=false);
  declareProperty("InputDataLoc",
                  m_inputDataLoc=VeloTELL1DataLocation::PedSubADCs);
  declareProperty("OutputDataLoc",
                  m_outputDataLoc=VeloTELL1DataLocation::ReorderedADCs);
  declareProperty("SectorCorrection", m_funnyfix=APPLY_SECTOR_CORR);  

  // Switches on a custom re-ordering algorithm
  declareProperty("CustomReordering", m_customReordering=false);
  // Chooses the specific custom method to use
  declareProperty("CustomMethod", m_customMethod=0);
}
//=============================================================================
// Destructor
//=============================================================================
VeloTELL1Reordering::~VeloTELL1Reordering() {}
//=============================================================================
// Initialization
//=============================================================================
StatusCode VeloTELL1Reordering::initialize() {
  StatusCode sc = VeloTELL1Algorithm::initialize(); // must be executed first
  if ( sc.isFailure() ) return sc;  // error printed already by GaudiAlgorithm

  debug() << "==> Initialize" << endmsg;
  //
  return StatusCode::SUCCESS;
}
//=============================================================================
// Main execution
//=============================================================================
StatusCode VeloTELL1Reordering::execute() {

  debug() << "==> Execute" << endmsg;
  //
  m_evtNumber++;
  if(!isInitialized()){
    INIT();
    prepareEngine();
  }
  StatusCode dataStatus;
  debug() << "Enable test " << this->isEnable() << endmsg;
  debug() << "Event no " << m_evtNumber << endmsg;

// we can use the re-ordering either if it is specified by the detector data, or if the CustomReordering option is set

  if((this->isEnable()||m_customReordering)&&(m_evtNumber>convergenceLimit())){
  debug() << "isEnabled" << endmsg;
    dataStatus=getData();
    if(dataStatus.isSuccess()){              /// check if data is in TES
      dataStatus=inputStream(inputData());
      if(dataStatus.isSuccess()){            /// check if not empty
        topologicalChannelReordering();
        writeData();
  debug() << "writeData function was used" << endmsg;
      }else{
        return ( dataStatus );
      }
    }else{
      return ( dataStatus );
    }
  }else{
    dataStatus=getData();
    if(dataStatus.isSuccess()){
      cloneData();
    }else{
      return ( dataStatus );
    }
  }
  return StatusCode::SUCCESS;
}
//=============================================================================
//  Finalize
//=============================================================================
StatusCode VeloTELL1Reordering::finalize() {

  debug() << "==> Finalize" << endmsg;
  // must be called after all other actions
  delete m_reorderEngine;
  return VeloTELL1Algorithm::finalize();
}
//=============================================================================
StatusCode VeloTELL1Reordering::topologicalChannelReordering()
{

// The usage of this function depends on whether custom re-ordering is set. For the moment, I'll just see how it goes.

  debug()<< " ==> topologicalChannelReordering() " <<endmsg;
  //
  VeloTELL1Datas::iterator sensIt=inputData()->begin();
  // loop over all sensors
  for( ; sensIt!=inputData()->end(); ++sensIt){
    // get info about current sensor
    const int tell1=(*sensIt)->key();
    unsigned int uTELL1=static_cast<unsigned int>(tell1);
    // !!for reordering studies!!
    const DeVeloSensor* sens=deVelo()->sensorByTell1Id(uTELL1);
    debug()<< " tell1: " << tell1 <<endmsg;
    // check if the sensor pointer is valid
    if(sens==0){
      Error(" ==> No match between TELL1 and sensor found! ");
      Error(" ==> Check your XML conditions file! ");
      return ( StatusCode::FAILURE );
    }
    unsigned int sensNb=sens->sensorNumber();
    debug()<< " the sensor number: " << sensNb << " for TELL1 number: "
          << uTELL1 <<endmsg;
    if(sens->isReadOut()){
      sdataVec inData=(*sensIt)->data();
      // check the sensor type and set the reordering mode
     // if we're using custom re-ordering, change the re-order mode and don't do the cable correction.
     
     if (m_customReordering==true){
     m_reorderEngine->setReorderMode(NO_REORDERING);
     debug()<< " Custom re-order - use no reordering for now " <<endmsg;
     }     
     else {
      if(sens->isR()||sens->isPileUp()){
        m_reorderEngine->setReorderMode(R_SENSOR);
        debug()<< " R sensor " <<endmsg;
        cableErrorCorrection(inData);
      }else if(sens->isPhi()){
        m_reorderEngine->setReorderMode(PHI_SENSOR);
        debug()<< " Phi sensor " <<endmsg;
        cableErrorCorrection(inData);
      }
     }

     // At this point, run custom re-ordering if required. Note it may be possible to introduce another parameter to switch between different algorithms.
      if (m_customReordering==true){
	//cableErrorCorrection(inData);
        customReorder(inData); 
        // store and write out reordered data
        VeloTELL1Data* reorderedData=new VeloTELL1Data(tell1, VeloFull);
        sdataVec outData=inData;
        debug()<< " Writing output of custom reordering " <<endmsg;
        reorderedData->setDecodedData(outData);
        outputData(reorderedData);
	}

     else{
      m_reorderEngine->setInData(inData);
      m_reorderEngine->runReordering();
      // store and write out reordered data
      VeloTELL1Data* reorderedData=new VeloTELL1Data(tell1, VeloFull);
      sdataVec outData=m_reorderEngine->outData();
  debug()<< " re-ordering engine used " <<endmsg;
      reorderedData->setDecodedData(outData);
      outputData(reorderedData);
     }
    }
  }
  return ( StatusCode::SUCCESS );
}
//=============================================================================
void VeloTELL1Reordering::cableErrorCorrection(sdataVec& inVec)
{
  debug()<< " ==> cableErrorCorrection() " <<endmsg;
  debug()<< " ==> No of strips in PPFPGA = " << STRIPS_IN_PPFPGA <<endmsg;
  debug()<< " ==> No of dummy strips = " << DUMMY_STRIPS <<endmsg;
  sdataVec tmp=inVec;
  sdatIt IT;
  sdatIt IU;
  inVec.clear();
  unsigned int indexI=0, indexF=0, indexP=0;
  indexI=3*STRIPS_IN_PPFPGA+3*DUMMY_STRIPS;
  indexF=indexI+STRIPS_IN_PPFPGA+DUMMY_STRIPS;
  indexP=0;
  IU=inVec.begin()+indexP;
  for(IT=tmp.begin()+indexI; IT!=tmp.begin()+indexF; ++IT){
    (*IU)=(*IT); // Transfer data from tmp to output vector 
    ++IU;
  }
  indexI=2*STRIPS_IN_PPFPGA+2*DUMMY_STRIPS;
  indexF=indexI+STRIPS_IN_PPFPGA+DUMMY_STRIPS;
  indexP=STRIPS_IN_PPFPGA+DUMMY_STRIPS;
  IU=inVec.begin()+indexP;
  for(IT=tmp.begin()+indexI; IT!=tmp.begin()+indexF; ++IT){
    (*IU)=(*IT); // Transfer data from tmp to output vector 
    ++IU;
  }
  indexI=STRIPS_IN_PPFPGA+DUMMY_STRIPS;
  indexF=indexI+STRIPS_IN_PPFPGA+DUMMY_STRIPS;
  indexP=2*STRIPS_IN_PPFPGA+2*DUMMY_STRIPS;
  IU=inVec.begin()+indexP;
  for(IT=tmp.begin()+indexI; IT!=tmp.begin()+indexF; ++IT){
    (*IU)=(*IT); // Transfer data from tmp to output vector 
    ++IU;
  }
  indexI=0;
  indexF=indexI+STRIPS_IN_PPFPGA+DUMMY_STRIPS;
  indexP=3*STRIPS_IN_PPFPGA+3*DUMMY_STRIPS;
  IU=inVec.begin()+indexP;
  for(IT=tmp.begin()+indexI; IT!=tmp.begin()+indexF; ++IT){
    (*IU)=(*IT); // Transfer data from tmp to output vector 
    ++IU;
  }
}
//=============================================================================
void VeloTELL1Reordering::prepareEngine()
{
  debug()<< " ==> prepareEngine() " <<endmsg;
  //
  m_reorderEngine=new TELL1ReorderProcessEngine();
  m_reorderEngine->setCableCorrection(0);
  // the engine runs actually with mode flag not process enable
  // flag, the mode flag is set depending on type of sensor
}
//--



//=============================================================================
void VeloTELL1Reordering::customReorder(sdataVec& inVec)
{

// This is code for doing custom re-ordering. Potentially I could have multiple functions available, and a parameter to choose which one to use.
// Here, I will switch the blocks of data around a bit.

  debug()<< " ==> customReorder() " <<endmsg;
  sdataVec tmp=inVec;
  sdatIt IT;
  sdatIt IU;
  debug()<< " ==> Size of inVec = " << inVec.size() <<endmsg;
  debug()<< " ==> Size of tmp = " << tmp.size() <<endmsg;
  unsigned int indexI=0, indexF=0, indexP=0;
  int channelIn=0, channelOut=0, channelTest=0;

// Use a case statement to choose the reordering method. Default just leaves order unchanged.

switch (m_customMethod){

  // CASE 1 - SHIFT THE 4 BLOCKS OF 512 STRIPS AS A TEST
  case 1:

  indexI=0;
  indexF=512; // in tmp, go from 0 to 511
  indexP=512; // in output vector, start at 512
  IU=inVec.begin()+indexP;
  for(IT=tmp.begin()+indexI; IT!=tmp.begin()+indexF; ++IT){
    (*IU)=(*IT); // Transfer data from tmp to output vector 
    ++IU;
  }
  
  debug()<< " Part 1 worked! " <<endmsg;

  indexI=512;
  indexF=1024; // in tmp, go from 512 to 1023
  indexP=0; // in output vector, start at 0
  IU=inVec.begin()+indexP;
  for(IT=tmp.begin()+indexI; IT!=tmp.begin()+indexF; ++IT){
    (*IU)=(*IT); // Transfer data from tmp to output vector 
    ++IU;
  }
  
  debug()<< " Part 2 worked! " <<endmsg;


  indexI=1024;
  indexF=1536; // work over strips 1024 to 1535 for now.
  indexP=1536; // in output vector, start at 1536
  IU=inVec.begin()+indexP;

  debug()<< "Start of loop." <<endmsg;
  for(IT=tmp.begin()+indexI; IT!=tmp.begin()+indexF; ++IT){
    (*IU)=(*IT);
    ++IU;
  }

  debug()<< " Part 3 worked! " <<endmsg;

  indexI=1536;
  indexF=2048; // work over strips 1536 to 2047 for now.
  indexP=1024;
  IU=inVec.begin()+indexP;
//  debug()<< "Start of loop." <<endmsg;
  for(IT=tmp.begin()+indexI; IT!=tmp.begin()+indexF; ++IT){
//  debug()<< "In the loop" <<endmsg;
    (*IU)=(*IT);
    ++IU;
  }

  debug()<< " Part 4 worked! " <<endmsg;
  break;




  //**********
  // CASE 2 - Freiburg sensors re-ordered, all others unchanged.
  case 2:

  // Fre2 sensor. Loop over only the strips which have the sensor attached.

  channelIn=0; // use these to keep track of input channels
  channelOut=0; // more intuitive than the indexes

  // Fre2 sensor. Loop over only the strips which have the sensor attached.
  debug()<< "Start of Fre 2" <<endmsg;
  for(channelIn=1175; channelIn<=1238; ++channelIn){ // loop over required input channels

    IT=tmp.begin()+channelIn; // set the pointer to strip channelIn in temporary data vector
	
    // Choose the channel we send this strip to, using a case statement
    channelTest=channelIn%4; // use modular arithmetic test

    switch (channelTest){
	case 0:
	channelOut = 2390-channelIn;
	break;

	case 1:
	channelOut = 2389-channelIn;
	break;

	case 2:
	channelOut = 2391-channelIn;
	break;

	case 3:
	channelOut = 2390-channelIn;
	break;
	}

    IU=inVec.begin()+channelOut; // set the pointer to strip channelOut in data vector
    (*IU)=(*IT); // copy data from temp to data vector, overwriting the existing entry
  }

  // Afterwards, want to clear all the other "Fre2" strips in the output, i.e. 1216 to 1279 inclusive
  for(channelOut=1216; channelOut<=1279; ++channelOut){  
    IU=inVec.begin()+channelOut;
    (*IU)=0; // clear this channel - hope it works ok.
  }
 
  // We also want to clear the channels where the headers appear. Basically, the headers no longer appear at the start of 32-blocks after re-ordering, so the header threshold won't help them. Blank 1172, 1174, 1204, 1206
  //IU=inVec.begin()+1172;
  //(*IU)=0; 
  //IU=inVec.begin()+1174;
  //(*IU)=0; 
  //IU=inVec.begin()+1204;
  //(*IU)=0; 
  //IU=inVec.begin()+1206;
  //(*IU)=0; 

  // Fre0 sensor. Take the same approach.

  channelIn=0; // use these to keep track of input channels
  channelOut=0; // more intuitive than the indexes

  // Fre2 sensor. Loop over only the strips which have the sensor attached.
  debug()<< "Start of Fre0." <<endmsg;
  for(channelIn=1451; channelIn<=1514; ++channelIn){ // loop over required input channels

    IT=tmp.begin()+channelIn; // set the pointer to strip channelIn in temporary data vector
	
    // Choose the channel we send this strip to, using a case statement
    channelTest=channelIn%4; // use modular arithmetic test

    switch (channelTest){
	case 0:
	channelOut = 2921-channelIn;
	break;

	case 1:
	channelOut = 2920-channelIn;
	break;

	case 2:
	channelOut = 2922-channelIn;
	break;

	case 3:
	channelOut = 2921-channelIn;
	break;
	}

    IU=inVec.begin()+channelOut; // set the pointer to strip channelOut in data vector
    (*IU)=(*IT); // copy data from temp to data vector, overwriting the existing entry
  }

  // Afterwards, want to clear all the other "Fre0" strips in the output, i.e. 1472 to 1535 inclusive
  for(channelOut=1472; channelOut<=1535; ++channelOut){  
    IU=inVec.begin()+channelOut;
    (*IU)=0; // clear this channel - hope it works ok.
  }

  // We also want to clear the channels where the headers appear. Basically, the headers no longer appear at the start of 32-blocks after re-ordering, so the header threshold won't help them. Blank 1415, 1417, 1447 and 1449
  //IU=inVec.begin()+1415; 
  //(*IU)=0; 
  //IU=inVec.begin()+1417;
  //(*IU)=0; 
  //IU=inVec.begin()+1447;
  //(*IU)=0; 
  //IU=inVec.begin()+1449;
  //(*IU)=0; 



  break;


  //**********
  // CASE 3 - Re-ordering for irrad 3D module - Oct '08.
  case 3:



  channelIn=0; // use these to keep track of input channels
  channelOut=0; // more intuitive than the indexes

  // Fre2 sensor. Loop over only the first module
  debug()<< "Start of irrad 3D" <<endmsg;
  for(channelIn=0; channelIn<=511; ++channelIn){ // loop over required input channels

    IT=tmp.begin()+channelIn; // set the pointer to strip channelIn in temporary data vector
	
    // Choose the channel we send this strip to, using a case statement
    channelTest=channelIn%4; // use modular arithmetic test
    // On strips 256, 260 etc don't change. Likewise, on 259, 263 don't change. However, other strips should be swapped.

    switch (channelTest){
	case 0:
	channelOut = channelIn;
	break;

	case 1:
	channelOut = channelIn+1;
	break;

	case 2:
	channelOut =channelIn-1;
	break;

	case 3:
	channelOut = channelIn;
	break;
	}

    IU=inVec.begin()+channelOut; // set the pointer to strip channelOut in data vector
    (*IU)=(*IT); // copy data from temp to data vector, overwriting the existing entry
  }

  break;


  // DEFAULT CASE - DON'T DO ANYTHING
  default: // Don't do reordering at all
  indexI=0; // Might as well do something

} // bracket at end of switch statement


} // bracket for end of declaration
//=============================================================================


//=============================================================================
void VeloTELL1Reordering::trialReorder(sdataVec& inVec)
{
  debug()<< " ==> cableErrorCorrection(), renamed trialReorder " <<endmsg;
  debug()<< " ==> No of strips in PPFPGA = " << STRIPS_IN_PPFPGA <<endmsg;
  debug()<< " ==> No of dummy strips = " << DUMMY_STRIPS <<endmsg;
  sdataVec tmp=inVec;
  sdatIt IT;
  inVec.clear();
  unsigned int indexI=0, indexF=0;
  indexI=3*STRIPS_IN_PPFPGA+3*DUMMY_STRIPS;
  indexF=indexI+STRIPS_IN_PPFPGA+DUMMY_STRIPS;
  std::insert_iterator<sdataVec> ii(inVec, (inVec.begin()));
  for(IT=tmp.begin()+indexI; IT!=tmp.begin()+indexF; ++IT){
  debug()<< "In the loop" <<endmsg;
    *ii=(*IT);
  debug()<< "Shift data from tmp to inVec" <<endmsg;
    *ii++;
  debug()<< "Increment ii pointer" <<endmsg;
  }
  indexI=2*STRIPS_IN_PPFPGA+2*DUMMY_STRIPS;
  indexF=indexI+STRIPS_IN_PPFPGA+DUMMY_STRIPS;
  std::insert_iterator<sdataVec> ij(inVec, (inVec.begin()+(indexI/2)));
  for(IT=tmp.begin()+indexI; IT!=tmp.begin()+indexF; ++IT){
    *ij++=(*IT);
  }
  indexI=STRIPS_IN_PPFPGA+DUMMY_STRIPS;
  indexF=indexI+STRIPS_IN_PPFPGA+DUMMY_STRIPS;
  std::insert_iterator<sdataVec> ik(inVec, (inVec.begin()+(2*indexI)));
  for(IT=tmp.begin()+indexI; IT!=tmp.begin()+indexF; ++IT){
    *ik++=(*IT);
  }
  indexI=0;
  indexF=indexI+STRIPS_IN_PPFPGA+DUMMY_STRIPS;
  std::insert_iterator<sdataVec> il(inVec, (inVec.begin()+(3*indexF)));
  for(IT=tmp.begin()+indexI; IT!=tmp.begin()+indexF; ++IT){
    *il++=(*IT);
  }
}
//=========================================================
