// $Id: VeloBitLimitMoni.cpp,v 1.4 2007/08/26 19:20:05 szumlat Exp $
// Include files 

// from Gaudi
#include "GaudiKernel/DeclareFactoryEntries.h" 

// local
#include "VeloBitLimitMoni.h"

// kernel
#include "Kernel/DecodeConf.h"
#include "Kernel/DecodeCore.h"
#include "Kernel/ContPrinter.h"

// boost
#include "boost/lexical_cast.hpp"

//-----------------------------------------------------------------------------
// Implementation file for class : VeloBitLimitMoni
//
// 2006-08-22 : Tomasz Szumlak
//-----------------------------------------------------------------------------

// Declaration of the Algorithm Factory
DECLARE_ALGORITHM_FACTORY( VeloBitLimitMoni );

using namespace LHCb;
using namespace VeloTELL1;

//=============================================================================
// Standard constructor, initializes variables
//=============================================================================
VeloBitLimitMoni::VeloBitLimitMoni( const std::string& name,
                                    ISvcLocator* pSvcLocator)
  : GaudiTupleAlg ( name , pSvcLocator ),
    m_bitLimit ( 0 ),
    m_bitLimitLoc ( LHCb::VeloTELL1DataLocation::BitLimitADCs ),
    m_numberOfEvents ( 0 )
{
  declareProperty("PlotBitLimitData", m_plotBitLimit=false);
  declareProperty("SubtractorType", m_subType=1);
  declareProperty("SizeOfNTupleArray", m_sizeOfNTupleArray=2048);
  declareProperty("StoreNTuple", m_storeNTuple=false);
}
//=============================================================================
// Destructor
//=============================================================================
VeloBitLimitMoni::~VeloBitLimitMoni() {}
//=============================================================================
// Initialization
//=============================================================================
StatusCode VeloBitLimitMoni::initialize() {
  StatusCode sc = GaudiTupleAlg::initialize(); // must be executed first
  if ( sc.isFailure() ) return sc;  // error printed already by GaudiAlgorithm

  debug() << "==> Initialize" << endmsg;
  setHistoTopDir("Vetra/");
  // 
  if(m_subType==FAST_SUBTRACTOR){
    m_convergenceLimit=FAST_CONVERGENCE_LIMIT;
  }else{
    m_convergenceLimit=CONVERGENCE_LIMIT;
  }
  //
  return StatusCode::SUCCESS;
}
//=============================================================================
// Main execution
//=============================================================================
StatusCode VeloBitLimitMoni::execute() {

  debug() << "==> Execute" << endmsg;
  //
  m_numberOfEvents++;
  if(m_numberOfEvents>m_convergenceLimit){
    StatusCode dataStatus=getData();
    if(dataStatus.isSuccess()){
      plotSaturatedData();
      if(m_storeNTuple) storeNTuple();
    }else{
      return ( dataStatus );
    }
  }
  //
  return StatusCode::SUCCESS;
}
//=============================================================================
//  Finalize
//=============================================================================
StatusCode VeloBitLimitMoni::finalize() {

  debug() << "==> Finalize" << endmsg;

  return GaudiTupleAlg::finalize();  // must be called after all other actions
}
//=============================================================================
StatusCode VeloBitLimitMoni::getData()
{
  debug()<< " ==> getData()" <<endmsg;
  // get ADC samples after bit saturation
  // it does not matter if one deals with real or
  // simulated data 
  if(!exist<LHCb::VeloTELL1Datas>(m_bitLimitLoc)){
    error()<< "There is no bit saturated ADC samples in TES!" 
           << "at location: " << m_bitLimitLoc <<endmsg;
  }else{
    m_bitLimit=get<LHCb::VeloTELL1Datas>(m_bitLimitLoc);
  }
  //
  return ( StatusCode::SUCCESS );
}
//=============================================================================
StatusCode VeloBitLimitMoni::plotSaturatedData()
{
  debug()<< " plotSaturatedData() " <<endmsg;
  //
  if(m_plotBitLimit){
    VeloTELL1Datas::const_iterator sensIt=m_bitLimit->begin();
    for( ; sensIt!=m_bitLimit->end(); ++sensIt){
      sdataVec ALinkADC;
      scdatIt iT;
      unsigned int channel=0;
      int TELL1=(*sensIt)->key();
      VeloTELL1Data* adcData=(*sensIt);
      //sdataVec data=adcData->data();
      //if(m_numberOfEvents>m_convergenceLimit) contPrinter(data, 5);
      for(int aLink=0; aLink<NumberOfALinks; ++aLink){
        ALinkADC=(*adcData)[aLink];
        for(iT=ALinkADC.begin(); iT!=ALinkADC.end(); ++iT){
          plot2D(channel, (*iT), TELL1,
                 "Saturated ADCs data for TELL1 number "+
                 boost::lexical_cast<std::string>(TELL1),
                 0., 2048, -50., 150., 2048, 200);
          ++channel;
          // projection of the saturated data
          plot((*iT), TELL1*10,
               "Saturated data projection (noise) for TELL1-"+
               boost::lexical_cast<std::string>(TELL1),
               //-10, 10, 20);
               -6, 6, 12);
          plot((*iT), TELL1*100,
               "Saturated data projection (signal) for TELL1-"+
               boost::lexical_cast<std::string>(TELL1),
               //10, 130, 120);
               6, 60, 54);
        }
      }
    }
  }
  //
  return ( StatusCode::SUCCESS );
}
//=============================================================================
StatusCode VeloBitLimitMoni::storeNTuple()
{
  debug()<< " ==> storeNTuple() " <<endmsg;
  //
  sdataVec data;
  std::string columnName;
  std::string columnTitle="RawData";
  Tuple rawTuple=nTuple(100, "RawTELL1Data", CLID_ColumnWiseTuple);
  VeloTELL1Datas::const_iterator sensIt=m_bitLimit->begin();
  for( ; sensIt!=m_bitLimit->end(); ++sensIt){
    int TELL1=(*sensIt)->key();
    VeloTELL1Data* adcData=(*sensIt);
    data=adcData->data();
    columnName="RawdataforTELL1"+
               boost::lexical_cast<std::string>(TELL1);
    rawTuple->farray(columnName, data, columnTitle, m_sizeOfNTupleArray);
  }
  rawTuple->column("EventNumber", m_numberOfEvents);
  // write out the NTuple
  rawTuple->write();
  //
  return ( StatusCode::SUCCESS );
}
//
