
// $Id: VeloLCMSMoni.cpp,v 1.2 2007/08/26 19:20:05 szumlat Exp $
// Include files 

// from Gaudi
#include "GaudiKernel/DeclareFactoryEntries.h" 

// local
#include "VeloLCMSMoni.h"

// kernel
#include "Kernel/DecodeConf.h"
#include "Kernel/DecodeCore.h"
#include "Kernel/ContPrinter.h"

// boost
#include "boost/lexical_cast.hpp"

//-----------------------------------------------------------------------------
// Implementation file for class : VeloLCMSMoni
//
// 2007-02-09 : Tomasz Szumlak
//-----------------------------------------------------------------------------

// Declaration of the Algorithm Factory
DECLARE_ALGORITHM_FACTORY( VeloLCMSMoni );

using namespace LHCb;
using namespace VeloTELL1;

//=============================================================================
// Standard constructor, initializes variables
//=============================================================================
VeloLCMSMoni::VeloLCMSMoni( const std::string& name,
                            ISvcLocator* pSvcLocator)
  : GaudiTupleAlg ( name , pSvcLocator ),
    m_procInfos ( 0 ),
    m_procInfosLoc ( VeloProcessInfoLocation::Default ),
    m_LCMSADCs ( 0 ),
    m_LCMSADCsLoc ( LHCb::VeloTELL1DataLocation::CMSuppressedADCs ),
    m_eventNumber ( 0 ),
    m_isInitialized ( false ),
    m_convergenceLimit ( ),

    m_hitPosX ( 0 ),
    m_hitADCValue ( 0 ),
    m_lADCValue ( 0 ), 
    m_rADCValue ( 0 ), 
    m_totADCValue ( 0 ), 

    m_hitId ( 0 ),

    m_numberOfHits ( 0 ),
    m_adcMax ( 0 ),
    m_posXMax ( -1 ),

 
    m_bunchCounter ( 0 ),
    m_BCntLoop ( 0 ),
    m_doesCounterLoop ( false ),


    m_BATTrig ( 0 ),
    m_orderRaw ( 0 ),

    m_evtInfos ( 0 ),
    m_evtInfoLoc ( EvtInfoLocation::Default )
{
  declareProperty("PlotLCMS", m_plotLCMS=true);
  declareProperty("HitThreshold", m_hitThreshold=6);
  declareProperty("HitThresholdGla0", m_hitThresholdGla0=6);
  declareProperty("HitThresholdGla1", m_hitThresholdGla1=6);
  declareProperty("HitThresholdGla2", m_hitThresholdGla2=6);
  declareProperty("HitThresholdFre0", m_hitThresholdFre0=6);
  declareProperty("HitThresholdFre1", m_hitThresholdFre1=6);
  declareProperty("HitThresholdFre2", m_hitThresholdFre2=6);
  declareProperty("HitThresholdGlaHeader", m_hitThresholdGlaHeader=20);
  declareProperty("HitThresholdFreHeader", m_hitThresholdFreHeader=20);

  declareProperty("HeaderWidth", m_headerWidth=2);
  declareProperty("nSamples", m_nSamples=1);

  declareProperty("ModuleNumber", m_moduleNumber=0);
  declareProperty("DetectorNumber", m_detectorNumber=0);

  declareProperty("WriteAll", m_writeAll=0);

}
//=============================================================================
// Destructor
//=============================================================================
VeloLCMSMoni::~VeloLCMSMoni() {} 

//=============================================================================
// Initialization
//=============================================================================
StatusCode VeloLCMSMoni::initialize() {
  StatusCode sc = GaudiAlgorithm::initialize(); // must be executed first
  if ( sc.isFailure() ) return sc;  // error printed already by GaudiAlgorithm

  debug() << "==> Initialize" << endmsg;
  //
  setHistoTopDir("Vetra/");
  //
  return StatusCode::SUCCESS;
}

//=============================================================================
// Main execution
//=============================================================================
StatusCode VeloLCMSMoni::execute() {

  debug() << "==> Execute" << endmsg;
  //
  ++m_eventNumber;
  if(!m_isInitialized){
    if(!exist<VeloProcessInfos>(m_procInfosLoc)){
      debug()<< " ==> No init info available! " <<endmsg;
    }else{
      m_procInfos=get<VeloProcessInfos>(m_procInfosLoc);
    }
    m_isInitialized=true;
    VeloProcessInfos::const_iterator procIt=m_procInfos->begin();
    if(!((*procIt)->convLimit())){
      Error("Convergence Limit not set!", StatusCode::FAILURE);
    }else{
      m_convergenceLimit=(*procIt)->convLimit();
    }
  }
  if(m_eventNumber>m_convergenceLimit){
    getData();
    plotLCMSData();
  }
  //
  return StatusCode::SUCCESS;
}
//=============================================================================
//  Finalize
//=============================================================================
StatusCode VeloLCMSMoni::finalize() {

  debug() << "==> Finalize" << endmsg;

  return GaudiAlgorithm::finalize();  // must be called after all other actions
}
//=============================================================================
StatusCode VeloLCMSMoni::getData()
{
  debug()<< " ==> getData()" <<endmsg;
  // get ADC samples after pedestal subtraction
  // it does not matter if one deals with real or
  // simulated data 
  if(!exist<LHCb::VeloTELL1Datas>(m_LCMSADCsLoc)){
    error()<< "There is no subtracted ADC samples in TES!" 
           << "at location: " << m_LCMSADCsLoc <<endmsg;
  }else{
    m_LCMSADCs=get<LHCb::VeloTELL1Datas>(m_LCMSADCsLoc);
  }
  //
  if(!exist<EvtInfos>(m_evtInfoLoc)){
    info()<< " ==> There is no event info at: "
           << m_evtInfoLoc <<endmsg;
  }else{  
    // get data banks from default TES location
    m_evtInfos=get<EvtInfos>(m_evtInfoLoc);
    debug()<< " ==> The EvtInfos have been read-in from location: "
           << m_evtInfoLoc
           << ", size of data container (number of read-out TELL1s): "
           << m_evtInfos->size() <<endmsg;
  }
  //
  return StatusCode::SUCCESS;
}
//=============================================================================
StatusCode VeloLCMSMoni::plotLCMSData()
{
  debug()<< " plotLCMSData() " <<endmsg;

  // One detector, one N-Tuple
  //
  // Ntuple for hits
  Tuple hitNtupleLCMS=nTuple("LCMSHit", "hitsTreeLCMS", CLID_ColumnWiseTuple);

  // Ntuple for events
  Tuple eventNtupleLCMS=nTuple("LCMSEvent", "eventTreeLCMS", CLID_ColumnWiseTuple);



  VeloTELL1Datas::const_iterator sensIt;
  if(m_plotLCMS){

    // Loop over all TELL1s - no effect here.
    for(sensIt=m_LCMSADCs->begin(); sensIt!=m_LCMSADCs->end(); ++sensIt){
      sdataVec ALinkADC;
      scdatIt iT;
      int channel=-1;  // For the very first point we're reading data, the strip on the right is no 0, so need to set this to -1.
      int TELL1=(*sensIt)->key();
      EvtInfo* anInfo=m_evtInfos->object(TELL1);
 
      VeloTELL1Data* adcData=(*sensIt);
      sdataVec data=adcData->data();

      // Get bunch counter 

      m_bunchCounter=anInfo->bunchCounter();


      // Given that we record m_nSamples consecutive samples every time the BAT sees a trigger, we need to keep track of which BAT trigger each sample corresponds to. Within each m_nSamples samples in the raw data, we also add an index. NOTE that within each batch of m_nSamples samples, they are NOT automatically consecutive; at some point, we will need to re-arrange them so that we have m_nSamples consecutive BunchCounter numbers.

      m_BATTrig=(m_eventNumber-1)/m_nSamples; 
      m_orderRaw=m_eventNumber-m_BATTrig*m_nSamples-1; 

      // Then, when we get the first sample of each group of m_nSamples, we want to change the m_counterLoops flag, so that we can create a m_BCntLoops variable where we no longer have the problem of the BunchCounter looping from 4097 to 0.



      if(m_orderRaw==0){
       if((m_bunchCounter>=(4097-m_nSamples))||(m_bunchCounter<m_nSamples)){
         m_doesCounterLoop=true;}
     else{
         m_doesCounterLoop=false;}
         }

     // With the flag set, we can then decide if we need to add 4096 to the low bunch counter nos. In this version we also add 5000 * the BAT trigger no, so that by sorting by this variable we can get all the events in order. After that sorting, we still need to fix the time bins.

     if((m_doesCounterLoop==true)&&(m_bunchCounter<=(2*m_nSamples))){
       m_BCntLoop=m_BATTrig*5000+m_bunchCounter+4096;}
     else{
       m_BCntLoop=m_BATTrig*5000+m_bunchCounter;}

      /// Write the event-related data here

      eventNtupleLCMS->column("Event", m_eventNumber);

      eventNtupleLCMS->column("BATTrig", m_BATTrig);

      eventNtupleLCMS->column("OrderRaw", m_orderRaw);

      eventNtupleLCMS->column("BunchCounter", m_bunchCounter);

      eventNtupleLCMS->column("BCntLoop", m_BCntLoop);

      // Temp storage for max ADC vals (note that we work with magnitude here)
      signed int tempMax=0;

      // Temp storage for max hit position
      signed int tempPosX=-1;

      // Temp storage for no of hits
      unsigned int tempHits=0;

      // Initialise strip variables
      m_totADCValue=0;
      m_hitADCValue=0;
      m_lADCValue=0;
      m_rADCValue=0;

      // Set first and last strip in detector, going by the module and detector nos

      int firstStrip=0;
      int lastStrip=127;

      // Each module has 512 strips plus 64 dummies. Each detector has 128.
      firstStrip=m_moduleNumber*576+m_detectorNumber*128;
      lastStrip=m_moduleNumber*576+m_detectorNumber*128+127;

      //if(m_eventNumber>m_convergenceLimit) contPrinter(data, 15);

      // loop over all strips in each event
      for(int aLink=0; aLink<ALL_ALINKS; ++aLink){
        ALinkADC=(*adcData)[aLink];
        for(iT=ALinkADC.begin(); iT!=ALinkADC.end(); ++iT){
          plot2D((channel+1), (*iT), TELL1,
                 "LCMS ADCs data for TELL1 number "+
                 boost::lexical_cast<std::string>(TELL1),
                 0., 2304, -130., 130., 2304, 260);

	  // Get ADC values. iT actually points to the rightmost strip, and we pass on the old values to update the central and left strips.
	  m_lADCValue=m_hitADCValue;
	  m_hitADCValue=m_rADCValue;
	  m_rADCValue=(*iT);
 	  m_totADCValue=m_hitADCValue+m_lADCValue+m_rADCValue;  // total of strips in cluster
          m_hitPosX=channel;


	  // If we haven't chosen to "writeAll", test that the value on the central strip is higher than the two adjacent ones. This prevents clusters from registering as multiple hits. Note that we work on magnitudes here, by squaring the values before comparison.
	  if((((m_hitADCValue*m_hitADCValue)>(m_lADCValue*m_lADCValue))&&((m_hitADCValue*m_hitADCValue)>(m_rADCValue*m_rADCValue)))||(m_writeAll==1)){

          // apply header cut as appropriate
	  if((channel%32)>=m_headerWidth){

	  // test to see if the strips are in the detector, using firstStrip and lastStrip, and also check threshold. If we chose m_writeAll, no need for threshold. Also, now apply threshold to magnitude of signal by squaring, rather than using absolute value.
          if((((m_hitADCValue*m_hitADCValue)>(m_hitThreshold*m_hitThreshold))||(m_writeAll==1))&&(m_hitADCValue<127)&&(m_hitPosX>firstStrip)&&(m_hitPosX<lastStrip)){
            ++tempHits;
	    if((m_hitADCValue*m_hitADCValue)>(tempMax*tempMax)){tempMax=m_hitADCValue; tempPosX=m_hitPosX;}
            hitNtupleLCMS->column("ADCVal", m_hitADCValue);
            hitNtupleLCMS->column("LVal", m_lADCValue);
            hitNtupleLCMS->column("RVal", m_rADCValue);
            hitNtupleLCMS->column("TotVal", m_totADCValue);
            hitNtupleLCMS->column("PosX", m_hitPosX);
            hitNtupleLCMS->column("BunchCounter", m_bunchCounter);
            hitNtupleLCMS->column("BCntLoop", m_BCntLoop);
            hitNtupleLCMS->column("Event", m_eventNumber);
	    hitNtupleLCMS->column("BATTrig", m_BATTrig);
	    hitNtupleLCMS->column("OrderRaw", m_orderRaw);

            hitNtupleLCMS->write();
	  }
  	  } // close bracket for header test
	  } // close bracket for cluster test

	  // next channel
          channel++;
        }
      }

      // write sums of ADC values and no of hits for each detector


      m_adcMax=tempMax;

      m_numberOfHits=tempHits;

      m_posXMax=tempPosX;

      eventNtupleLCMS->column("ADCMax", m_adcMax);

      eventNtupleLCMS->column("PosXMax", m_posXMax);

      eventNtupleLCMS->column("nHits", m_numberOfHits);

      eventNtupleLCMS->write();

    }
  }

  return ( StatusCode::SUCCESS );
}
//
