// $Id: VeloPedestalSubtractorMoni.cpp,v 1.4 2007/08/26 19:20:05 szumlat Exp $
// Include files 

// from Gaudi
#include "GaudiKernel/DeclareFactoryEntries.h" 

// local
#include "VeloPedestalSubtractorMoni.h"

// kernel
#include "Kernel/DecodeConf.h"
#include "Kernel/DecodeCore.h"
#include "Kernel/ContPrinter.h"

// boost
#include "boost/lexical_cast.hpp"

//-----------------------------------------------------------------------------
// Implementation file for class : VeloPedestalSubtractorMoni
//
// 2006-06-16 : Tomasz Szumlak
//-----------------------------------------------------------------------------

// Declaration of the Algorithm Factory
DECLARE_ALGORITHM_FACTORY( VeloPedestalSubtractorMoni );

using namespace LHCb;
using namespace VeloTELL1;

//=============================================================================
// Standard constructor, initializes variables
//=============================================================================
VeloPedestalSubtractorMoni::VeloPedestalSubtractorMoni( 
                            const std::string& name,ISvcLocator* pSvcLocator)
  : GaudiTupleAlg ( name , pSvcLocator ),
    m_procInfos ( 0 ),
    m_procInfosLoc ( VeloProcessInfoLocation::Default ),
    m_subADCs ( 0 ),
    m_subADCsLoc ( LHCb::VeloTELL1DataLocation::PedSubADCs ),
    m_peds ( 0 ),
    m_pedsLoc ( LHCb::VeloTELL1DataLocation::SubPeds ),
    m_numberOfEvt ( 0 ),
    m_isPed ( false ),
    m_isInitialized ( false ),
    m_hitPosX ( 0 ),
    m_hitADCValue ( 0 ),
    m_hitId ( 0 ),
    m_numberOfHits ( 0 ),
    m_adcSum ( 0 ),
    m_bunchCounter ( 0 ),
    m_evtInfos ( 0 ),
    m_evtInfoLoc ( EvtInfoLocation::Default )
{
  declareProperty("PlotSubADCs", m_plotSubADCs=true);
  declareProperty("HitThreshold", m_hitThreshold=4);
  //declareProperty("PlotPeds", m_plotPeds=true);
}
//=============================================================================
// Destructor
//=============================================================================
VeloPedestalSubtractorMoni::~VeloPedestalSubtractorMoni() {}
//=============================================================================
// Initialization
//=============================================================================
StatusCode VeloPedestalSubtractorMoni::initialize() {
  StatusCode sc = GaudiAlgorithm::initialize(); // must be executed first
  if ( sc.isFailure() ) return sc;  // error printed already by GaudiAlgorithm

  debug() << "==> Initialize" << endmsg;
  setHistoTopDir("Vetra/");
  // 
  return StatusCode::SUCCESS;
}
//=============================================================================
// Main execution
//=============================================================================
StatusCode VeloPedestalSubtractorMoni::execute() {

  debug() << "==> Execute" << endmsg;
  //
  ++m_numberOfEvt;
  if(!m_isInitialized){
    if(!exist<VeloProcessInfos>(m_procInfosLoc)){
      debug()<< " ==> No init info available! " <<endmsg;
    }else{
      m_procInfos=get<VeloProcessInfos>(m_procInfosLoc);
    }
    m_isInitialized=true;
    VeloProcessInfos::const_iterator procIt=m_procInfos->begin();
    if((*procIt)->convLimit()<=0){
      Error("Convergence Limit not set!", StatusCode::FAILURE);
    }else{
      m_convergenceLimit=(*procIt)->convLimit();
    }
  }
  if(m_numberOfEvt>m_convergenceLimit){
    getData();
    if(pedFlag()){
      plotSubtractedData();
      //plotPedestals();
      unsetPedFlag();
    }
  }
  //
  return StatusCode::SUCCESS;
}

//=============================================================================
//  Finalize
//=============================================================================
StatusCode VeloPedestalSubtractorMoni::finalize() {

  debug() << "==> Finalize" << endmsg;

  return GaudiAlgorithm::finalize();  // must be called after all other actions
}
//=============================================================================
StatusCode VeloPedestalSubtractorMoni::getData()
{
  debug()<< " ==> getData()" <<endmsg;
  // get ADC samples after pedestal subtraction
  // it does not matter if one deals with real or
  // simulated data 
  if(!exist<LHCb::VeloTELL1Datas>(m_subADCsLoc)){
    error()<< "There is no subtracted ADC samples in TES!" 
           << "at location: " << m_subADCsLoc <<endmsg;
  }else{
    m_subADCs=get<LHCb::VeloTELL1Datas>(m_subADCsLoc);
    setPedFlag();
  }
  // get subtracted pedestals (determined during pedestal
  // subtraction stage), simulated and real data are treated
  // in the same way
  if(!exist<VeloTELL1Datas>(m_pedsLoc)){
    debug()<< "There is no pedestals in TES!" 
          << "at location: " << m_pedsLoc <<endmsg;
  }else{
    m_peds=get<VeloTELL1Datas>(m_pedsLoc);
  }
  //
  if(!exist<EvtInfos>(m_evtInfoLoc)){
    info()<< " ==> There is no event info at: "
           << m_evtInfoLoc <<endmsg;
  }else{  
    // get data banks from default TES location
    m_evtInfos=get<EvtInfos>(m_evtInfoLoc);
    debug()<< " ==> The EvtInfos have been read-in from location: "
           << m_evtInfoLoc
           << ", size of data container (number of read-out TELL1s): "
           << m_evtInfos->size() <<endmsg;
  }
  //
  return StatusCode::SUCCESS;
}
//=============================================================================
StatusCode VeloPedestalSubtractorMoni::plotSubtractedData()
{
  debug()<< " plotSubtractedData() " <<endmsg;
  //
  // Ntuple for hits
  Tuple hitNtuplePed=nTuple("PedHit", "hitsTreePed", CLID_ColumnWiseTuple);

  // Ntuple for events
  Tuple eventNtuplePed=nTuple("PedEvent", "eventTreePed", CLID_ColumnWiseTuple);

  eventNtuplePed->column("Event", m_numberOfEvt);

  VeloTELL1Datas::const_iterator sensIt;
  if(m_plotSubADCs){
    for(sensIt=m_subADCs->begin(); sensIt!=m_subADCs->end(); ++sensIt){
      sdataVec ALinkADC;
      scdatIt iT;
      unsigned int channel=0;
      int TELL1=(*sensIt)->key();
      EvtInfo* anInfo=m_evtInfos->object(TELL1);
      eventNtuplePed->column("BunchCounter", anInfo->bunchCounter());
      VeloTELL1Data* adcData=(*sensIt);
      sdataVec data=adcData->data();
      signed int tempSum=0;
      unsigned int tempHits=0;
      //if(m_numberOfEvt>m_convergenceLimit) contPrinter(data, 15);
      for(int aLink=0; aLink<ALL_ALINKS; ++aLink){
        ALinkADC=(*adcData)[aLink];
        for(iT=ALinkADC.begin(); iT!=ALinkADC.end(); ++iT){
          plot2D(channel, (*iT), TELL1,
                 "Subtracted ADCs data for TELL1 number "+
                 boost::lexical_cast<std::string>(TELL1),
                 0., 2304, -130., 130., 2304, 260);
          m_hitADCValue=(*iT);
          m_hitPosX=channel;
          if((m_hitADCValue>m_hitThreshold)&&(m_hitADCValue<127)){
            tempSum+=m_hitADCValue;
            ++tempHits;
            hitNtuplePed->column("ADCVal", m_hitADCValue);
            hitNtuplePed->column("PosX", m_hitPosX);
            hitNtuplePed->column("BunchCounter", anInfo->bunchCounter());
            hitNtuplePed->column("Event", m_numberOfEvt);


            hitNtuplePed->write();
	  }
          channel++;
        }
      }
      m_adcSum=tempSum;
      m_numberOfHits=tempHits;
      eventNtuplePed->column("ADCSum", m_adcSum);
      eventNtuplePed->column("nHits", m_numberOfHits);
      eventNtuplePed->write();
    }
  }
  //
  return ( StatusCode::SUCCESS );
}
//=============================================================================
StatusCode VeloPedestalSubtractorMoni::plotPedestals()
{
  debug()<< " plotSubtractedData() " <<endmsg;
  //
  VeloTELL1Datas::const_iterator sensIt;
  if(m_plotPeds){
    for(sensIt=m_peds->begin(); sensIt!=m_peds->end(); ++sensIt){
      sdataVec ALinkADC;
      scdatIt iT;
      unsigned int channel=0;
      int TELL1=(*sensIt)->key()+1000;
      VeloTELL1Data* adcData=(*sensIt);
      for(int aLink=0; aLink<NumberOfALinks; ++aLink){
        ALinkADC=(*adcData)[aLink];
        for(iT=ALinkADC.begin(); iT!=ALinkADC.end(); ++iT){
          plot2D(channel, (*iT), TELL1,
                 "Subtracted pedestals for TELL1 number "+
                 boost::lexical_cast<std::string>(TELL1),
                 0., 2048., 0., 550., 2048, 550);
          channel++;
        }
      }
    }
  }
  //
  return ( StatusCode::SUCCESS );
}
//=============================================================================
void VeloPedestalSubtractorMoni::setPedFlag()
{
  m_isPed=true;
}
//=============================================================================
void VeloPedestalSubtractorMoni::unsetPedFlag()
{
  m_isPed=false;
}
//=============================================================================
bool VeloPedestalSubtractorMoni::pedFlag()
{
  return ( m_isPed );
}
//
