// $Id: VeloTELL1ADCReader.cpp,v 1.2 2007/08/27 10:57:49 szumlat Exp $
// Include files 

// from Gaudi
#include "GaudiKernel/DeclareFactoryEntries.h" 

// local
#include "VeloTELL1ADCReader.h"

// Kernel
#include "Kernel/DecodeConf.h"
#include "Kernel/DecodeCore.h"
#include "Kernel/ContPrinter.h"

// boost
#include "boost/lexical_cast.hpp"

using namespace VeloTELL1;

//-----------------------------------------------------------------------------
// Implementation file for class : VeloTELL1ADCReader
//
// 2006-05-25 : Tomasz Szumlak
//-----------------------------------------------------------------------------

// Declaration of the Algorithm Factory
DECLARE_ALGORITHM_FACTORY( VeloTELL1ADCReader );

//=============================================================================
// Standard constructor, initializes variables
//=============================================================================
VeloTELL1ADCReader::VeloTELL1ADCReader( const std::string& name,
                                        ISvcLocator* pSvcLocator)
  : GaudiTupleAlg ( name , pSvcLocator ),
    m_procInfos ( 0 ),
    m_procInfosLoc ( VeloProcessInfoLocation::Default ),
    m_decodedADCLoc ( VeloTELL1DataLocation::ADCs ),
    m_decodedHeaderLoc ( VeloTELL1DataLocation::Headers ),
    m_evtInfoLoc ( EvtInfoLocation::Default ),
    m_odinLoc ( VeloODINBankLocation::Default ),
    m_isADCBank ( false ),
    m_isEvtInfo ( false ),
    m_eventNumber ( 0 ),
    m_isInitialized ( false ),
    m_hitPosX ( 0 ),
    m_hitADCValue ( 0 ),
    m_hitId ( 0 ),
    m_numberOfHits ( 0 ),
    m_adcSum ( 0 )
{ 
  declareProperty("PrintInfo", m_printInfo=false);
  declareProperty("PlotAllData", m_plotAllData=false);
  declareProperty("PlotALinks", m_plotALinks=false);
  declareProperty("DumpODIN", m_odinDump=false);
  declareProperty("SubtractorType", m_subType=1);
  declareProperty("MaxChannel", m_maxChannel=2304);
  declareProperty("MinChannel", m_minChannel=0);
}
//=============================================================================
// Destructor
//=============================================================================
VeloTELL1ADCReader::~VeloTELL1ADCReader() {}
//=============================================================================
// Initialization
//=============================================================================
StatusCode VeloTELL1ADCReader::initialize() {
  StatusCode sc = GaudiAlgorithm::initialize(); // must be executed first
  if ( sc.isFailure() ) return sc;  // error printed already by GaudiAlgorithm

  debug() << "==> Initialize" << endmsg;
  setHistoTopDir( "Vetra/" );
  //
  return StatusCode::SUCCESS;
}
//=============================================================================
// Main execution
//=============================================================================
StatusCode VeloTELL1ADCReader::execute() {

  debug() << "==> Execute" << endmsg;
  //
  m_eventNumber++;
  if(!m_isInitialized){
    if(!exist<VeloProcessInfos>(m_procInfosLoc)){
      debug()<< " ==> No init info available! " <<endmsg;
    }else{
      m_procInfos=get<VeloProcessInfos>(m_procInfosLoc);
    }
    m_isInitialized=true;
    VeloProcessInfos::const_iterator procIt=m_procInfos->begin();
    if((*procIt)->convLimit()<=0){
      Error("Convergence Limit not set!", StatusCode::FAILURE);
    }else{
      m_convergenceLimit=(*procIt)->convLimit();
    }
  }
  if(m_eventNumber>m_convergenceLimit){
    getData();
    if(adcBankFlag()){
      getAllSensors();
    }
    if(m_odinDump) odinInfo();
    finalProcess();
  }
  //
  return StatusCode::SUCCESS;
}
//=============================================================================
//  Finalize
//=============================================================================
StatusCode VeloTELL1ADCReader::finalize() {

  debug() << "==> Finalize" << endmsg;

  return GaudiAlgorithm::finalize();  // must be called after all other actions
}
//=============================================================================
StatusCode VeloTELL1ADCReader::getData()
{
  debug()<< " ==> getData() " <<endmsg;
  // check and get the decoded ADCs
  if(!exist<VeloTELL1Datas>(m_decodedADCLoc)){
    info()<< " ==> There is no decoded ADCs at: "
           << m_decodedADCLoc <<endmsg;
    return ( StatusCode::FAILURE );
  }else{  
    // get data banks from default TES location
    m_decodedADCs=get<VeloTELL1Datas>(m_decodedADCLoc);
    debug()<< " ==> The data banks have been read-in from location: "
           << m_decodedADCLoc
           << ", size of data container (number of read-out TELL1s): "
           << m_decodedADCs->size() <<endmsg;
    setADCBankFlag();                 // there is data so set the flag!
  }
  // check and get decoded ADC headers
  if(!exist<VeloTELL1Datas>(m_decodedHeaderLoc)){
    info()<< " ==> There is no decoded ADC headers at: "
           << m_decodedHeaderLoc <<endmsg;
  }else{  
    // get data banks from default TES location
    m_decodedHeaders=get<VeloTELL1Datas>(m_decodedHeaderLoc);
    debug()<< " ==> The decoded ADC headers have been read-in from location: "
           << m_decodedHeaderLoc
           << ", size of ADCH container (number of read-out TELL1s): "
           << m_decodedHeaders->size() <<endmsg;
  }
  //
  if(!exist<EvtInfos>(m_evtInfoLoc)){
    info()<< " ==> There is no event info at: "
           << m_evtInfoLoc <<endmsg;
  }else{  
    // get data banks from default TES location
    m_evtInfos=get<EvtInfos>(m_evtInfoLoc);
    debug()<< " ==> The EvtInfos have been read-in from location: "
           << m_evtInfoLoc
           << ", size of data container (number of read-out TELL1s): "
           << m_evtInfos->size() <<endmsg;
    setEvtInfoFlag();                 // there is data so set the flag!
  }
  //
  //if(!exist<VeloODINBanks>(m_odinLoc)){
  //  info()<< " ==> There is no ODIN at: "
  //          << m_odinLoc <<endmsg;
  //}else{
  //    m_odin=get<VeloODINBanks>(m_odinLoc);
  //  debug()<< " ==> odin container size: " << m_odin->size() <<endmsg;
  //  }
  //
  return ( StatusCode::SUCCESS );
}
//=============================================================================
StatusCode VeloTELL1ADCReader::getAllSensors()
{
  debug() << " ==> getAllSensors " <<endmsg;
  //
  // declare iterator for data object with ADC samples 
  // for each sensor
  VeloTELL1Datas::const_iterator sensIt;
  // vector for 32 channels SIGNED values!
  sdataVec ALinkADC;
  sdataVec ALinkADCH;
  // declare iterator for navigate through 
  // vector with channels cdatIt -> signed_const_data_iterator
  scdatIt aLinkItADC;
  scdatIt aLinkItADCH;
  // loop over sensor data
  if(adcBankFlag()){
    for(sensIt=m_decodedADCs->begin(); sensIt!=m_decodedADCs->end(); ++sensIt){
      // temporary object for ADCs 
      VeloTELL1Data* adcData=(*sensIt);
      // make some not very sophisticated plots...
      if(m_plotAllData) plotAll(adcData);
      if(m_plotALinks) plotALinks(adcData);
      const int TELL1No=adcData->key();
      // get headers for the given TELL1 number
      VeloTELL1Data* adcHeaders=m_decodedHeaders->object(TELL1No);
      // check Event info block for the data
      EvtInfo* anInfo=m_evtInfos->object(TELL1No);
      if(m_printInfo&&evtInfoFlag()) dumpEvtInfo(anInfo);
      // get ALinks
      if(m_printInfo)
      info()<< " ==> Decoded data for TELL1 no: " << TELL1No <<endmsg;
      for(int ALinkCounter=0; ALinkCounter<NumberOfALinks; ++ALinkCounter){
        // returns vector with 32 channels (one ALink)
        if(m_printInfo){
          info()<< " ==> Decoded data from ALink [" << ALinkCounter
                << "] " <<endmsg;
          std::cout<< " > ADC samples " <<std::endl;
          //
          ALinkADC=(*adcData)[ALinkCounter];
          ALinkADCH=(*adcHeaders)[ALinkCounter];
          aLinkItADC=ALinkADC.begin();
          aLinkItADCH=ALinkADCH.begin();
          int channelCounter=0;
          // check output
          for(int blocks=0; blocks<32; blocks+=4){
            for(int chanCounter=0; chanCounter<4; ++chanCounter){
              std::cout<< " chan[" << channelCounter << "]: ";
              if(channelCounter<10) std::cout<< " ";
              std::cout<< *(aLinkItADC+chanCounter+blocks) << "\t";
              channelCounter++;
              }
            std::cout<<std::endl;
          }
          //
          std::cout<< " > Decoded ADC Headers: " <<std::endl;
          for(int block=0; block<4; ++block){
            std::cout<< " adcHead[" << block << "]: "
                     << *(aLinkItADCH+block);
          }
          std::cout<<std::endl;
        }
      }
    }
  }
  //
  return ( StatusCode::SUCCESS );
}
//=============================================================================
StatusCode VeloTELL1ADCReader::dumpEvtInfo(EvtInfo*& anInfo)
{
  debug()<< " ==> dumpEvtInfo() " <<endmsg;
  //
  if(anInfo==0){
    info()<< " ==> Event Info block is empty! " <<endmsg;
    return ( StatusCode::FAILURE );
  }
  //
  info()<< " ==> Event Info for TELL1 no: " << anInfo->key() <<endmsg;
  info()<< "detectorID: " << anInfo->detectorID() <<endmsg;
  info()<< " bunchCounter: " << anInfo->bunchCounter() <<endmsg;
  info()<< " l0EventID " << anInfo->l0EventID() <<endmsg;
  //
  return ( StatusCode::SUCCESS );
}
//=============================================================================
StatusCode VeloTELL1ADCReader::plotAll(VeloTELL1Data*& inData)
{
  debug()<< " ==> plotAll() " <<endmsg;
  //
  // Ntuple for hit data - updated whenever we see a hit
  Tuple hitNtuple=nTuple("RawHit", "hitsTree", CLID_ColumnWiseTuple);

  // Ntuple for event data - one entry per event
  Tuple eventNtuple=nTuple("RawEvent", "eventsTree", CLID_ColumnWiseTuple);

  sdataVec ALinkADC;
  scdatIt iT;
  unsigned int channel=0;
  int TELL1=inData->key();
  EvtInfo* anInfo=m_evtInfos->object(TELL1);
  sdataVec data=inData->data();
  signed int tempSum=0;
  unsigned int tempHits=0;
  //if(m_eventNumber>m_convergenceLimit) contPrinter(data, 535);

  // Stick event and bunch counter info into a tree
  eventNtuple->column("Event", m_eventNumber);
  eventNtuple->column("BunchCounter", anInfo->bunchCounter());

  for(int aLink=0; aLink<NumberOfALinks; ++aLink){
    ALinkADC=(*inData)[aLink];
    for(iT=ALinkADC.begin(); iT!=ALinkADC.end(); ++iT){ // inner loop over all channels
      plot2D(channel, (*iT), TELL1,
             "Decoded data for TELL1 number "+
             boost::lexical_cast<std::string>(TELL1),
             0., 2048, 450., 1050., 2048, 600);
      m_hitADCValue=(*iT);
      m_hitPosX=channel;

      // We ignore strips that have a raw ADC value of 1023 (not connected), and also those outside our strip range
      if((m_hitADCValue<1023)&&(m_hitPosX>=m_minChannel)&&(m_hitPosX<=m_maxChannel)){ 
        tempSum+=m_hitADCValue;
        ++tempHits;
	// Write hit data to n-tuple. Also various debugging statements
        //info()<< " A hit is registered! "  <<endmsg;
        hitNtuple->column("ADCVal", m_hitADCValue); 
        //info()<< " ==> Wrote m_hitADCValue " << m_hitADCValue <<endmsg;
        //info()<< " ==> Should match (*iT) " << (*iT) <<endmsg;
        hitNtuple->column("PosX", m_hitPosX);
        //info()<< " ==> Wrote m_hitPosX: " << m_hitPosX <<endmsg;
        //info()<< " ==> Should match channel: " << channel <<endmsg;
        hitNtuple->column("Event", m_eventNumber);
        hitNtuple->column("BunchCounter", anInfo->bunchCounter());

	hitNtuple->write(); 
      }

      // plot2D(channel, (*iT), 2,  // example of another 2D histogram
      //       "Histogram with few bins ",
      //       0., 2048, 450., 1050., 128, 600);

      ++channel;
    }
  }

  // sum of all ADC values
  m_adcSum=tempSum;
  eventNtuple->column("ADCSum", m_adcSum);

  m_numberOfHits=tempHits;
  eventNtuple->column("nHits", m_numberOfHits);
  // Outside loop
  //plot1D( value690, "ch690", 450., 1050., 600);
  eventNtuple->write();
  //
  return ( StatusCode::SUCCESS );
}
//=============================================================================
StatusCode VeloTELL1ADCReader::plotALinks(VeloTELL1Data*& inData)
{
  debug()<< " ==> plotALinks() " <<endmsg;
  //
  sdataVec ALinkADC;
  scdatIt iT;
  int TELL1=inData->key();
  for(int aLink=0; aLink<NumberOfALinks; ++aLink){
    unsigned int channel=0;
    ALinkADC=(*inData)[aLink];
    for(iT=ALinkADC.begin(); iT!=ALinkADC.end(); ++iT){
      plot2D(channel, (*iT), aLink+1000*TELL1,
             "Decoded data for Sensor/ALink number "+
             boost::lexical_cast<std::string>(TELL1)+
             boost::lexical_cast<std::string>("/")+
             boost::lexical_cast<std::string>(aLink),
             0., 32., 350., 650., 32, 350);
      channel++;
    }
  }
  //
  return ( StatusCode::SUCCESS );
}
//=============================================================================
void VeloTELL1ADCReader::setADCBankFlag()
{
  debug()<< " ==> setADCBankFlag() " <<endmsg;
  //
  m_isADCBank=true;
}
//=============================================================================
void VeloTELL1ADCReader::unsetADCBankFlag()
{
  debug()<< " ==> unsetADCBankFlag() " <<endmsg;
  //
  m_isADCBank=false;
}
//=============================================================================
void VeloTELL1ADCReader::setEvtInfoFlag()
{
  debug()<< " ==> setEvtInfoFlag() " <<endmsg;
  //
  m_isEvtInfo=true;
}
//=============================================================================
void VeloTELL1ADCReader::unsetEvtInfoFlag()
{
  debug()<< " ==> unsetEvtInfoFlag() " <<endmsg;
  //
  m_isEvtInfo=false;
}
//=============================================================================
bool VeloTELL1ADCReader::adcBankFlag()
{
  return ( m_isADCBank );
}
//=============================================================================
bool VeloTELL1ADCReader::evtInfoFlag()
{
  return ( m_isEvtInfo );
}
//=============================================================================
void VeloTELL1ADCReader::finalProcess()
{
  debug()<< " ==> finalProcess() " <<endmsg;
  //
  if(adcBankFlag()) unsetADCBankFlag();
  if(evtInfoFlag()) unsetEvtInfoFlag();
}
//=============================================================================
void VeloTELL1ADCReader::odinInfo()
{
  debug()<< " ==> odinInfo() " <<endmsg;
  //
  VeloODINBanks::const_iterator odinIt=m_odin->begin();
  VeloODINBank* odin=(*odinIt);
  info()<< " Size of the ODIN bank: " << odin->odinSize() <<endmsg;
  info()<< " ODIN source ID: " << odin->odinSourceID() <<endmsg;
  info()<< " ODIN version: " << odin->odinVersion() <<endmsg;
  info()<< " ODIN type: " << odin->odinType() <<endmsg;
  info()<< " Run number: " << odin->runNumber() <<endmsg;
  info()<< " Event type: " << odin->eventType() <<endmsg;
}
//--
