// $Id: VeloTELL1ErrorBankReader.cpp,v 1.1 2007/08/26 19:20:05 szumlat Exp $
// Include files 

// from Gaudi
#include "GaudiKernel/DeclareFactoryEntries.h" 

// local
#include "VeloTELL1ErrorBankReader.h"
#include "Kernel/DecodeConf.h"
#include "Kernel/DecodeCore.h"

using namespace VeloTELL1;

//-----------------------------------------------------------------------------
// Implementation file for class : VeloTELL1ErrorBankReader
//
// 2006-05-26 : Tomasz Szumlak
//-----------------------------------------------------------------------------

// Declaration of the Algorithm Factory
DECLARE_ALGORITHM_FACTORY( VeloTELL1ErrorBankReader );


//=============================================================================
// Standard constructor, initializes variables
//=============================================================================
VeloTELL1ErrorBankReader::VeloTELL1ErrorBankReader( const std::string& name,
                                                    ISvcLocator* pSvcLocator)
  : GaudiTupleAlg ( name , pSvcLocator ),
    m_errorBankLoc ( ErrorBankLocation::Default ),
    m_isErrorBank ( false ),
    m_printInfo ( false )
{
  declareProperty("PrintInfo", m_printInfo);
}
//=============================================================================
// Destructor
//=============================================================================
VeloTELL1ErrorBankReader::~VeloTELL1ErrorBankReader() {}
//=============================================================================
// Initialization
//=============================================================================
StatusCode VeloTELL1ErrorBankReader::initialize() {
  StatusCode sc = GaudiAlgorithm::initialize(); // must be executed first
  if ( sc.isFailure() ) return sc;  // error printed already by GaudiAlgorithm

  debug() << "==> Initialize" << endmsg;
  setHistoTopDir( "Vetra/" );
  //
  return StatusCode::SUCCESS;
}
//=============================================================================
// Main execution
//=============================================================================
StatusCode VeloTELL1ErrorBankReader::execute() {

  debug() << "==> Execute" << endmsg;
  //
  getData();
  if(m_isErrorBank){
    getAllSensors();
  }
  //
  return StatusCode::SUCCESS;
}
//=============================================================================
//  Finalize
//=============================================================================
StatusCode VeloTELL1ErrorBankReader::finalize() {

  debug() << "==> Finalize" << endmsg;

  return GaudiAlgorithm::finalize();  // must be called after all other actions
}
//=============================================================================
StatusCode VeloTELL1ErrorBankReader::getData()
{
  debug()<< " ==> getData() " <<endmsg;
  //
  if(!exist<ErrorBanks>(m_errorBankLoc)){
    info()<< " ==> There is no Error Bank at: "
           << m_errorBankLoc <<endmsg;
  }else{  
    // get data banks from default TES location
    m_errorBanks=get<ErrorBanks>(m_errorBankLoc);
    debug()<< " ==> The ErrorBanks have been read-in from location: "
           << m_errorBankLoc
           << ", size of error bank container (number of read-out TELL1s): "
           << m_errorBanks->size() <<endmsg;
    m_isErrorBank=true;
  }
  return ( StatusCode::SUCCESS );
}
//=============================================================================
StatusCode VeloTELL1ErrorBankReader::getAllSensors()
{
  debug()<< " ==> getAllSensores() " <<endmsg;
  //
  if(m_isErrorBank){
    ErrorBanks::const_iterator eBIt;
    //
    for(eBIt=m_errorBanks->begin(); eBIt!=m_errorBanks->end(); eBIt++){
      ErrorBank* aBank=(*eBIt);
      if(!aBank->isEmpty()){
        // check Error Bank content
        dataVec errorSources=aBank->errorSources();
        if(m_printInfo){
          for(unsigned int src=0; src<errorSources.size(); src++){
            info()<< " ==> Decoded Error Bank contents: " <<endmsg;
            info()<< " TELL1 number: " << aBank->key() <<endmsg;
            info()<< " error for PPFPGA = " << errorSources[src] <<endmsg;
            info()<< " l0 from error bank: " 
                  << aBank->l0EventID(errorSources[src]) <<endmsg;
            info()<< " detector ID: " 
                  << aBank->detectorID(errorSources[src]) <<endmsg;
            info()<< " clusterDataSectionLength: "
                  << aBank->clusterDataSectionLength(errorSources[src])
                  <<endmsg;
            info()<< " adcDataSectionLength: "
                  << aBank->adcDataSectionLength(errorSources[src]) <<endmsg;
            info()<< " numberOfCluster: "
                  << aBank->numberOfCluster(errorSources[src]) <<endmsg;
            info()<< " errorBankLength: "
                  << aBank->errorBankLength(errorSources[src]) <<endmsg;
            info()<< " BankLength: "
                  << aBank->bankLength(errorSources[src]) <<endmsg;
          }
        }
        //
        EvtInfo* infoPtr=aBank->evtInfo();
        if(m_printInfo) dumpEvtInfo(infoPtr);
        delete infoPtr;             /// ALWAYS DELETE THE POINTER YOU GET!
      }else{
        info()<< " ==> error bank is empty! :) " <<endmsg;
      }
    }
  }
  //
  return ( StatusCode::SUCCESS );
}
//=============================================================================
StatusCode VeloTELL1ErrorBankReader::dumpEvtInfo(EvtInfo*& anInfo)
{
  debug()<< " ==> dumpEvtInfo() " <<endmsg;
  //
  if(anInfo==0){
    info()<< " ==> Event Info block is empty! " <<endmsg;
    return ( StatusCode::FAILURE );
  }
  //
  info()<< " ==> Event Info for TELL1 no: " << anInfo->key() <<endmsg;
  info()<< "detectorID: " << anInfo->detectorID() <<endmsg;
  info()<< " bunchCounter: " << anInfo->bunchCounter() <<endmsg;
  info()<< " l0EventID " << anInfo->l0EventID() <<endmsg;
  //
  return ( StatusCode::SUCCESS );
}
//
