// $Id: VeloTELL1PedestalsReader.cpp,v 1.1 2007/08/26 19:20:05 szumlat Exp $
// Include files 

// from Gaudi
#include "GaudiKernel/DeclareFactoryEntries.h" 

// local
#include "VeloTELL1PedestalsReader.h"
#include "Kernel/DecodeConf.h"
#include "Kernel/DecodeCore.h"

// boost
#include "boost/lexical_cast.hpp"

using namespace VeloTELL1;

//-----------------------------------------------------------------------------
// Implementation file for class : VeloTELL1PedestalsReader
//
// 2006-05-25 : Tomasz Szumlak
//-----------------------------------------------------------------------------

// Declaration of the Algorithm Factory
DECLARE_ALGORITHM_FACTORY( VeloTELL1PedestalsReader );


//=============================================================================
// Standard constructor, initializes variables
//=============================================================================
VeloTELL1PedestalsReader::VeloTELL1PedestalsReader( const std::string& name,
                                                    ISvcLocator* pSvcLocator)
  : GaudiTupleAlg ( name , pSvcLocator ),
    m_decodedPedLoc ( VeloTELL1DataLocation::Pedestals ),
    m_isPedBank ( false )
{ 
  declareProperty("PrintInfo", m_printInfo=false);
  declareProperty("PlotAllPed", m_plotAllPed=false);
  declareProperty("PlotALinks", m_plotALinks=false);
}
//=============================================================================
// Destructor
//=============================================================================
VeloTELL1PedestalsReader::~VeloTELL1PedestalsReader() {}
//=============================================================================
// Initialization
//=============================================================================
StatusCode VeloTELL1PedestalsReader::initialize() {
  StatusCode sc = GaudiAlgorithm::initialize(); // must be executed first
  if ( sc.isFailure() ) return sc;  // error printed already by GaudiAlgorithm

  debug() << "==> Initialize" << endmsg;
  setHistoTopDir( "Vetra/" );
  //
  return StatusCode::SUCCESS;
}
//=============================================================================
// Main execution
//=============================================================================
StatusCode VeloTELL1PedestalsReader::execute() {

  debug() << "==> Execute" << endmsg;
  //
  getData();
  if(m_isPedBank){
    getAllSensors();
  }
  if(pedBankFlag()) unsetPedBankFlag();
  //
  return StatusCode::SUCCESS;
}
//=============================================================================
//  Finalize
//=============================================================================
StatusCode VeloTELL1PedestalsReader::finalize() {

  debug() << "==> Finalize" << endmsg;

  return GaudiAlgorithm::finalize();  // must be called after all other actions
}
//=============================================================================
StatusCode VeloTELL1PedestalsReader::getData()
{
  debug()<< " ==> getData() " <<endmsg;
  //
  if(!exist<VeloTELL1Datas>(m_decodedPedLoc)){
    info()<< " ==> There is no decoded Pedestals at: "
           << m_decodedPedLoc <<endmsg;
  }else{  
    // get data banks from default TES location
    m_decodedPeds=get<VeloTELL1Datas>(m_decodedPedLoc);
    info()<< " ==> The data banks have been read-in from location: "
           << m_decodedPedLoc
           << ", size of data container (number of read-out TELL1s): "
           << m_decodedPeds->size() <<endmsg;
    setPedBankFlag();
  }
  //
  return ( StatusCode::SUCCESS );
}
//=============================================================================
StatusCode VeloTELL1PedestalsReader::getAllSensors()
{
  debug()<< " ==> getAllSensors() " <<endmsg;
  //
  // get decoded pedestals
  if(pedBankFlag()){
    // iterator for pedestal bank's container
    VeloTELL1Datas::const_iterator sensIt;
    // vector for 32 pedestals (one ALink)
    sdataVec ALink;
    scdatIt aLinkIt;
    //
    for(sensIt=m_decodedPeds->begin(); sensIt!=m_decodedPeds->end(); ++sensIt){
      // one pedestal bank
      VeloTELL1Data* pedData=(*sensIt);
      if(m_plotAllPed) plotAllPed(pedData);
      if(m_plotALinks) plotALinks(pedData);
      const int TELL1No=pedData->key();
      debug()<< " TELL1 number: " << TELL1No <<endmsg;
      if(m_printInfo){
        for(int ALinkCounter=0; ALinkCounter<NumberOfALinks; ++ALinkCounter){
          info()<< " ==> Decoded data from ALink [" << ALinkCounter
                << "] " <<endmsg;
          std::cout<< " > Pedestals " <<std::endl;
          //
          ALink=(*pedData)[ALinkCounter];
          aLinkIt=ALink.begin();
          int channelCounter=0;
          for(int blocks=0; blocks<32; blocks+=4){
            for(int chanCounter=0; chanCounter<4; ++chanCounter){
              std::cout<< " chan[" << channelCounter << "]: ";
              if(channelCounter<10) std::cout<< " ";
                std::cout<< *(aLinkIt+chanCounter+blocks) << "\t";
                channelCounter++;
              }
            std::cout<<std::endl;
          }
        }
      }
    }
  }
  //
  return ( StatusCode::SUCCESS );
}
//=============================================================================
StatusCode VeloTELL1PedestalsReader::plotAllPed(VeloTELL1Data*& inData)
{
  debug()<< " ==> plotAll() " <<endmsg;
  //
  sdataVec ALinkPed;
  scdatIt iT;
  unsigned int channel=0;
  int TELL1=inData->key();
  for(int aLink=0; aLink<NumberOfALinks; ++aLink){
    ALinkPed=(*inData)[aLink];
    for(iT=ALinkPed.begin(); iT!=ALinkPed.end(); ++iT){
      plot2D(channel, (*iT), TELL1,
             "Decoded pedestals for TELL1 number "+
             boost::lexical_cast<std::string>(TELL1),
             0., 2048, 300., 650., 2048, 350);
      channel++;
    }
  }
  //
  return ( StatusCode::SUCCESS );
}
//=============================================================================
StatusCode VeloTELL1PedestalsReader::plotALinks(VeloTELL1Data*& inData)
{
  debug()<< " ==> plotALinks() " <<endmsg;
  //
  sdataVec ALinkPed;
  scdatIt iT;
  int TELL1=inData->key();
  for(int aLink=0; aLink<NumberOfALinks; ++aLink){
    unsigned int channel=0;
    ALinkPed=(*inData)[aLink];
    for(iT=ALinkPed.begin(); iT!=ALinkPed.end(); ++iT){
      plot2D(channel, (*iT), aLink+1000*TELL1,
             "Decoded data for Sensor/ALink number "+
             boost::lexical_cast<std::string>(TELL1)+
             boost::lexical_cast<std::string>("/")+
             boost::lexical_cast<std::string>(aLink),
             0., 32., 350., 650., 32, 350);
      channel++;
    }
  }
  //
  return ( StatusCode::SUCCESS );
}
//=============================================================================
void VeloTELL1PedestalsReader::setPedBankFlag()
{
  m_isPedBank=true;
}
//=============================================================================
void VeloTELL1PedestalsReader::unsetPedBankFlag()
{
  m_isPedBank=false;
}
//=============================================================================
bool VeloTELL1PedestalsReader::pedBankFlag()
{
  return ( m_isPedBank );
}
//
