// $Id: VeloTELL1SimMoni.cpp,v 1.1.1.1 2006/10/24 14:16:17 szumlat Exp $
// Include files 

// from Gaudi
#include "GaudiKernel/AlgFactory.h" 
#include "GaudiAlg/Tuples.h"
#include "GaudiKernel/SmartRefVector.h"
#include "Kernel/VeloChannelID.h"

// local
#include "VeloTELL1SimMoni.h"
#include "Kernel/DecodeCore.h"
#include "Kernel/DecodeConf.h"

// Velo Event classes
#include "Event/MCHit.h"
#include "Event/MCVeloFE.h"
#include "Kernel/IMCVeloFEType.h"

// boost
#include "boost/lexical_cast.hpp"

using namespace Gaudi::Units;
using namespace VeloTELL1;

//-----------------------------------------------------------------------------
// Implementation file for class : VeloTELL1SimMoni
//
// 2005-11-03 : Tomasz Szumlak & Chris Parkes
//-----------------------------------------------------------------------------

DECLARE_ALGORITHM_FACTORY( VeloTELL1SimMoni );

//=============================================================================
// Standard constructor, initializes variables
//=============================================================================
VeloTELL1SimMoni::VeloTELL1SimMoni( const std::string& name,
                          ISvcLocator* pSvcLocator)
  : GaudiTupleAlg ( name , pSvcLocator ),
    m_veloDet( 0 ),
    m_nMCVeloFE( 0 ),
    m_nMCVeloFE2( 0. ),
    m_nMCVeloFEs( 0 ),
    m_nMCVeloFEn( 0 ),
    m_nMCVeloFEo( 0 ),
    m_NumberOfEvents( 0 ),
    m_adcLocation ( VeloTELL1DataLocation::SimADCs ),
    m_pedLocation ( VeloTELL1DataLocation::SimPeds ),
    m_adcs ( 0 ),
    m_peds ( 0 )
{
  declareProperty( "PrintInfo",       m_printInfo = false );
  declareProperty( "DetailedMonitor", m_detailedMonitor = false );
  declareProperty( "VeloSiliconMoni",      m_veloFEMoni = false );
  declareProperty( "NZSDataMoni", m_nzsDataMoni=false );
  declareProperty( "PlotADCs", m_plotADCs=true );
  declareProperty( "PlotPeds", m_plotPeds=false );
}
//=============================================================================
// Destructor
//=============================================================================
VeloTELL1SimMoni::~VeloTELL1SimMoni() {}; 

//=============================================================================
// Initialization
//=============================================================================
StatusCode VeloTELL1SimMoni::initialize() {
  StatusCode sc = GaudiAlgorithm::initialize(); // must be executed first
  if ( sc.isFailure() ) return sc;  // error printed already by GaudiAlgorithm

  debug() << "==> Initialize" << endmsg;
  m_feTypeTool=tool<IMCVeloFEType>("MCVeloFEType/feTypeTool");

  m_veloDet = getDet<DeVelo>( DeVeloLocation::Default );

  setHistoTopDir("Velo/");
  //  
  return StatusCode::SUCCESS;
};

//=============================================================================
// Main execution
//=============================================================================
StatusCode VeloTELL1SimMoni::execute() {

  debug() << "==> Execute" << endmsg;
  //
  m_NumberOfEvents++;
  getData();
  if(m_veloFEMoni){
    VeloFEMonitor();
  }else{
    basicMonitor();
  }
  if(m_nzsDataMoni) VeloTELL1DigiMoni();
  //  
  return StatusCode::SUCCESS;
};

//=============================================================================
//  Finalize
//=============================================================================
StatusCode VeloTELL1SimMoni::finalize() {

  debug() << "==> Finalize" << endmsg;
  //
  double errnMCVeloFE = 0;
  if( 0 != m_NumberOfEvents ) {
    m_nMCVeloFE/=m_NumberOfEvents;
    m_nMCVeloFE2/=m_NumberOfEvents;
    errnMCVeloFE=
      sqrt((m_nMCVeloFE2-(m_nMCVeloFE*m_nMCVeloFE))/m_NumberOfEvents);
  }
  //
  info()<< "------------------------------------------------------" <<endmsg;
  info()<< "                - VeloTELL1SimMoni table -                 " <<endmsg;
  info()<< "------------------------------------------------------" <<endmsg;
  info()<< "| Number of MCVeloFEs/Event: " << m_nMCVeloFE << " +/- " 
        << errnMCVeloFE <<endmsg;
  info()<< "| Number of read-out sensors: " << m_adcs->size() <<endmsg;
  double allFEs=m_nMCVeloFEs+m_nMCVeloFEn+m_nMCVeloFEo;
  if(m_NumberOfEvents>0){
    info().precision(4);
    if(allFEs>0){
    info()<< "| FEs from signal:                      " 
          << (m_nMCVeloFEs/allFEs)*100
          << "%" <<endmsg;
    info()<< "| FEs from noise:                       " 
          << (m_nMCVeloFEn/allFEs)*100
          << "%" <<endmsg;
    info()<< "| FEs from other (spillover/coupling):  " 
          << (m_nMCVeloFEo/allFEs)*100
          << "%" <<endmsg;
    }
  }else{
    info()<< "| ==> No MCVeloFEs found! " <<endmsg;
  }
  info()<< "------------------------------------------------------" <<endmsg;
  //
  return GaudiAlgorithm::finalize();  // must be called after all other actions
}
//=============================================================================
StatusCode VeloTELL1SimMoni::getData()
{
  debug()<< " ==> VeloTELL1SimMoni::getData" <<endmsg;
  // 
  if(!exist<LHCb::MCVeloFEs>(LHCb::MCVeloFELocation::Default)){
    error()<< "There is no VeloFEs in TES!" <<endmsg;
  }else{
    m_mcVFEsCont=get<LHCb::MCVeloFEs>(LHCb::MCVeloFELocation::Default);
  }
  //
  if(!exist<VeloTELL1Datas>(m_adcLocation)){
    info()<< "There is no simulated adcs in TES!" <<endmsg;
  }else{
    m_adcs=get<VeloTELL1Datas>(m_adcLocation);
  }
  //
  if(!exist<VeloTELL1Datas>(m_pedLocation)){
    info()<< "There is no simulated pedestals in TES!" <<endmsg;
  }else{
    m_peds=get<VeloTELL1Datas>(m_pedLocation);
  }
  //
  return StatusCode::SUCCESS;
}
//=============================================================================
StatusCode VeloTELL1SimMoni::VeloFEMonitor()
{
  info()<< " ==> VeloTELL1SimMoni::VeloFEMonitor" <<endmsg;
  //  
  long size=m_mcVFEsCont->size();
  m_nMCVeloFE+=size;
  m_nMCVeloFE2+=double(size*size);
  plot(size, 100, "Number of MCVeloFE/event", 0., 12000., 50);
  // main loop over FEs container
  LHCb::MCVeloFEs::iterator ItFE;
  for(ItFE=m_mcVFEsCont->begin(); ItFE!=m_mcVFEsCont->end(); ItFE++){
    LHCb::MCVeloFE* localFE=(*ItFE);
    int feType=0;
    bool Signal=false, Noise=false, Other=false;
    m_feTypeTool->FEType(localFE, feType);
    if(!feType){ Signal=true; }
    else if(1==feType){ Noise=true; }
    else if(2==feType){ Other=true; }
    if(Signal) { m_nMCVeloFEs++; }
    if(Noise) { m_nMCVeloFEn++; }
    if(Other) { m_nMCVeloFEo++; }
    if(Other){
    debug()<< " sig: " << Signal << ", noise: " << Noise
          << ", so: " << Other <<endmsg;
    debug()<< " local FE strip: " << localFE->strip() <<endmsg;
    }
    // print some info if asked
    if(m_printInfo){
      info()<< " ==> MCVeloFE Monitor: "
            << " ChannelID: " << localFE->channelID()
            << ", sensor number: " << localFE->sensor()
            << ", strip number: " << localFE->strip()
            << "\n \t\t" << " Added noise: " << localFE->addedNoise()
            << ", added CM noise: " << localFE->addedCMNoise()
            << ", added pedestal: " << localFE->addedPedestal()
            << "\n \t\t" << " Added signal: " << localFE->addedSignal()
            << ", charge: " << localFE->charge()
      <<endmsg;
    }
  //
  plot(localFE->charge(), 101, 
       "Charge in Si strip (electrons)",
       300000., 550000., 100);
  plot(localFE->addedSignal(), 102, 
       "Signal deposited in Si strip (electrons)",
       0., 80000., 100);
  plot(localFE->addedSignal(), 142, 
       "Signal deposited in Si strip (electrons) - zoom high values",
       80000., 200000., 200);
  plot(localFE->addedNoise(), 103, 
       "Noise added in Si strip (electrons)",
       -5000., 10000, 100);
  plot(localFE->addedCMNoise(), 113,
       "Common-Mode noise added in Si strip (electrons)",
       -20000., 20000., 100);
  plot(localFE->addedPedestal(), 114,
       "Pedestal added in Si strip (electrons)",
       400000., 420000., 2000);
  plot2D(localFE->strip(), localFE->addedPedestal(), 150,
         "Pedestal and strip number",
         0, 2047, 350000., 500000., 2048, 150);
  plot2D(localFE->sensor(), localFE->strip(), 115,
         "Sensor and strip number",
         0., 140., 0., 2050., 140, 50);
  //
  if(Signal){
    plot(localFE->charge(), 104, 
         "Signal dominated - charge in Si (electrons)",
         300000., 550000., 100);
    plot(localFE->addedSignal(), 105, 
         "Signal domianted - Signal deposited in Si (electrons)",
         0., 50000., 100);
    plot(localFE->addedNoise(), 106,
         "Signal dominated - Noise added in Si strip (electrons)",
         -10000., -10000., 100);
    plot2D(localFE->sensor(), localFE->strip(), 116,
           "Signal dominated - sensor and strip number",
           0., 100., 0., 5000., 100, 50);
  }
  //
  if(Noise){
    plot(localFE->charge(), 107, 
         "Noise dominated - Charge in Si strip (electrons)",
         0., 80000., 100);
    plot(localFE->addedSignal(), 108,
         "Noise dominated - Signal deposited in Si strip (electrons)",
         300000., 550000., 100);
    plot(localFE->addedNoise(), 109,
         "Noise domianted - Noise added in Si strip (electrons)",
         -10000., 10000., 100);
    plot2D(localFE->sensor(), localFE->strip(), 117,
           "Noise dominated - sensor and strip number",
           0., 100., 0., 5000., 100, 50);
  }
  //
  if(Other){
    plot(localFE->charge(), 110,
         "Other dominated - Charge in Si strip (electrons)",
         0., 80000., 100);
    plot(localFE->addedSignal(), 111,
         "Other dominated - Signal deposited in Si strip (electrons)",
         0., 800000., 100);
    plot(localFE->addedNoise(), 112,
         "Other dominated - Noise added in Si strip (electrons)",
         -10000., 10000., 100);
    plot2D(localFE->sensor(), localFE->strip(), 118,
           "Other dominated - sensor and strip number",
           0., 100., 0., 5000., 100, 50);
  }
  //
  // Get access to the MCVeloHits from which MCVeloFE was made, store energy
  //
  SmartRefVector<LHCb::MCHit> myMCVeloHits=localFE->mcHits();
  SmartRefVector<LHCb::MCHit>::iterator ItH;
  std::vector<double> myMCVeloHitsCharge=localFE->mcHitsCharge();
  std::vector<double>::iterator ItCh;
  //
  if(m_printInfo){
    info()<< "Test MCVeloFE: " << "MCVeloHit size:"
          << myMCVeloHits.size() <<endmsg;
  }
  plot(myMCVeloHits.size(), 119,
       "Number of MCVeloHits/MCVeloFE",
       -0.5, 5.5, 6);
  //
  double totalSignal=0.;
  ItCh=myMCVeloHitsCharge.begin();
  for(ItH=myMCVeloHits.begin(); ItH!=myMCVeloHits.end(); ItH++){
    if(m_printInfo){
      info()<< "Test MCVeloFE: " << "MCVeloHit, "
            << "charge in current strip (electrons)" << (*ItCh)
            << ", energy (eV)" << (*ItH)->energy()/eV
            << ", sensor " << (*ItH)->sensDetID() <<endmsg;
    }
    totalSignal+=(*ItCh);
    plot((*ItCh), 120,
         "MC hit signal deposited in MCVeloFE",
         0., 200000., 100);
    ItCh++;
  }
  //
  plot(totalSignal, 121,
       "MC total signal deposited in MCVeloFE",
       0., 50000., 100);
  if(Signal){
    plot(totalSignal, 122,
         "Signal dominated - MC total signal deposited in MCVeloFE",
         0., 100000., 100);
    plot(totalSignal, 140,
         "Signal dominated - MC total signal - zoom low values",
         0., 80000., 100);
    plot(totalSignal, 141,
         "Signal dominated - MC total signal - zoom high values",
         80000., 200000., 150);
  }
  //
  if(Noise){
    plot(totalSignal, 123,
         "Noise dominated - MC total signal deposited in MCVeloFE",
         0., 50000., 100);
  }
  //
  if(Other){
    plot(totalSignal, 124,
         "Other dominated - MC total signal deposited in MCVeloFE",
         0., 50000., 100);
  }
//
// Some other details to check
//
  if(m_detailedMonitor){
  LHCb::VeloChannelID channel=localFE->channelID();
  const DeVeloSensor* sens=m_veloDet->sensor(localFE->channelID().sensor());
  double sensorZ=sens->z()/cm;
  if(m_printInfo){
    info()<< "Channel: " << channel << ", sensor (from channel): "
          << channel.sensor() << ", sensor (from MCVeloFE): "
          << localFE->sensor() << ", Z position: "
          << sensorZ <<endmsg;
  }
  //
  if(sens->isR()){
    const DeVeloRType* rSens=dynamic_cast<const DeVeloRType*>(sens);
    double testRadius=rSens->rOfStrip(channel.strip());
    unsigned int zone=rSens->zoneOfStrip(channel.strip());
    if(m_printInfo){
      info()<< "Sensor: " << channel.sensor()
            << ", strip: " << channel.strip()
            << ", zone: " << zone
            << ", sensorZ: " << sensorZ
            << ", testRadius: " << testRadius
            << ", testRadius/cm: " << testRadius/cm <<endmsg;
    }
    //
    if(0==zone){
      plot2D(sensorZ, testRadius/cm, 125,
             "MCVeloFE R position vs. Z (cm), Zone 0",
             -20., 80., 0., 5., 1000, 50);
    }else if(1==zone){
      plot2D(sensorZ, testRadius/cm, 126,
             "MCVeloFE R position vs. Z (cm), Zone 1",
             -20., 80., 0., 5., 1000, 50);
    }else if(2==zone){
      plot2D(sensorZ, testRadius/cm, 127,
             "MCVeloFE R position vs. Z (cm), Zone 2",
             -20., 80., 0., 5., 1000, 50);
    }else if(3==zone){
      plot2D(sensorZ, testRadius/cm, 128,
             "MCVeloFE R position vs. Z (cm), Zone 3",
             -20., 80., 0., 5., 1000, 50);
    }
  }else if(sens->isPhi()){
    unsigned int strip=channel.strip();
    unsigned int zone=sens->zoneOfStrip(strip);
    const DeVeloPhiType* phiSens=dynamic_cast<const DeVeloPhiType*>(sens);
    double testRadius=phiSens->rMin(zone);
    double testPhi=phiSens->phiOfStrip(strip, 0., testRadius);
    double trgPhi=phiSens->trgPhiOfStrip(strip, 0., testRadius);
    double testPhiDirec=phiSens->angleOfStrip(strip,0.);
    double trgPhiDirec=phiSens->trgPhiDirectionOfStrip(strip,0.);
    if(m_printInfo){
      info()<< "Sensor: " << channel.sensor()
            << ", type: " << (sens->type())
            << ", strip: " << channel.strip()
            << ", sensorZ: " << sensorZ
            << ", testPhi: " << testPhi/degree
            << ", trgPhi: " << trgPhi/degree
            << ", testPhiDirec: " << testPhiDirec/degree
            << ", trgPhiDirec: " << trgPhiDirec/degree
            <<endmsg;
    }
    //
    if(0==zone){
      plot2D(sensorZ, testPhi/degree, 129,
             "MCVeloFE Phi position vs. Z (cm), Inner zone",
             -20., 80., -180., 180., 1000, 60);
      plot2D(sensorZ, trgPhi/degree, 130,
             "MCVeloFE Phi (trg) position vs. Z (cm), Inner zone",
             -20., 80., -180., 180., 1000, 60);
      plot2D(sensorZ, testPhiDirec/degree, 131,
             "MCVeloFE Phi direction vs. Z (cm), Inner zone",
             -20., 80., -180., 180., 1000, 60);
      plot2D(sensorZ, trgPhiDirec/degree, 132,
             "MCVeloFE Phi direction (trg) vs. Z (cm), Inner zone",
             -20., 80., -180., 180., 1000, 60);
    }else{
      plot2D(sensorZ, testPhi/degree, 129,
             "MCVeloFE Phi position vs. Z (cm), Outer zone",
             -20., 80., -180., 180., 1000, 60);
      plot2D(sensorZ, trgPhi/degree, 130,
             "MCVeloFE Phi (trg) position vs. Z (cm), Outer zone",
             -20., 80., -180., 180., 1000, 60);
      plot2D(sensorZ, testPhiDirec/degree, 131,
             "MCVeloFE Phi direction vs. Z (cm), Outer zone",
             -20., 80., -180., 180., 1000, 60);
      plot2D(sensorZ, trgPhiDirec/degree, 132,
             "MCVeloFE Phi direction (trg) vs. Z (cm), Outer zone",
             -20., 80., -180., 180., 1000, 60);
     }
    }
   }
  }
  return StatusCode::SUCCESS;
}
//=============================================================================
StatusCode VeloTELL1SimMoni::VeloTELL1DigiMoni()
{
  debug()<< " ==> VeloTELL1DigiMoni() " <<endmsg;
  //
  VeloTELL1Datas::const_iterator sensIt;
  if(m_plotADCs){
    for(sensIt=m_adcs->begin(); sensIt!=m_adcs->end(); sensIt++){
      sdataVec ALinkADC;
      scdatIt iT;
      unsigned int channel=0;
      int TELL1=(*sensIt)->key();
      VeloTELL1Data* adcData=(*sensIt);
      for(int aLink=0; aLink<NumberOfALinks; aLink++){
        ALinkADC=(*adcData)[aLink];
        for(iT=ALinkADC.begin(); iT!=ALinkADC.end(); iT++){
          plot2D(channel, (*iT), TELL1,
                 "Simulated ADCs for TELL1 number "+
                 boost::lexical_cast<std::string>(TELL1),
                 0., 2048, 450., 1050., 2048, 600);
          channel++;
        }
      }
    }
  }
  // plot the pedestals
  if(m_plotPeds){    
    for(sensIt=m_peds->begin(); sensIt!=m_peds->end(); sensIt++){
      sdataVec ALinkADC;
      scdatIt iT;
      unsigned int channel=0;
      int TELL1=((*sensIt)->key())+1000;
      VeloTELL1Data* adcData=(*sensIt);
      for(int aLink=0; aLink<NumberOfALinks; aLink++){
        ALinkADC=(*adcData)[aLink];
        for(iT=ALinkADC.begin(); iT!=ALinkADC.end(); iT++){
          plot2D(channel, (*iT), TELL1,
                 "Simulated Peds for TELL1 number "+
                 boost::lexical_cast<std::string>(TELL1),
                 0., 2048, 450., 580., 2048, 130);
          channel++;
        }
      }
    }
  }
  //
  return ( StatusCode::SUCCESS );
}
//=============================================================================
StatusCode VeloTELL1SimMoni::basicMonitor()
{
  debug()<< " ==> basicMonitor() " <<endmsg;
  //
    long size=m_mcVFEsCont->size();
  m_nMCVeloFE+=size;
  m_nMCVeloFE2+=double(size*size);
  plot(size, 100, "Number of MCVeloFE/event", 0., 12000., 50);
  // main loop over FEs container
  LHCb::MCVeloFEs::iterator ItFE;
  for(ItFE=m_mcVFEsCont->begin(); ItFE!=m_mcVFEsCont->end(); ItFE++){
    LHCb::MCVeloFE* localFE=(*ItFE);
    int feType=0;
    bool Signal=false, Noise=false, Other=false;
    m_feTypeTool->FEType(localFE, feType);
    if(!feType){ Signal=true; }
    else if(1==feType){ Noise=true; }
    else if(2==feType){ Other=true; }
    if(Signal) { m_nMCVeloFEs++; }
    if(Noise) { m_nMCVeloFEn++; }
    if(Other) { m_nMCVeloFEo++; }
    if(Other){
    debug()<< " sig: " << Signal << ", noise: " << Noise
          << ", so: " << Other <<endmsg;
    debug()<< " local FE strip: " << localFE->strip() <<endmsg;
    }
    // print some info if asked
    if(m_printInfo){
      info()<< " ==> MCVeloFE Monitor: "
            << " ChannelID: " << localFE->channelID()
            << ", sensor number: " << localFE->sensor()
            << ", strip number: " << localFE->strip()
            << "\n \t\t" << " Added noise: " << localFE->addedNoise()
            << ", added CM noise: " << localFE->addedCMNoise()
            << ", added pedestal: " << localFE->addedPedestal()
            << "\n \t\t" << " Added signal: " << localFE->addedSignal()
            << ", charge: " << localFE->charge()
      <<endmsg;
    }
  }
  //
  return ( StatusCode::SUCCESS );
}
